/*
 * Copyright (C) 2004 John Ellis
 * Copyright (C) 2008 - 2016 The Geeqie Team
 *
 * Author: John Ellis
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */

#include <stddef.h>

char *keymap_template [] = {
"<?xml version=\"1.0\" encoding=\"UTF-8\" standalone=\"no\"?>",
"<!-- Created with Inkscape (http://www.inkscape.org/) -->",
"",
"<svg",
"   xmlns:dc=\"http://purl.org/dc/elements/1.1/\"",
"   xmlns:cc=\"http://creativecommons.org/ns#\"",
"   xmlns:rdf=\"http://www.w3.org/1999/02/22-rdf-syntax-ns#\"",
"   xmlns:svg=\"http://www.w3.org/2000/svg\"",
"   xmlns=\"http://www.w3.org/2000/svg\"",
"   xmlns:sodipodi=\"http://sodipodi.sourceforge.net/DTD/sodipodi-0.dtd\"",
"   xmlns:inkscape=\"http://www.inkscape.org/namespaces/inkscape\"",
"   width=\"3000\"",
"   height=\"3000\"",
"   id=\"svg2\"",
"   inkscape:label=\"Pozadí\"",
"   sodipodi:version=\"0.32\"",
"   inkscape:version=\"0.91 r13725\"",
"   sodipodi:docname=\"keymap_template.svg\"",
"   inkscape:output_extension=\"org.inkscape.output.svg.inkscape\"",
"   version=\"1.0\">",
"  <defs",
"     id=\"defs3\">",
"    <inkscape:perspective",
"       sodipodi:type=\"inkscape:persp3d\"",
"       inkscape:vp_x=\"0 : 526.18109 : 1\"",
"       inkscape:vp_y=\"0 : 1000 : 0\"",
"       inkscape:vp_z=\"744.09448 : 526.18109 : 1\"",
"       inkscape:persp3d-origin=\"372.04724 : 350.78739 : 1\"",
"       id=\"perspective9\" />",
"  </defs>",
"  <sodipodi:namedview",
"     id=\"base\"",
"     pagecolor=\"#ffffff\"",
"     bordercolor=\"#666666\"",
"     borderopacity=\"1.0\"",
"     inkscape:pageopacity=\"0\"",
"     inkscape:pageshadow=\"2\"",
"     inkscape:zoom=\"3.959798\"",
"     inkscape:cx=\"1684.6349\"",
"     inkscape:cy=\"1412.463\"",
"     inkscape:document-units=\"px\"",
"     inkscape:current-layer=\"layer1\"",
"     showgrid=\"true\"",
"     inkscape:snap-global=\"true\"",
"     objecttolerance=\"10000\"",
"     gridtolerance=\"10000\"",
"     guidetolerance=\"10000\"",
"     inkscape:window-width=\"1920\"",
"     inkscape:window-height=\"1026\"",
"     inkscape:window-x=\"0\"",
"     inkscape:window-y=\"25\"",
"     inkscape:showpageshadow=\"false\"",
"     inkscape:window-maximized=\"1\">",
"    <inkscape:grid",
"       type=\"xygrid\"",
"       id=\"grid2382\"",
"       visible=\"true\"",
"       enabled=\"true\"",
"       dotted=\"false\"",
"       spacingx=\"1px\"",
"       spacingy=\"1px\"",
"       empspacing=\"10\" />",
"  </sodipodi:namedview>",
"  <metadata",
"     id=\"metadata6\">",
"    <rdf:RDF>",
"      <cc:Work",
"         rdf:about=\"\">",
"        <dc:format>image/svg+xml</dc:format>",
"        <dc:type",
"           rdf:resource=\"http://purl.org/dc/dcmitype/StillImage\" />",
"        <dc:title />",
"      </cc:Work>",
"    </rdf:RDF>",
"  </metadata>",
"  <g",
"     inkscape:label=\"Vrstva 1\"",
"     inkscape:groupmode=\"layer\"",
"     id=\"layer1\">",
"    <rect",
"       style=\"fill:#ffffff;fill-opacity:1;fill-rule:evenodd;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1\"",
"       id=\"rect4343\"",
"       width=\"3000\"",
"       height=\"3000\"",
"       x=\"0\"",
"       y=\"0\" />",
"    <g",
"       id=\"g2481\"",
"       transform=\"translate(450,530)\">",
"      <rect",
"         style=\"fill:#0000ff;fill-opacity:0.1;stroke:#000000;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1\"",
"         id=\"rect2483\"",
"         width=\"170\"",
"         height=\"170\"",
"         x=\"0\"",
"         y=\"2.9921875\" />",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:48px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#000000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         x=\"40\"",
"         y=\"42.992188\"",
"         id=\"text2505\"",
"         sodipodi:linespacing=\"100%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan2507\"",
"           x=\"40\"",
"           y=\"42.992188\">Q</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#000000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"60\"",
"         id=\"text2485\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan2487\"",
"           x=\"5\"",
"           y=\"60\">key:Q</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#ff0000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"75\"",
"         id=\"text2489\"><tspan",
"           sodipodi:role=\"line\"",
"           x=\"5\"",
"           y=\"75\"",
"           id=\"tspan2491\">key:&lt;Primary&gt;Q</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#008000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"90\"",
"         id=\"text2493\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan2495\"",
"           x=\"5\"",
"           y=\"90\">key:&lt;Shift&gt;Q</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#0000ff;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"105\"",
"         id=\"text2497\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan2499\"",
"           x=\"5\"",
"           y=\"105\">key:&lt;Alt&gt;Q</tspan></text>",
"      <text",
"         id=\"text6489\"",
"         y=\"120\"",
"         x=\"5\"",
"         style=\"font-style:normal;font-weight:normal;font-size:12.00000048px;font-family:'Bitstream Vera Sans';fill:#464646;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;-inkscape-font-specification:'Bitstream Vera Sans, Normal';font-stretch:normal;font-variant:normal;text-anchor:start;text-align:start;writing-mode:lr;line-height:125%;\"",
"         xml:space=\"preserve\"",
"         sodipodi:linespacing=\"125%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan6493\">key:&lt;Super&gt;Q</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#808000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"135\"",
"         id=\"text2501\"",
"         sodipodi:linespacing=\"100%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan2503\"",
"           x=\"5\"",
"           y=\"135\">key:&lt;Primary&gt;&lt;Shift&gt;Q</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#ff00ff;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"150\"",
"         id=\"text23\"",
"         sodipodi:linespacing=\"100%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan25\"",
"           x=\"5\"",
"           y=\"150\">key:&lt;Primary&gt;&lt;Alt&gt;Q</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#008080;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"165\"",
"         id=\"text27\"",
"         sodipodi:linespacing=\"100%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan29\"",
"           x=\"5\"",
"           y=\"165\">key:&lt;Shift&gt;&lt;Alt&gt;Q</tspan></text>",
"    </g>",
"    <g",
"       id=\"g2402\"",
"       transform=\"translate(630,530)\">",
"      <rect",
"         y=\"2.9921875\"",
"         x=\"0\"",
"         height=\"170\"",
"         width=\"170\"",
"         id=\"rect2404\"",
"         style=\"fill:#0000ff;fill-opacity:0.1;stroke:#000000;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1\" />",
"      <text",
"         sodipodi:linespacing=\"100%\"",
"         id=\"text2406\"",
"         y=\"42.992188\"",
"         x=\"40\"",
"         style=\"font-size:48px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#000000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"42.992188\"",
"           x=\"40\"",
"           id=\"tspan2408\"",
"           sodipodi:role=\"line\">W</tspan></text>",
"      <text",
"         id=\"text2410\"",
"         y=\"60\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#000000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"60\"",
"           x=\"5\"",
"           id=\"tspan2412\"",
"           sodipodi:role=\"line\">key:W</tspan></text>",
"      <text",
"         id=\"text2414\"",
"         y=\"75\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#ff0000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           id=\"tspan2416\"",
"           y=\"75\"",
"           x=\"5\"",
"           sodipodi:role=\"line\">key:&lt;Primary&gt;W</tspan></text>",
"      <text",
"         id=\"text2418\"",
"         y=\"90\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#008000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"90\"",
"           x=\"5\"",
"           id=\"tspan2420\"",
"           sodipodi:role=\"line\">key:&lt;Shift&gt;W</tspan></text>",
"      <text",
"         id=\"text2422\"",
"         y=\"105\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#0000ff;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"105\"",
"           x=\"5\"",
"           id=\"tspan2424\"",
"           sodipodi:role=\"line\">key:&lt;Alt&gt;W</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-style:normal;font-weight:normal;font-size:12.00000048px;font-family:'Bitstream Vera Sans';fill:#464646;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;-inkscape-font-specification:'Bitstream Vera Sans, Normal';font-stretch:normal;font-variant:normal;text-anchor:start;text-align:start;writing-mode:lr;line-height:125%;\"",
"         x=\"5\"",
"         y=\"120\"",
"         id=\"text6495\"",
"         sodipodi:linespacing=\"125%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan6499\">key:&lt;Super&gt;W</tspan></text>",
"      <text",
"         sodipodi:linespacing=\"100%\"",
"         id=\"text2426\"",
"         y=\"135\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#808000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"135\"",
"           x=\"5\"",
"           id=\"tspan2428\"",
"           sodipodi:role=\"line\">key:&lt;Primary&gt;&lt;Shift&gt;W</tspan></text>",
"      <text",
"         sodipodi:linespacing=\"100%\"",
"         id=\"text2430\"",
"         y=\"150\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#ff00ff;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"150\"",
"           x=\"5\"",
"           id=\"tspan2432\"",
"           sodipodi:role=\"line\">key:&lt;Primary&gt;&lt;Alt&gt;W</tspan></text>",
"      <text",
"         sodipodi:linespacing=\"100%\"",
"         id=\"text2434\"",
"         y=\"165\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#008080;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"165\"",
"           x=\"5\"",
"           id=\"tspan2436\"",
"           sodipodi:role=\"line\">key:&lt;Shift&gt;&lt;Alt&gt;W</tspan></text>",
"    </g>",
"    <g",
"       transform=\"translate(1040,710)\"",
"       id=\"g2438\">",
"      <rect",
"         style=\"fill:#0000ff;fill-opacity:0.1;stroke:#000000;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1\"",
"         id=\"rect2440\"",
"         width=\"170\"",
"         height=\"170\"",
"         x=\"0\"",
"         y=\"2.9921875\" />",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:48px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#000000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         x=\"40\"",
"         y=\"42.992188\"",
"         id=\"text2442\"",
"         sodipodi:linespacing=\"100%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan2444\"",
"           x=\"40\"",
"           y=\"42.992188\">F</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#000000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"60\"",
"         id=\"text2446\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan2448\"",
"           x=\"5\"",
"           y=\"60\">key:F</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#ff0000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"75\"",
"         id=\"text2450\"><tspan",
"           sodipodi:role=\"line\"",
"           x=\"5\"",
"           y=\"75\"",
"           id=\"tspan2452\">key:&lt;Primary&gt;F</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#008000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"90\"",
"         id=\"text2454\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan2456\"",
"           x=\"5\"",
"           y=\"90\">key:&lt;Shift&gt;F</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#0000ff;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"105\"",
"         id=\"text2458\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan2460\"",
"           x=\"5\"",
"           y=\"105\">key:&lt;Alt&gt;F</tspan></text>",
"      <text",
"         id=\"text6579\"",
"         y=\"120\"",
"         x=\"5\"",
"         style=\"font-style:normal;font-weight:normal;font-size:12px;font-family:'Bitstream Vera Sans';fill:#464646;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"105\"",
"           x=\"5\"",
"           id=\"tspan6581\"",
"           sodipodi:role=\"line\">key:&lt;Alt&gt;F</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#808000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"135\"",
"         id=\"text2462\"",
"         sodipodi:linespacing=\"100%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan2464\"",
"           x=\"5\"",
"           y=\"135\">key:&lt;Primary&gt;&lt;Shift&gt;F</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#ff00ff;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"150\"",
"         id=\"text2466\"",
"         sodipodi:linespacing=\"100%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan2468\"",
"           x=\"5\"",
"           y=\"150\">key:&lt;Primary&gt;&lt;Alt&gt;F</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#008080;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"165\"",
"         id=\"text2470\"",
"         sodipodi:linespacing=\"100%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan2472\"",
"           x=\"5\"",
"           y=\"165\">key:&lt;Shift&gt;&lt;Alt&gt;F</tspan></text>",
"    </g>",
"    <g",
"       id=\"g2474\"",
"       transform=\"translate(860,710)\">",
"      <rect",
"         y=\"2.9921875\"",
"         x=\"0\"",
"         height=\"170\"",
"         width=\"170\"",
"         id=\"rect2476\"",
"         style=\"fill:#0000ff;fill-opacity:0.1;stroke:#000000;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1\" />",
"      <text",
"         sodipodi:linespacing=\"100%\"",
"         id=\"text2478\"",
"         y=\"42.992188\"",
"         x=\"40\"",
"         style=\"font-size:48px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#000000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"42.992188\"",
"           x=\"40\"",
"           id=\"tspan2480\"",
"           sodipodi:role=\"line\">D</tspan></text>",
"      <text",
"         id=\"text2482\"",
"         y=\"60\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#000000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"60\"",
"           x=\"5\"",
"           id=\"tspan2484\"",
"           sodipodi:role=\"line\">key:D</tspan></text>",
"      <text",
"         id=\"text2486\"",
"         y=\"75\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#ff0000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           id=\"tspan2488\"",
"           y=\"75\"",
"           x=\"5\"",
"           sodipodi:role=\"line\">key:&lt;Primary&gt;D</tspan></text>",
"      <text",
"         id=\"text2490\"",
"         y=\"90\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#008000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"90\"",
"           x=\"5\"",
"           id=\"tspan2492\"",
"           sodipodi:role=\"line\">key:&lt;Shift&gt;D</tspan></text>",
"      <text",
"         id=\"text2494\"",
"         y=\"105\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#0000ff;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"105\"",
"           x=\"5\"",
"           id=\"tspan2496\"",
"           sodipodi:role=\"line\">key:&lt;Alt&gt;D</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-style:normal;font-weight:normal;font-size:12.00000048px;font-family:'Bitstream Vera Sans';fill:#464646;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;-inkscape-font-specification:'Bitstream Vera Sans, Normal';font-stretch:normal;font-variant:normal;text-anchor:start;text-align:start;writing-mode:lr;line-height:125%;\"",
"         x=\"5\"",
"         y=\"120\"",
"         id=\"text6573\"",
"         sodipodi:linespacing=\"125%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan6577\">key:&lt;Super&gt;D</tspan></text>",
"      <text",
"         sodipodi:linespacing=\"100%\"",
"         id=\"text2498\"",
"         y=\"135\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#808000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"135\"",
"           x=\"5\"",
"           id=\"tspan2500\"",
"           sodipodi:role=\"line\">key:&lt;Primary&gt;&lt;Shift&gt;D</tspan></text>",
"      <text",
"         sodipodi:linespacing=\"100%\"",
"         id=\"text2502\"",
"         y=\"150\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#ff00ff;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"150\"",
"           x=\"5\"",
"           id=\"tspan2504\"",
"           sodipodi:role=\"line\">key:&lt;Primary&gt;&lt;Alt&gt;D</tspan></text>",
"      <text",
"         sodipodi:linespacing=\"100%\"",
"         id=\"text2506\"",
"         y=\"165\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#008080;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"165\"",
"           x=\"5\"",
"           id=\"tspan2508\"",
"           sodipodi:role=\"line\">key:&lt;Shift&gt;&lt;Alt&gt;D</tspan></text>",
"    </g>",
"    <g",
"       transform=\"translate(680,710)\"",
"       id=\"g2510\">",
"      <rect",
"         style=\"fill:#0000ff;fill-opacity:0.1;stroke:#000000;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1\"",
"         id=\"rect2512\"",
"         width=\"170\"",
"         height=\"170\"",
"         x=\"0\"",
"         y=\"2.9921875\" />",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:48px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#000000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         x=\"40\"",
"         y=\"42.992188\"",
"         id=\"text2514\"",
"         sodipodi:linespacing=\"100%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan2516\"",
"           x=\"40\"",
"           y=\"42.992188\">S</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#000000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"60\"",
"         id=\"text2518\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan2520\"",
"           x=\"5\"",
"           y=\"60\">key:S</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#ff0000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"75\"",
"         id=\"text2522\"><tspan",
"           sodipodi:role=\"line\"",
"           x=\"5\"",
"           y=\"75\"",
"           id=\"tspan2524\">key:&lt;Primary&gt;S</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#008000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"90\"",
"         id=\"text2526\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan2528\"",
"           x=\"5\"",
"           y=\"90\">key:&lt;Shift&gt;S</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#0000ff;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"105\"",
"         id=\"text2530\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan2532\"",
"           x=\"5\"",
"           y=\"105\">key:&lt;Alt&gt;S</tspan></text>",
"      <text",
"         id=\"text6567\"",
"         y=\"120\"",
"         x=\"5\"",
"         style=\"font-style:normal;font-weight:normal;font-size:12.00000048px;font-family:'Bitstream Vera Sans';fill:#464646;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;-inkscape-font-specification:'Bitstream Vera Sans, Normal';font-stretch:normal;font-variant:normal;text-anchor:start;text-align:start;writing-mode:lr;line-height:125%;\"",
"         xml:space=\"preserve\"",
"         sodipodi:linespacing=\"125%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan6571\">key:&lt;Super&gt;S</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#808000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"135\"",
"         id=\"text2534\"",
"         sodipodi:linespacing=\"100%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan2536\"",
"           x=\"5\"",
"           y=\"135\">key:&lt;Primary&gt;&lt;Shift&gt;S</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#ff00ff;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"150\"",
"         id=\"text2538\"",
"         sodipodi:linespacing=\"100%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan2540\"",
"           x=\"5\"",
"           y=\"150\">key:&lt;Primary&gt;&lt;Alt&gt;S</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#008080;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"165\"",
"         id=\"text2542\"",
"         sodipodi:linespacing=\"100%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan2544\"",
"           x=\"5\"",
"           y=\"165\">key:&lt;Shift&gt;&lt;Alt&gt;S</tspan></text>",
"    </g>",
"    <g",
"       id=\"g2546\"",
"       transform=\"translate(500,710)\">",
"      <rect",
"         y=\"2.9921875\"",
"         x=\"0\"",
"         height=\"170\"",
"         width=\"170\"",
"         id=\"rect2548\"",
"         style=\"fill:#0000ff;fill-opacity:0.1;stroke:#000000;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1\" />",
"      <text",
"         sodipodi:linespacing=\"100%\"",
"         id=\"text2550\"",
"         y=\"42.992188\"",
"         x=\"40\"",
"         style=\"font-size:48px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#000000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"42.992188\"",
"           x=\"40\"",
"           id=\"tspan2552\"",
"           sodipodi:role=\"line\">A</tspan></text>",
"      <text",
"         id=\"text2554\"",
"         y=\"60\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#000000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"60\"",
"           x=\"5\"",
"           id=\"tspan2556\"",
"           sodipodi:role=\"line\">key:A</tspan></text>",
"      <text",
"         id=\"text2558\"",
"         y=\"75\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#ff0000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           id=\"tspan2560\"",
"           y=\"75\"",
"           x=\"5\"",
"           sodipodi:role=\"line\">key:&lt;Primary&gt;A</tspan></text>",
"      <text",
"         id=\"text2562\"",
"         y=\"90\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#008000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"90\"",
"           x=\"5\"",
"           id=\"tspan2564\"",
"           sodipodi:role=\"line\">key:&lt;Shift&gt;A</tspan></text>",
"      <text",
"         id=\"text2566\"",
"         y=\"105\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#0000ff;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"105\"",
"           x=\"5\"",
"           id=\"tspan2568\"",
"           sodipodi:role=\"line\">key:&lt;Alt&gt;A</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-style:normal;font-weight:normal;font-size:12.00000048px;font-family:'Bitstream Vera Sans';fill:#464646;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;-inkscape-font-specification:'Bitstream Vera Sans, Normal';font-stretch:normal;font-variant:normal;text-anchor:start;text-align:start;writing-mode:lr;line-height:125%;\"",
"         x=\"5\"",
"         y=\"120\"",
"         id=\"text6561\"",
"         sodipodi:linespacing=\"125%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan6565\">key:&lt;Super&gt;A</tspan></text>",
"      <text",
"         sodipodi:linespacing=\"100%\"",
"         id=\"text2570\"",
"         y=\"135\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#808000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"135\"",
"           x=\"5\"",
"           id=\"tspan2572\"",
"           sodipodi:role=\"line\">key:&lt;Primary&gt;&lt;Shift&gt;A</tspan></text>",
"      <text",
"         sodipodi:linespacing=\"100%\"",
"         id=\"text2574\"",
"         y=\"150\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#ff00ff;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"150\"",
"           x=\"5\"",
"           id=\"tspan2576\"",
"           sodipodi:role=\"line\">key:&lt;Primary&gt;&lt;Alt&gt;A</tspan></text>",
"      <text",
"         sodipodi:linespacing=\"100%\"",
"         id=\"text2578\"",
"         y=\"165\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#008080;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"165\"",
"           x=\"5\"",
"           id=\"tspan2580\"",
"           sodipodi:role=\"line\">key:&lt;Shift&gt;&lt;Alt&gt;A</tspan></text>",
"    </g>",
"    <g",
"       transform=\"translate(1530,530)\"",
"       id=\"g2582\">",
"      <rect",
"         style=\"fill:#0000ff;fill-opacity:0.1;stroke:#000000;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1\"",
"         id=\"rect2584\"",
"         width=\"170\"",
"         height=\"170\"",
"         x=\"0\"",
"         y=\"2.9921875\" />",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:48px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#000000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         x=\"40\"",
"         y=\"42.992188\"",
"         id=\"text2586\"",
"         sodipodi:linespacing=\"100%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan2588\"",
"           x=\"40\"",
"           y=\"42.992188\">U</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#000000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"60\"",
"         id=\"text2590\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan2592\"",
"           x=\"5\"",
"           y=\"60\">key:U</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#ff0000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"75\"",
"         id=\"text2594\"><tspan",
"           sodipodi:role=\"line\"",
"           x=\"5\"",
"           y=\"75\"",
"           id=\"tspan2596\">key:&lt;Primary&gt;U</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#008000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"90\"",
"         id=\"text2598\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan2600\"",
"           x=\"5\"",
"           y=\"90\">key:&lt;Shift&gt;U</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#0000ff;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"105\"",
"         id=\"text2602\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan2604\"",
"           x=\"5\"",
"           y=\"105\">key:&lt;Alt&gt;U</tspan></text>",
"      <text",
"         id=\"text6525\"",
"         y=\"120\"",
"         x=\"5\"",
"         style=\"font-style:normal;font-weight:normal;font-size:12.00000048px;font-family:'Bitstream Vera Sans';fill:#464646;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;-inkscape-font-specification:'Bitstream Vera Sans, Normal';font-stretch:normal;font-variant:normal;text-anchor:start;text-align:start;writing-mode:lr;line-height:125%;\"",
"         xml:space=\"preserve\"",
"         sodipodi:linespacing=\"125%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan6529\">key:&lt;Super&gt;U</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#808000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"135\"",
"         id=\"text2606\"",
"         sodipodi:linespacing=\"100%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan2608\"",
"           x=\"5\"",
"           y=\"135\">key:&lt;Primary&gt;&lt;Shift&gt;U</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#ff00ff;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"150\"",
"         id=\"text2610\"",
"         sodipodi:linespacing=\"100%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan2612\"",
"           x=\"5\"",
"           y=\"150\">key:&lt;Primary&gt;&lt;Alt&gt;U</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#008080;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"165\"",
"         id=\"text2614\"",
"         sodipodi:linespacing=\"100%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan2616\"",
"           x=\"5\"",
"           y=\"165\">key:&lt;Shift&gt;&lt;Alt&gt;U</tspan></text>",
"    </g>",
"    <g",
"       id=\"g2618\"",
"       transform=\"translate(1350,530)\">",
"      <rect",
"         y=\"2.9921875\"",
"         x=\"0\"",
"         height=\"170\"",
"         width=\"170\"",
"         id=\"rect2620\"",
"         style=\"fill:#0000ff;fill-opacity:0.1;stroke:#000000;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1\" />",
"      <text",
"         sodipodi:linespacing=\"100%\"",
"         id=\"text2622\"",
"         y=\"42.992188\"",
"         x=\"40\"",
"         style=\"font-size:48px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#000000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"42.992188\"",
"           x=\"40\"",
"           id=\"tspan2624\"",
"           sodipodi:role=\"line\">Y</tspan></text>",
"      <text",
"         id=\"text2626\"",
"         y=\"60\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#000000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"60\"",
"           x=\"5\"",
"           id=\"tspan2628\"",
"           sodipodi:role=\"line\">key:Y</tspan></text>",
"      <text",
"         id=\"text2630\"",
"         y=\"75\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#ff0000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           id=\"tspan2632\"",
"           y=\"75\"",
"           x=\"5\"",
"           sodipodi:role=\"line\">key:&lt;Primary&gt;Y</tspan></text>",
"      <text",
"         id=\"text2634\"",
"         y=\"90\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#008000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"90\"",
"           x=\"5\"",
"           id=\"tspan2636\"",
"           sodipodi:role=\"line\">key:&lt;Shift&gt;Y</tspan></text>",
"      <text",
"         id=\"text2638\"",
"         y=\"105\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#0000ff;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"105\"",
"           x=\"5\"",
"           id=\"tspan2640\"",
"           sodipodi:role=\"line\">key:&lt;Alt&gt;Y</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-style:normal;font-weight:normal;font-size:12.00000048px;font-family:'Bitstream Vera Sans';fill:#464646;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;-inkscape-font-specification:'Bitstream Vera Sans, Normal';font-stretch:normal;font-variant:normal;text-anchor:start;text-align:start;writing-mode:lr;line-height:125%;\"",
"         x=\"5\"",
"         y=\"120\"",
"         id=\"text6519\"",
"         sodipodi:linespacing=\"125%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan6523\">key:&lt;Super&gt;Y</tspan></text>",
"      <text",
"         sodipodi:linespacing=\"100%\"",
"         id=\"text2642\"",
"         y=\"135\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#808000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"135\"",
"           x=\"5\"",
"           id=\"tspan2644\"",
"           sodipodi:role=\"line\">key:&lt;Primary&gt;&lt;Shift&gt;Y</tspan></text>",
"      <text",
"         sodipodi:linespacing=\"100%\"",
"         id=\"text2646\"",
"         y=\"150\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#ff00ff;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"150\"",
"           x=\"5\"",
"           id=\"tspan2648\"",
"           sodipodi:role=\"line\">key:&lt;Primary&gt;&lt;Alt&gt;Y</tspan></text>",
"      <text",
"         sodipodi:linespacing=\"100%\"",
"         id=\"text2650\"",
"         y=\"165\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#008080;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"165\"",
"           x=\"5\"",
"           id=\"tspan2652\"",
"           sodipodi:role=\"line\">key:&lt;Shift&gt;&lt;Alt&gt;Y</tspan></text>",
"    </g>",
"    <g",
"       transform=\"translate(1170,530)\"",
"       id=\"g2654\">",
"      <rect",
"         style=\"fill:#0000ff;fill-opacity:0.1;stroke:#000000;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1\"",
"         id=\"rect2656\"",
"         width=\"170\"",
"         height=\"170\"",
"         x=\"0\"",
"         y=\"2.9921875\" />",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:48px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#000000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         x=\"40\"",
"         y=\"42.992188\"",
"         id=\"text2658\"",
"         sodipodi:linespacing=\"100%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan2660\"",
"           x=\"40\"",
"           y=\"42.992188\">T</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#000000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"60\"",
"         id=\"text2662\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan2664\"",
"           x=\"5\"",
"           y=\"60\">key:T</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#ff0000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"75\"",
"         id=\"text2666\"><tspan",
"           sodipodi:role=\"line\"",
"           x=\"5\"",
"           y=\"75\"",
"           id=\"tspan2668\">key:&lt;Primary&gt;T</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#008000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"90\"",
"         id=\"text2670\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan2672\"",
"           x=\"5\"",
"           y=\"90\">key:&lt;Shift&gt;T</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#0000ff;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"105\"",
"         id=\"text2674\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan2676\"",
"           x=\"5\"",
"           y=\"105\">key:&lt;Alt&gt;T</tspan></text>",
"      <text",
"         id=\"text6513\"",
"         y=\"120\"",
"         x=\"5\"",
"         style=\"font-style:normal;font-weight:normal;font-size:12.00000048px;font-family:'Bitstream Vera Sans';fill:#464646;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;-inkscape-font-specification:'Bitstream Vera Sans, Normal';font-stretch:normal;font-variant:normal;text-anchor:start;text-align:start;writing-mode:lr;line-height:125%;\"",
"         xml:space=\"preserve\"",
"         sodipodi:linespacing=\"125%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan6517\">key:&lt;Super&gt;T</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#808000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"135\"",
"         id=\"text2678\"",
"         sodipodi:linespacing=\"100%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan2680\"",
"           x=\"5\"",
"           y=\"135\">key:&lt;Primary&gt;&lt;Shift&gt;T</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#ff00ff;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"150\"",
"         id=\"text2682\"",
"         sodipodi:linespacing=\"100%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan2684\"",
"           x=\"5\"",
"           y=\"150\">key:&lt;Primary&gt;&lt;Alt&gt;T</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#008080;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"165\"",
"         id=\"text2686\"",
"         sodipodi:linespacing=\"100%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan2688\"",
"           x=\"5\"",
"           y=\"165\">key:&lt;Shift&gt;&lt;Alt&gt;T</tspan></text>",
"    </g>",
"    <g",
"       id=\"g2690\"",
"       transform=\"translate(990,530)\">",
"      <rect",
"         y=\"2.9921875\"",
"         x=\"0\"",
"         height=\"170\"",
"         width=\"170\"",
"         id=\"rect2692\"",
"         style=\"fill:#0000ff;fill-opacity:0.1;stroke:#000000;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1\" />",
"      <text",
"         sodipodi:linespacing=\"100%\"",
"         id=\"text2694\"",
"         y=\"42.992188\"",
"         x=\"40\"",
"         style=\"font-size:48px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#000000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"42.992188\"",
"           x=\"40\"",
"           id=\"tspan2696\"",
"           sodipodi:role=\"line\">R</tspan></text>",
"      <text",
"         id=\"text2698\"",
"         y=\"60\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#000000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"60\"",
"           x=\"5\"",
"           id=\"tspan2700\"",
"           sodipodi:role=\"line\">key:R</tspan></text>",
"      <text",
"         id=\"text2702\"",
"         y=\"75\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#ff0000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           id=\"tspan2704\"",
"           y=\"75\"",
"           x=\"5\"",
"           sodipodi:role=\"line\">key:&lt;Primary&gt;R</tspan></text>",
"      <text",
"         id=\"text2706\"",
"         y=\"90\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#008000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"90\"",
"           x=\"5\"",
"           id=\"tspan2708\"",
"           sodipodi:role=\"line\">key:&lt;Shift&gt;R</tspan></text>",
"      <text",
"         id=\"text2710\"",
"         y=\"105\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#0000ff;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"105\"",
"           x=\"5\"",
"           id=\"tspan2712\"",
"           sodipodi:role=\"line\">key:&lt;Alt&gt;R</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-style:normal;font-weight:normal;font-size:12.00000048px;font-family:'Bitstream Vera Sans';fill:#464646;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;-inkscape-font-specification:'Bitstream Vera Sans, Normal';font-stretch:normal;font-variant:normal;text-anchor:start;text-align:start;writing-mode:lr;line-height:125%;\"",
"         x=\"5\"",
"         y=\"120\"",
"         id=\"text6507\"",
"         sodipodi:linespacing=\"125%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan6511\">key:&lt;Super&gt;R</tspan></text>",
"      <text",
"         sodipodi:linespacing=\"100%\"",
"         id=\"text2714\"",
"         y=\"135\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#808000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"135\"",
"           x=\"5\"",
"           id=\"tspan2716\"",
"           sodipodi:role=\"line\">key:&lt;Primary&gt;&lt;Shift&gt;R</tspan></text>",
"      <text",
"         sodipodi:linespacing=\"100%\"",
"         id=\"text2718\"",
"         y=\"150\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#ff00ff;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"150\"",
"           x=\"5\"",
"           id=\"tspan2720\"",
"           sodipodi:role=\"line\">key:&lt;Primary&gt;&lt;Alt&gt;R</tspan></text>",
"      <text",
"         sodipodi:linespacing=\"100%\"",
"         id=\"text2722\"",
"         y=\"165\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#008080;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"165\"",
"           x=\"5\"",
"           id=\"tspan2724\"",
"           sodipodi:role=\"line\">key:&lt;Shift&gt;&lt;Alt&gt;R</tspan></text>",
"    </g>",
"    <g",
"       transform=\"translate(810,530)\"",
"       id=\"g2726\">",
"      <rect",
"         style=\"fill:#0000ff;fill-opacity:0.1;stroke:#000000;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1\"",
"         id=\"rect2728\"",
"         width=\"170\"",
"         height=\"170\"",
"         x=\"0\"",
"         y=\"2.9921875\" />",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:48px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#000000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         x=\"40\"",
"         y=\"42.992188\"",
"         id=\"text2730\"",
"         sodipodi:linespacing=\"100%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan2732\"",
"           x=\"40\"",
"           y=\"42.992188\">E</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#000000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"60\"",
"         id=\"text2734\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan2736\"",
"           x=\"5\"",
"           y=\"60\">key:E</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#ff0000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"75\"",
"         id=\"text2738\"><tspan",
"           sodipodi:role=\"line\"",
"           x=\"5\"",
"           y=\"75\"",
"           id=\"tspan2740\">key:&lt;Primary&gt;E</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#008000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"90\"",
"         id=\"text2742\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan2744\"",
"           x=\"5\"",
"           y=\"90\">key:&lt;Shift&gt;E</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#0000ff;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"105\"",
"         id=\"text2746\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan2748\"",
"           x=\"5\"",
"           y=\"105\">key:&lt;Alt&gt;E</tspan></text>",
"      <text",
"         id=\"text6501\"",
"         y=\"120\"",
"         x=\"5\"",
"         style=\"font-style:normal;font-weight:normal;font-size:12.00000048px;font-family:'Bitstream Vera Sans';fill:#464646;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;-inkscape-font-specification:'Bitstream Vera Sans, Normal';font-stretch:normal;font-variant:normal;text-anchor:start;text-align:start;writing-mode:lr;line-height:125%;\"",
"         xml:space=\"preserve\"",
"         sodipodi:linespacing=\"125%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan6505\">key:&lt;Super&gt;E</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#808000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"135\"",
"         id=\"text2750\"",
"         sodipodi:linespacing=\"100%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan2752\"",
"           x=\"5\"",
"           y=\"135\">key:&lt;Primary&gt;&lt;Shift&gt;E</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#ff00ff;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"150\"",
"         id=\"text2754\"",
"         sodipodi:linespacing=\"100%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan2756\"",
"           x=\"5\"",
"           y=\"150\">key:&lt;Primary&gt;&lt;Alt&gt;E</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#008080;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"165\"",
"         id=\"text2758\"",
"         sodipodi:linespacing=\"100%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan2760\"",
"           x=\"5\"",
"           y=\"165\">key:&lt;Shift&gt;&lt;Alt&gt;E</tspan></text>",
"    </g>",
"    <g",
"       id=\"g2762\"",
"       transform=\"translate(1580,710)\">",
"      <rect",
"         y=\"2.9921875\"",
"         x=\"0\"",
"         height=\"170\"",
"         width=\"170\"",
"         id=\"rect2764\"",
"         style=\"fill:#0000ff;fill-opacity:0.1;stroke:#000000;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1\" />",
"      <text",
"         sodipodi:linespacing=\"100%\"",
"         id=\"text2766\"",
"         y=\"42.992188\"",
"         x=\"40\"",
"         style=\"font-size:48px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#000000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"42.992188\"",
"           x=\"40\"",
"           id=\"tspan2768\"",
"           sodipodi:role=\"line\">J</tspan></text>",
"      <text",
"         id=\"text2770\"",
"         y=\"60\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#000000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"60\"",
"           x=\"5\"",
"           id=\"tspan2772\"",
"           sodipodi:role=\"line\">key:J</tspan></text>",
"      <text",
"         id=\"text2774\"",
"         y=\"75\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#ff0000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           id=\"tspan2776\"",
"           y=\"75\"",
"           x=\"5\"",
"           sodipodi:role=\"line\">key:&lt;Primary&gt;J</tspan></text>",
"      <text",
"         id=\"text2778\"",
"         y=\"90\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#008000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"90\"",
"           x=\"5\"",
"           id=\"tspan2780\"",
"           sodipodi:role=\"line\">key:&lt;Shift&gt;J</tspan></text>",
"      <text",
"         id=\"text2782\"",
"         y=\"105\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#0000ff;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"105\"",
"           x=\"5\"",
"           id=\"tspan2784\"",
"           sodipodi:role=\"line\">key:&lt;Alt&gt;J</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-style:normal;font-weight:normal;font-size:12.00000048px;font-family:'Bitstream Vera Sans';fill:#464646;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;-inkscape-font-specification:'Bitstream Vera Sans, Normal';font-stretch:normal;font-variant:normal;text-anchor:start;text-align:start;writing-mode:lr;line-height:125%;\"",
"         x=\"5\"",
"         y=\"120\"",
"         id=\"text6595\"",
"         sodipodi:linespacing=\"125%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan6599\">key:&lt;Super&gt;J</tspan></text>",
"      <text",
"         sodipodi:linespacing=\"100%\"",
"         id=\"text2786\"",
"         y=\"135\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#808000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"135\"",
"           x=\"5\"",
"           id=\"tspan2788\"",
"           sodipodi:role=\"line\">key:&lt;Primary&gt;&lt;Shift&gt;J</tspan></text>",
"      <text",
"         sodipodi:linespacing=\"100%\"",
"         id=\"text2790\"",
"         y=\"150\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#ff00ff;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"150\"",
"           x=\"5\"",
"           id=\"tspan2792\"",
"           sodipodi:role=\"line\">key:&lt;Primary&gt;&lt;Alt&gt;J</tspan></text>",
"      <text",
"         sodipodi:linespacing=\"100%\"",
"         id=\"text2794\"",
"         y=\"165\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#008080;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"165\"",
"           x=\"5\"",
"           id=\"tspan2796\"",
"           sodipodi:role=\"line\">key:&lt;Shift&gt;&lt;Alt&gt;J</tspan></text>",
"    </g>",
"    <g",
"       transform=\"translate(1400,710)\"",
"       id=\"g2798\">",
"      <rect",
"         style=\"fill:#0000ff;fill-opacity:0.1;stroke:#000000;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1\"",
"         id=\"rect2800\"",
"         width=\"170\"",
"         height=\"170\"",
"         x=\"0\"",
"         y=\"2.9921875\" />",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:48px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#000000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         x=\"40\"",
"         y=\"42.992188\"",
"         id=\"text2802\"",
"         sodipodi:linespacing=\"100%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan2804\"",
"           x=\"40\"",
"           y=\"42.992188\">H</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#000000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"60\"",
"         id=\"text2806\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan2808\"",
"           x=\"5\"",
"           y=\"60\">key:H</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#ff0000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"75\"",
"         id=\"text2810\"><tspan",
"           sodipodi:role=\"line\"",
"           x=\"5\"",
"           y=\"75\"",
"           id=\"tspan2812\">key:&lt;Primary&gt;H</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#008000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"90\"",
"         id=\"text2814\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan2816\"",
"           x=\"5\"",
"           y=\"90\">key:&lt;Shift&gt;H</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#0000ff;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"105\"",
"         id=\"text2818\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan2820\"",
"           x=\"5\"",
"           y=\"105\">key:&lt;Alt&gt;H</tspan></text>",
"      <text",
"         id=\"text6589\"",
"         y=\"120\"",
"         x=\"5\"",
"         style=\"font-style:normal;font-weight:normal;font-size:12.00000048px;font-family:'Bitstream Vera Sans';fill:#464646;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;-inkscape-font-specification:'Bitstream Vera Sans, Normal';font-stretch:normal;font-variant:normal;text-anchor:start;text-align:start;writing-mode:lr;line-height:125%;\"",
"         xml:space=\"preserve\"",
"         sodipodi:linespacing=\"125%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan6593\">key:&lt;Super&gt;H</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#808000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"135\"",
"         id=\"text2822\"",
"         sodipodi:linespacing=\"100%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan2824\"",
"           x=\"5\"",
"           y=\"135\">key:&lt;Primary&gt;&lt;Shift&gt;H</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#ff00ff;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"150\"",
"         id=\"text2826\"",
"         sodipodi:linespacing=\"100%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan2828\"",
"           x=\"5\"",
"           y=\"150\">key:&lt;Primary&gt;&lt;Alt&gt;H</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#008080;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"165\"",
"         id=\"text2830\"",
"         sodipodi:linespacing=\"100%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan2832\"",
"           x=\"5\"",
"           y=\"165\">key:&lt;Shift&gt;&lt;Alt&gt;H</tspan></text>",
"    </g>",
"    <g",
"       id=\"g2834\"",
"       transform=\"translate(1220,710)\">",
"      <rect",
"         y=\"2.9921875\"",
"         x=\"0\"",
"         height=\"170\"",
"         width=\"170\"",
"         id=\"rect2836\"",
"         style=\"fill:#0000ff;fill-opacity:0.1;stroke:#000000;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1\" />",
"      <text",
"         sodipodi:linespacing=\"100%\"",
"         id=\"text2838\"",
"         y=\"42.992188\"",
"         x=\"40\"",
"         style=\"font-size:48px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#000000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"42.992188\"",
"           x=\"40\"",
"           id=\"tspan2840\"",
"           sodipodi:role=\"line\">G</tspan></text>",
"      <text",
"         id=\"text2842\"",
"         y=\"60\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#000000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"60\"",
"           x=\"5\"",
"           id=\"tspan2844\"",
"           sodipodi:role=\"line\">key:G</tspan></text>",
"      <text",
"         id=\"text2846\"",
"         y=\"75\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#ff0000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           id=\"tspan2848\"",
"           y=\"75\"",
"           x=\"5\"",
"           sodipodi:role=\"line\">key:&lt;Primary&gt;G</tspan></text>",
"      <text",
"         id=\"text2850\"",
"         y=\"90\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#008000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"90\"",
"           x=\"5\"",
"           id=\"tspan2852\"",
"           sodipodi:role=\"line\">key:&lt;Shift&gt;G</tspan></text>",
"      <text",
"         id=\"text2854\"",
"         y=\"105\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#0000ff;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"105\"",
"           x=\"5\"",
"           id=\"tspan2856\"",
"           sodipodi:role=\"line\">key:&lt;Alt&gt;G</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-style:normal;font-weight:normal;font-size:12.00000048px;font-family:'Bitstream Vera Sans';fill:#464646;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;-inkscape-font-specification:'Bitstream Vera Sans, Normal';font-stretch:normal;font-variant:normal;text-anchor:start;text-align:start;writing-mode:lr;line-height:125%;\"",
"         x=\"5\"",
"         y=\"120\"",
"         id=\"text6583\"",
"         sodipodi:linespacing=\"125%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan6587\">key:&lt;Super&gt;G</tspan></text>",
"      <text",
"         sodipodi:linespacing=\"100%\"",
"         id=\"text2858\"",
"         y=\"135\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#808000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"135\"",
"           x=\"5\"",
"           id=\"tspan2860\"",
"           sodipodi:role=\"line\">key:&lt;Primary&gt;&lt;Shift&gt;G</tspan></text>",
"      <text",
"         sodipodi:linespacing=\"100%\"",
"         id=\"text2862\"",
"         y=\"150\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#ff00ff;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"150\"",
"           x=\"5\"",
"           id=\"tspan2864\"",
"           sodipodi:role=\"line\">key:&lt;Primary&gt;&lt;Alt&gt;G</tspan></text>",
"      <text",
"         sodipodi:linespacing=\"100%\"",
"         id=\"text2866\"",
"         y=\"165\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#008080;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"165\"",
"           x=\"5\"",
"           id=\"tspan2868\"",
"           sodipodi:role=\"line\">key:&lt;Shift&gt;&lt;Alt&gt;G</tspan></text>",
"    </g>",
"    <g",
"       transform=\"translate(1130,890)\"",
"       id=\"g2870\">",
"      <rect",
"         style=\"fill:#0000ff;fill-opacity:0.1;stroke:#000000;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1\"",
"         id=\"rect2872\"",
"         width=\"170\"",
"         height=\"170\"",
"         x=\"0\"",
"         y=\"2.9921875\" />",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:48px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#000000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         x=\"40\"",
"         y=\"42.992188\"",
"         id=\"text2874\"",
"         sodipodi:linespacing=\"100%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan2876\"",
"           x=\"40\"",
"           y=\"42.992188\">V</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#000000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"60\"",
"         id=\"text2878\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan2880\"",
"           x=\"5\"",
"           y=\"60\">key:V</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#ff0000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"75\"",
"         id=\"text2882\"><tspan",
"           sodipodi:role=\"line\"",
"           x=\"5\"",
"           y=\"75\"",
"           id=\"tspan2884\">key:&lt;Primary&gt;V</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#008000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"90\"",
"         id=\"text2886\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan2888\"",
"           x=\"5\"",
"           y=\"90\">key:&lt;Shift&gt;V</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#0000ff;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"105\"",
"         id=\"text2890\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan2892\"",
"           x=\"5\"",
"           y=\"105\">key:&lt;Alt&gt;V</tspan></text>",
"      <text",
"         id=\"text6655\"",
"         y=\"120\"",
"         x=\"5\"",
"         style=\"font-style:normal;font-weight:normal;font-size:12.00000048px;font-family:'Bitstream Vera Sans';fill:#464646;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;-inkscape-font-specification:'Bitstream Vera Sans, Normal';font-stretch:normal;font-variant:normal;text-anchor:start;text-align:start;writing-mode:lr;line-height:125%;\"",
"         xml:space=\"preserve\"",
"         sodipodi:linespacing=\"125%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan6659\">key:&lt;Super&gt;V</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#808000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"135\"",
"         id=\"text2894\"",
"         sodipodi:linespacing=\"100%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan2896\"",
"           x=\"5\"",
"           y=\"135\">key:&lt;Primary&gt;&lt;Shift&gt;V</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#ff00ff;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"150\"",
"         id=\"text2898\"",
"         sodipodi:linespacing=\"100%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan2900\"",
"           x=\"5\"",
"           y=\"150\">key:&lt;Primary&gt;&lt;Alt&gt;V</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#008080;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"165\"",
"         id=\"text2902\"",
"         sodipodi:linespacing=\"100%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan2904\"",
"           x=\"5\"",
"           y=\"165\">key:&lt;Shift&gt;&lt;Alt&gt;V</tspan></text>",
"    </g>",
"    <g",
"       id=\"g2906\"",
"       transform=\"translate(950,890)\">",
"      <rect",
"         y=\"2.9921875\"",
"         x=\"0\"",
"         height=\"170\"",
"         width=\"170\"",
"         id=\"rect2908\"",
"         style=\"fill:#0000ff;fill-opacity:0.1;stroke:#000000;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1\" />",
"      <text",
"         sodipodi:linespacing=\"100%\"",
"         id=\"text2910\"",
"         y=\"42.992188\"",
"         x=\"40\"",
"         style=\"font-size:48px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#000000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"42.992188\"",
"           x=\"40\"",
"           id=\"tspan2912\"",
"           sodipodi:role=\"line\">C</tspan></text>",
"      <text",
"         id=\"text2914\"",
"         y=\"60\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#000000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"60\"",
"           x=\"5\"",
"           id=\"tspan2916\"",
"           sodipodi:role=\"line\">key:C</tspan></text>",
"      <text",
"         id=\"text2918\"",
"         y=\"75\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#ff0000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           id=\"tspan2920\"",
"           y=\"75\"",
"           x=\"5\"",
"           sodipodi:role=\"line\">key:&lt;Primary&gt;C</tspan></text>",
"      <text",
"         id=\"text2922\"",
"         y=\"90\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#008000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"90\"",
"           x=\"5\"",
"           id=\"tspan2924\"",
"           sodipodi:role=\"line\">key:&lt;Shift&gt;C</tspan></text>",
"      <text",
"         id=\"text2926\"",
"         y=\"105\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#0000ff;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"105\"",
"           x=\"5\"",
"           id=\"tspan2928\"",
"           sodipodi:role=\"line\">key:&lt;Alt&gt;C</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-style:normal;font-weight:normal;font-size:12.00000048px;font-family:'Bitstream Vera Sans';fill:#464646;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;-inkscape-font-specification:'Bitstream Vera Sans, Normal';font-stretch:normal;font-variant:normal;text-anchor:start;text-align:start;writing-mode:lr;line-height:125%;\"",
"         x=\"5\"",
"         y=\"120\"",
"         id=\"text6649\"",
"         sodipodi:linespacing=\"125%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan6653\">key:&lt;Super&gt;C</tspan></text>",
"      <text",
"         sodipodi:linespacing=\"100%\"",
"         id=\"text2930\"",
"         y=\"135\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#808000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"135\"",
"           x=\"5\"",
"           id=\"tspan2932\"",
"           sodipodi:role=\"line\">key:&lt;Primary&gt;&lt;Shift&gt;C</tspan></text>",
"      <text",
"         sodipodi:linespacing=\"100%\"",
"         id=\"text2934\"",
"         y=\"150\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#ff00ff;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"150\"",
"           x=\"5\"",
"           id=\"tspan2936\"",
"           sodipodi:role=\"line\">key:&lt;Primary&gt;&lt;Alt&gt;C</tspan></text>",
"      <text",
"         sodipodi:linespacing=\"100%\"",
"         id=\"text2938\"",
"         y=\"165\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#008080;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"165\"",
"           x=\"5\"",
"           id=\"tspan2940\"",
"           sodipodi:role=\"line\">key:&lt;Shift&gt;&lt;Alt&gt;C</tspan></text>",
"    </g>",
"    <g",
"       transform=\"translate(770,890)\"",
"       id=\"g2942\">",
"      <rect",
"         style=\"fill:#0000ff;fill-opacity:0.1;stroke:#000000;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1\"",
"         id=\"rect2944\"",
"         width=\"170\"",
"         height=\"170\"",
"         x=\"0\"",
"         y=\"2.9921875\" />",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:48px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#000000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         x=\"40\"",
"         y=\"42.992188\"",
"         id=\"text2946\"",
"         sodipodi:linespacing=\"100%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan2948\"",
"           x=\"40\"",
"           y=\"42.992188\">X</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#000000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"60\"",
"         id=\"text2950\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan2952\"",
"           x=\"5\"",
"           y=\"60\">key:X</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#ff0000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"75\"",
"         id=\"text2954\"><tspan",
"           sodipodi:role=\"line\"",
"           x=\"5\"",
"           y=\"75\"",
"           id=\"tspan2956\">key:&lt;Primary&gt;X</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#008000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"90\"",
"         id=\"text2958\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan2960\"",
"           x=\"5\"",
"           y=\"90\">key:&lt;Shift&gt;X</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#0000ff;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"105\"",
"         id=\"text2962\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan2964\"",
"           x=\"5\"",
"           y=\"105\">key:&lt;Alt&gt;X</tspan></text>",
"      <text",
"         id=\"text6643\"",
"         y=\"120\"",
"         x=\"5\"",
"         style=\"font-style:normal;font-weight:normal;font-size:12.00000048px;font-family:'Bitstream Vera Sans';fill:#464646;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;-inkscape-font-specification:'Bitstream Vera Sans, Normal';font-stretch:normal;font-variant:normal;text-anchor:start;text-align:start;writing-mode:lr;line-height:125%;\"",
"         xml:space=\"preserve\"",
"         sodipodi:linespacing=\"125%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan6647\">key:&lt;Super&gt;X</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#808000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"135\"",
"         id=\"text2966\"",
"         sodipodi:linespacing=\"100%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan2968\"",
"           x=\"5\"",
"           y=\"135\">key:&lt;Primary&gt;&lt;Shift&gt;X</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#ff00ff;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"150\"",
"         id=\"text2970\"",
"         sodipodi:linespacing=\"100%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan2972\"",
"           x=\"5\"",
"           y=\"150\">key:&lt;Primary&gt;&lt;Alt&gt;X</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#008080;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"165\"",
"         id=\"text2974\"",
"         sodipodi:linespacing=\"100%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan2976\"",
"           x=\"5\"",
"           y=\"165\">key:&lt;Shift&gt;&lt;Alt&gt;X</tspan></text>",
"    </g>",
"    <g",
"       id=\"g2978\"",
"       transform=\"translate(590,890)\">",
"      <rect",
"         y=\"2.9921875\"",
"         x=\"0\"",
"         height=\"170\"",
"         width=\"170\"",
"         id=\"rect2980\"",
"         style=\"fill:#0000ff;fill-opacity:0.1;stroke:#000000;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1\" />",
"      <text",
"         sodipodi:linespacing=\"100%\"",
"         id=\"text2982\"",
"         y=\"42.992188\"",
"         x=\"40\"",
"         style=\"font-size:48px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#000000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"42.992188\"",
"           x=\"40\"",
"           id=\"tspan2984\"",
"           sodipodi:role=\"line\">Z</tspan></text>",
"      <text",
"         id=\"text2986\"",
"         y=\"60\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#000000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"60\"",
"           x=\"5\"",
"           id=\"tspan2988\"",
"           sodipodi:role=\"line\">key:Z</tspan></text>",
"      <text",
"         id=\"text2990\"",
"         y=\"75\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#ff0000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           id=\"tspan2992\"",
"           y=\"75\"",
"           x=\"5\"",
"           sodipodi:role=\"line\">key:&lt;Primary&gt;Z</tspan></text>",
"      <text",
"         id=\"text2994\"",
"         y=\"90\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#008000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"90\"",
"           x=\"5\"",
"           id=\"tspan2996\"",
"           sodipodi:role=\"line\">key:&lt;Shift&gt;Z</tspan></text>",
"      <text",
"         id=\"text2998\"",
"         y=\"105\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#0000ff;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"105\"",
"           x=\"5\"",
"           id=\"tspan3000\"",
"           sodipodi:role=\"line\">key:&lt;Alt&gt;Z</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-style:normal;font-weight:normal;font-size:12.00000048px;font-family:'Bitstream Vera Sans';fill:#464646;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;-inkscape-font-specification:'Bitstream Vera Sans, Normal';font-stretch:normal;font-variant:normal;text-anchor:start;text-align:start;writing-mode:lr;line-height:125%;\"",
"         x=\"5\"",
"         y=\"120\"",
"         id=\"text6637\"",
"         sodipodi:linespacing=\"125%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan6641\">key:&lt;Super&gt;Z</tspan></text>",
"      <text",
"         sodipodi:linespacing=\"100%\"",
"         id=\"text3002\"",
"         y=\"135\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#808000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"135\"",
"           x=\"5\"",
"           id=\"tspan3004\"",
"           sodipodi:role=\"line\">key:&lt;Primary&gt;&lt;Shift&gt;Z</tspan></text>",
"      <text",
"         sodipodi:linespacing=\"100%\"",
"         id=\"text3006\"",
"         y=\"150\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#ff00ff;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"150\"",
"           x=\"5\"",
"           id=\"tspan3008\"",
"           sodipodi:role=\"line\">key:&lt;Primary&gt;&lt;Alt&gt;Z</tspan></text>",
"      <text",
"         sodipodi:linespacing=\"100%\"",
"         id=\"text3010\"",
"         y=\"165\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#008080;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"165\"",
"           x=\"5\"",
"           id=\"tspan3012\"",
"           sodipodi:role=\"line\">key:&lt;Shift&gt;&lt;Alt&gt;Z</tspan></text>",
"    </g>",
"    <g",
"       id=\"g3014\"",
"       transform=\"translate(700,350)\">",
"      <rect",
"         y=\"2.9921875\"",
"         x=\"0\"",
"         height=\"170\"",
"         width=\"170\"",
"         id=\"rect3016\"",
"         style=\"fill:#0000ff;fill-opacity:0.1;stroke:#000000;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1\" />",
"      <text",
"         sodipodi:linespacing=\"100%\"",
"         id=\"text3018\"",
"         y=\"42.992188\"",
"         x=\"40\"",
"         style=\"font-size:48px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#000000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"42.992188\"",
"           x=\"40\"",
"           id=\"tspan3020\"",
"           sodipodi:role=\"line\">3</tspan></text>",
"      <text",
"         id=\"text3022\"",
"         y=\"60\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#000000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"60\"",
"           x=\"5\"",
"           id=\"tspan3024\"",
"           sodipodi:role=\"line\">key:3</tspan></text>",
"      <text",
"         id=\"text3026\"",
"         y=\"75\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#ff0000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           id=\"tspan3028\"",
"           y=\"75\"",
"           x=\"5\"",
"           sodipodi:role=\"line\">key:&lt;Primary&gt;3</tspan></text>",
"      <text",
"         id=\"text3030\"",
"         y=\"90\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#008000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"90\"",
"           x=\"5\"",
"           id=\"tspan3032\"",
"           sodipodi:role=\"line\">key:&lt;Shift&gt;sterling</tspan></text>",
"      <text",
"         id=\"text3034\"",
"         y=\"105\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#0000ff;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"105\"",
"           x=\"5\"",
"           id=\"tspan3036\"",
"           sodipodi:role=\"line\">key:&lt;Alt&gt;3</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-style:normal;font-weight:normal;font-size:12.00000048px;font-family:'Bitstream Vera Sans';fill:#464646;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;-inkscape-font-specification:'Bitstream Vera Sans, Normal';font-stretch:normal;font-variant:normal;text-anchor:start;text-align:start;writing-mode:lr;line-height:125%;\"",
"         x=\"5\"",
"         y=\"120\"",
"         id=\"text6339\"",
"         sodipodi:linespacing=\"125%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan6343\">key:&lt;Super&gt;3</tspan></text>",
"      <text",
"         sodipodi:linespacing=\"100%\"",
"         id=\"text3038\"",
"         y=\"135\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#808000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"135\"",
"           x=\"5\"",
"           id=\"tspan3040\"",
"           sodipodi:role=\"line\">key:&lt;Primary&gt;&lt;Shift&gt;sterling</tspan></text>",
"      <text",
"         sodipodi:linespacing=\"100%\"",
"         id=\"text3042\"",
"         y=\"150\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#ff00ff;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"150\"",
"           x=\"5\"",
"           id=\"tspan3044\"",
"           sodipodi:role=\"line\">key:&lt;Primary&gt;&lt;Alt&gt;3</tspan></text>",
"      <text",
"         sodipodi:linespacing=\"100%\"",
"         id=\"text3046\"",
"         y=\"165\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#008080;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"165\"",
"           x=\"5\"",
"           id=\"tspan3048\"",
"           sodipodi:role=\"line\">key:&lt;Shift&gt;&lt;Alt&gt;sterling</tspan></text>",
"    </g>",
"    <g",
"       id=\"g3050\"",
"       transform=\"translate(520,350)\">",
"      <rect",
"         style=\"fill:#0000ff;fill-opacity:0.1;stroke:#000000;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1\"",
"         id=\"rect3052\"",
"         width=\"170\"",
"         height=\"170\"",
"         x=\"0\"",
"         y=\"2.9921875\" />",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:48px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#000000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         x=\"40\"",
"         y=\"42.992188\"",
"         id=\"text3054\"",
"         sodipodi:linespacing=\"100%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan3056\"",
"           x=\"40\"",
"           y=\"42.992188\">2</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#000000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"60\"",
"         id=\"text3058\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan3060\"",
"           x=\"5\"",
"           y=\"60\">key:2</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#ff0000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"75\"",
"         id=\"text3062\"><tspan",
"           sodipodi:role=\"line\"",
"           x=\"5\"",
"           y=\"75\"",
"           id=\"tspan3064\">key:&lt;Primary&gt;2</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#008000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"90\"",
"         id=\"text3066\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan3068\"",
"           x=\"5\"",
"           y=\"90\">key:&lt;Shift&gt;quotedbl</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#0000ff;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"105\"",
"         id=\"text3070\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan3072\"",
"           x=\"5\"",
"           y=\"105\">key:&lt;Alt&gt;2</tspan></text>",
"      <text",
"         id=\"text6333\"",
"         y=\"120\"",
"         x=\"5\"",
"         style=\"font-style:normal;font-weight:normal;font-size:12.00000048px;font-family:'Bitstream Vera Sans';fill:#464646;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;-inkscape-font-specification:'Bitstream Vera Sans, Normal';font-stretch:normal;font-variant:normal;text-anchor:start;text-align:start;writing-mode:lr;line-height:125%;\"",
"         xml:space=\"preserve\"",
"         sodipodi:linespacing=\"125%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan6337\">key:&lt;Super&gt;2</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#808000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"135\"",
"         id=\"text3074\"",
"         sodipodi:linespacing=\"100%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan3076\"",
"           x=\"5\"",
"           y=\"135\">key:&lt;Primary&gt;&lt;Shift&gt;quotedbl</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#ff00ff;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"150\"",
"         id=\"text3078\"",
"         sodipodi:linespacing=\"100%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan3080\"",
"           x=\"5\"",
"           y=\"150\">key:&lt;Primary&gt;&lt;Alt&gt;2</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#008080;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"165\"",
"         id=\"text3082\"",
"         sodipodi:linespacing=\"100%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan3084\"",
"           x=\"5\"",
"           y=\"165\">key:&lt;Shift&gt;&lt;Alt&gt;quotedbl</tspan></text>",
"    </g>",
"    <g",
"       id=\"g3086\"",
"       transform=\"translate(340,350)\">",
"      <rect",
"         y=\"2.9921875\"",
"         x=\"0\"",
"         height=\"170\"",
"         width=\"170\"",
"         id=\"rect3088\"",
"         style=\"fill:#0000ff;fill-opacity:0.1;stroke:#000000;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1\" />",
"      <text",
"         sodipodi:linespacing=\"100%\"",
"         id=\"text3090\"",
"         y=\"42.992188\"",
"         x=\"40\"",
"         style=\"font-size:48px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#000000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"42.992188\"",
"           x=\"40\"",
"           id=\"tspan3092\"",
"           sodipodi:role=\"line\">1</tspan></text>",
"      <text",
"         id=\"text3094\"",
"         y=\"60\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#000000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"60\"",
"           x=\"5\"",
"           id=\"tspan3096\"",
"           sodipodi:role=\"line\">key:1</tspan></text>",
"      <text",
"         id=\"text3098\"",
"         y=\"75\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#ff0000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           id=\"tspan3100\"",
"           y=\"75\"",
"           x=\"5\"",
"           sodipodi:role=\"line\">key:&lt;Primary&gt;1</tspan></text>",
"      <text",
"         id=\"text3102\"",
"         y=\"90\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#008000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"90\"",
"           x=\"5\"",
"           id=\"tspan3104\"",
"           sodipodi:role=\"line\">key:&lt;Shift&gt;exclam</tspan></text>",
"      <text",
"         id=\"text3106\"",
"         y=\"105\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#0000ff;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"105\"",
"           x=\"5\"",
"           id=\"tspan3108\"",
"           sodipodi:role=\"line\">key:&lt;Alt&gt;1</tspan></text>",
"      <text",
"         sodipodi:linespacing=\"100%\"",
"         id=\"text3110\"",
"         y=\"135.35715\"",
"         x=\"5\"",
"         style=\"font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;font-size:12px;line-height:100%;font-family:'Bitstream Vera Sans';-inkscape-font-specification:'Bitstream Vera Sans';text-align:start;writing-mode:lr-tb;text-anchor:start;fill:#808000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"135.35715\"",
"           x=\"5\"",
"           id=\"tspan3112\"",
"           sodipodi:role=\"line\">key:&lt;Primary&gt;&lt;Shift&gt;exclam</tspan></text>",
"      <text",
"         sodipodi:linespacing=\"100%\"",
"         id=\"text3114\"",
"         y=\"150.35715\"",
"         x=\"5\"",
"         style=\"font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;font-size:12px;line-height:100%;font-family:'Bitstream Vera Sans';-inkscape-font-specification:'Bitstream Vera Sans';text-align:start;writing-mode:lr-tb;text-anchor:start;fill:#ff00ff;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"150.35715\"",
"           x=\"5\"",
"           id=\"tspan3116\"",
"           sodipodi:role=\"line\">key:&lt;Primary&gt;&lt;Alt&gt;1</tspan></text>",
"      <text",
"         sodipodi:linespacing=\"100%\"",
"         id=\"text3118\"",
"         y=\"165.35715\"",
"         x=\"5\"",
"         style=\"font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;font-size:12px;line-height:100%;font-family:'Bitstream Vera Sans';-inkscape-font-specification:'Bitstream Vera Sans';text-align:start;writing-mode:lr-tb;text-anchor:start;fill:#008080;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"165.35715\"",
"           x=\"5\"",
"           id=\"tspan3120\"",
"           sodipodi:role=\"line\">key:&lt;Shift&gt;&lt;Alt&gt;exclam</tspan></text>",
"      <text",
"         id=\"text3106-8\"",
"         y=\"120\"",
"         x=\"5\"",
"         style=\"font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;font-size:12.00000095px;line-height:125%;font-family:'Bitstream Vera Sans';-inkscape-font-specification:'Bitstream Vera Sans, Normal';text-align:start;writing-mode:lr-tb;text-anchor:start;fill:#464646;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;\"",
"         xml:space=\"preserve\"",
"         sodipodi:linespacing=\"125%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan6191\"",
"           x=\"5\"",
"           y=\"120\">key:&lt;Super&gt;1</tspan></text>",
"    </g>",
"    <g",
"       transform=\"translate(1780,350)\"",
"       id=\"g3122\">",
"      <rect",
"         style=\"fill:#0000ff;fill-opacity:0.1;stroke:#000000;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1\"",
"         id=\"rect3124\"",
"         width=\"170\"",
"         height=\"170\"",
"         x=\"0\"",
"         y=\"2.9921875\" />",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:48px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#000000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         x=\"40\"",
"         y=\"42.992188\"",
"         id=\"text3126\"",
"         sodipodi:linespacing=\"100%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan3128\"",
"           x=\"40\"",
"           y=\"42.992188\">9</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#000000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"60\"",
"         id=\"text3130\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan3132\"",
"           x=\"5\"",
"           y=\"60\">key:9</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#ff0000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"75\"",
"         id=\"text3134\"><tspan",
"           sodipodi:role=\"line\"",
"           x=\"5\"",
"           y=\"75\"",
"           id=\"tspan3136\">key:&lt;Primary&gt;9</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#008000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"90\"",
"         id=\"text3138\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan3140\"",
"           x=\"5\"",
"           y=\"90\">key:&lt;Shift&gt;parenleft</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#0000ff;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"105\"",
"         id=\"text3142\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan3144\"",
"           x=\"5\"",
"           y=\"105\">key:&lt;Alt&gt;9</tspan></text>",
"      <text",
"         id=\"text6375\"",
"         y=\"120\"",
"         x=\"5\"",
"         style=\"font-style:normal;font-weight:normal;font-size:12.00000048px;font-family:'Bitstream Vera Sans';fill:#464646;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;-inkscape-font-specification:'Bitstream Vera Sans, Normal';font-stretch:normal;font-variant:normal;text-anchor:start;text-align:start;writing-mode:lr;line-height:125%;\"",
"         xml:space=\"preserve\"",
"         sodipodi:linespacing=\"125%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan6477\">key:&lt;Super&gt;9</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#808000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"135\"",
"         id=\"text3146\"",
"         sodipodi:linespacing=\"100%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan3148\"",
"           x=\"5\"",
"           y=\"135\">key:&lt;Primary&gt;&lt;Shift&gt;parenleft</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#ff00ff;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"150\"",
"         id=\"text3150\"",
"         sodipodi:linespacing=\"100%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan3152\"",
"           x=\"5\"",
"           y=\"150\">key:&lt;Primary&gt;&lt;Alt&gt;9</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#008080;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"165\"",
"         id=\"text3154\"",
"         sodipodi:linespacing=\"100%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan3156\"",
"           x=\"5\"",
"           y=\"165\">key:&lt;Shift&gt;&lt;Alt&gt;parenleft</tspan></text>",
"    </g>",
"    <g",
"       id=\"g3158\"",
"       transform=\"translate(1600,350)\">",
"      <rect",
"         y=\"2.9921875\"",
"         x=\"0\"",
"         height=\"170\"",
"         width=\"170\"",
"         id=\"rect3160\"",
"         style=\"fill:#0000ff;fill-opacity:0.1;stroke:#000000;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1\" />",
"      <text",
"         sodipodi:linespacing=\"100%\"",
"         id=\"text3162\"",
"         y=\"42.992188\"",
"         x=\"40\"",
"         style=\"font-size:48px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#000000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"42.992188\"",
"           x=\"40\"",
"           id=\"tspan3164\"",
"           sodipodi:role=\"line\">8</tspan></text>",
"      <text",
"         id=\"text3166\"",
"         y=\"60\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#000000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"60\"",
"           x=\"5\"",
"           id=\"tspan3168\"",
"           sodipodi:role=\"line\">key:8</tspan></text>",
"      <text",
"         id=\"text3170\"",
"         y=\"75\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#ff0000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           id=\"tspan3172\"",
"           y=\"75\"",
"           x=\"5\"",
"           sodipodi:role=\"line\">key:&lt;Primary&gt;8</tspan></text>",
"      <text",
"         id=\"text3174\"",
"         y=\"90\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#008000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"90\"",
"           x=\"5\"",
"           id=\"tspan3176\"",
"           sodipodi:role=\"line\">key:&lt;Shift&gt;asterisk</tspan></text>",
"      <text",
"         id=\"text3178\"",
"         y=\"105\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#0000ff;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"105\"",
"           x=\"5\"",
"           id=\"tspan3180\"",
"           sodipodi:role=\"line\">key:&lt;Alt&gt;8</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-style:normal;font-weight:normal;font-size:12.00000048px;font-family:'Bitstream Vera Sans';fill:#464646;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;-inkscape-font-specification:'Bitstream Vera Sans, Normal';font-stretch:normal;font-variant:normal;text-anchor:start;text-align:start;writing-mode:lr;line-height:125%;\"",
"         x=\"5\"",
"         y=\"120\"",
"         id=\"text6371\"",
"         sodipodi:linespacing=\"125%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan6475\">key:&lt;Super&gt;8</tspan></text>",
"      <text",
"         sodipodi:linespacing=\"100%\"",
"         id=\"text3182\"",
"         y=\"135\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#808000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"135\"",
"           x=\"5\"",
"           id=\"tspan3184\"",
"           sodipodi:role=\"line\">key:&lt;Primary&gt;&lt;Shift&gt;asterisk</tspan></text>",
"      <text",
"         sodipodi:linespacing=\"100%\"",
"         id=\"text3186\"",
"         y=\"150\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#ff00ff;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"150\"",
"           x=\"5\"",
"           id=\"tspan3188\"",
"           sodipodi:role=\"line\">key:&lt;Primary&gt;&lt;Alt&gt;8</tspan></text>",
"      <text",
"         sodipodi:linespacing=\"100%\"",
"         id=\"text3190\"",
"         y=\"165\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#008080;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"165\"",
"           x=\"5\"",
"           id=\"tspan3192\"",
"           sodipodi:role=\"line\">key:&lt;Shift&gt;&lt;Alt&gt;asterisk</tspan></text>",
"    </g>",
"    <g",
"       transform=\"translate(1420,350)\"",
"       id=\"g3194\">",
"      <rect",
"         style=\"fill:#0000ff;fill-opacity:0.1;stroke:#000000;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1\"",
"         id=\"rect3196\"",
"         width=\"170\"",
"         height=\"170\"",
"         x=\"0\"",
"         y=\"2.9921875\" />",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:48px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#000000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         x=\"40\"",
"         y=\"42.992188\"",
"         id=\"text3198\"",
"         sodipodi:linespacing=\"100%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan3200\"",
"           x=\"40\"",
"           y=\"42.992188\">7</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#000000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"60\"",
"         id=\"text3202\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan3204\"",
"           x=\"5\"",
"           y=\"60\">key:7</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#ff0000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"75\"",
"         id=\"text3206\"><tspan",
"           sodipodi:role=\"line\"",
"           x=\"5\"",
"           y=\"75\"",
"           id=\"tspan3208\">key:&lt;Primary&gt;7</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#008000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"90\"",
"         id=\"text3210\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan3212\"",
"           x=\"5\"",
"           y=\"90\">key:&lt;Shift&gt;ampersand</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#0000ff;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"105\"",
"         id=\"text3214\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan3216\"",
"           x=\"5\"",
"           y=\"105\">key:&lt;Alt&gt;7</tspan></text>",
"      <text",
"         id=\"text6367\"",
"         y=\"120\"",
"         x=\"5\"",
"         style=\"font-style:normal;font-weight:normal;font-size:12.00000048px;font-family:'Bitstream Vera Sans';fill:#464646;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;-inkscape-font-specification:'Bitstream Vera Sans, Normal';font-stretch:normal;font-variant:normal;text-anchor:start;text-align:start;writing-mode:lr;line-height:125%;\"",
"         xml:space=\"preserve\"",
"         sodipodi:linespacing=\"125%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan6473\">key:&lt;Super&gt;7</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#808000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"135\"",
"         id=\"text3218\"",
"         sodipodi:linespacing=\"100%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan3220\"",
"           x=\"5\"",
"           y=\"135\">key:&lt;Primary&gt;&lt;Shift&gt;ampersand</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#ff00ff;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"150\"",
"         id=\"text3222\"",
"         sodipodi:linespacing=\"100%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan3224\"",
"           x=\"5\"",
"           y=\"150\">key:&lt;Primary&gt;&lt;Alt&gt;7</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#008080;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"165\"",
"         id=\"text3226\"",
"         sodipodi:linespacing=\"100%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan3228\"",
"           x=\"5\"",
"           y=\"165\">key:&lt;Shift&gt;&lt;Alt&gt;ampersand</tspan></text>",
"    </g>",
"    <g",
"       id=\"g3230\"",
"       transform=\"translate(1240,350)\">",
"      <rect",
"         y=\"2.9921875\"",
"         x=\"0\"",
"         height=\"170\"",
"         width=\"170\"",
"         id=\"rect3232\"",
"         style=\"fill:#0000ff;fill-opacity:0.1;stroke:#000000;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1\" />",
"      <text",
"         sodipodi:linespacing=\"100%\"",
"         id=\"text3234\"",
"         y=\"42.992188\"",
"         x=\"40\"",
"         style=\"font-size:48px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#000000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"42.992188\"",
"           x=\"40\"",
"           id=\"tspan3236\"",
"           sodipodi:role=\"line\">6</tspan></text>",
"      <text",
"         id=\"text3238\"",
"         y=\"60\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#000000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"60\"",
"           x=\"5\"",
"           id=\"tspan3240\"",
"           sodipodi:role=\"line\">key:6</tspan></text>",
"      <text",
"         id=\"text3242\"",
"         y=\"75\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#ff0000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           id=\"tspan3244\"",
"           y=\"75\"",
"           x=\"5\"",
"           sodipodi:role=\"line\">key:&lt;Primary&gt;6</tspan></text>",
"      <text",
"         id=\"text3246\"",
"         y=\"90\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#008000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"90\"",
"           x=\"5\"",
"           id=\"tspan3248\"",
"           sodipodi:role=\"line\">key:&lt;Shift&gt;asciicircum</tspan></text>",
"      <text",
"         id=\"text3250\"",
"         y=\"105\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#0000ff;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"105\"",
"           x=\"5\"",
"           id=\"tspan3252\"",
"           sodipodi:role=\"line\">key:&lt;Alt&gt;6</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-style:normal;font-weight:normal;font-size:12.00000048px;font-family:'Bitstream Vera Sans';fill:#464646;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;-inkscape-font-specification:'Bitstream Vera Sans, Normal';font-stretch:normal;font-variant:normal;text-anchor:start;text-align:start;writing-mode:lr;line-height:125%;\"",
"         x=\"5\"",
"         y=\"120\"",
"         id=\"text6355\"",
"         sodipodi:linespacing=\"125%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan6471\">key:&lt;Super&gt;6</tspan></text>",
"      <text",
"         sodipodi:linespacing=\"100%\"",
"         id=\"text3254\"",
"         y=\"135\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#808000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"135\"",
"           x=\"5\"",
"           id=\"tspan3256\"",
"           sodipodi:role=\"line\">key:&lt;Primary&gt;&lt;Shift&gt;asciicircum</tspan></text>",
"      <text",
"         sodipodi:linespacing=\"100%\"",
"         id=\"text3258\"",
"         y=\"150\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#ff00ff;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"150\"",
"           x=\"5\"",
"           id=\"tspan3260\"",
"           sodipodi:role=\"line\">key:&lt;Primary&gt;&lt;Alt&gt;6</tspan></text>",
"      <text",
"         sodipodi:linespacing=\"100%\"",
"         id=\"text3262\"",
"         y=\"165\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#008080;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"165\"",
"           x=\"5\"",
"           id=\"tspan3264\"",
"           sodipodi:role=\"line\">key:&lt;Shift&gt;&lt;Alt&gt;asciicircum</tspan></text>",
"    </g>",
"    <g",
"       transform=\"translate(1060,350)\"",
"       id=\"g3266\">",
"      <rect",
"         style=\"fill:#0000ff;fill-opacity:0.1;stroke:#000000;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1\"",
"         id=\"rect3268\"",
"         width=\"170\"",
"         height=\"170\"",
"         x=\"0\"",
"         y=\"2.9921875\" />",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:48px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#000000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         x=\"40\"",
"         y=\"42.992188\"",
"         id=\"text3270\"",
"         sodipodi:linespacing=\"100%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan3272\"",
"           x=\"40\"",
"           y=\"42.992188\">5</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#000000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"60\"",
"         id=\"text3274\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan3276\"",
"           x=\"5\"",
"           y=\"60\">key:5</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#ff0000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"75\"",
"         id=\"text3278\"><tspan",
"           sodipodi:role=\"line\"",
"           x=\"5\"",
"           y=\"75\"",
"           id=\"tspan3280\">key:&lt;Primary&gt;5</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#008000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"90\"",
"         id=\"text3282\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan3284\"",
"           x=\"5\"",
"           y=\"90\">key:&lt;Shift&gt;percent</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#0000ff;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"105\"",
"         id=\"text3286\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan3288\"",
"           x=\"5\"",
"           y=\"105\">key:&lt;Alt&gt;5</tspan></text>",
"      <text",
"         id=\"text6351\"",
"         y=\"120\"",
"         x=\"5\"",
"         style=\"font-style:normal;font-weight:normal;font-size:12.00000048px;font-family:'Bitstream Vera Sans';fill:#464646;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;-inkscape-font-specification:'Bitstream Vera Sans, Normal';font-stretch:normal;font-variant:normal;text-anchor:start;text-align:start;writing-mode:lr;line-height:125%;\"",
"         xml:space=\"preserve\"",
"         sodipodi:linespacing=\"125%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan6469\">key:&lt;Super&gt;5</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#808000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"135\"",
"         id=\"text3290\"",
"         sodipodi:linespacing=\"100%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan3292\"",
"           x=\"5\"",
"           y=\"135\">key:&lt;Primary&gt;&lt;Shift&gt;percent</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#ff00ff;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"150\"",
"         id=\"text3294\"",
"         sodipodi:linespacing=\"100%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan3296\"",
"           x=\"5\"",
"           y=\"150\">key:&lt;Primary&gt;&lt;Alt&gt;5</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#008080;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"165\"",
"         id=\"text3298\"",
"         sodipodi:linespacing=\"100%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan3300\"",
"           x=\"5\"",
"           y=\"165\">key:&lt;Shift&gt;&lt;Alt&gt;percent</tspan></text>",
"    </g>",
"    <g",
"       id=\"g3302\"",
"       transform=\"translate(880,350)\">",
"      <rect",
"         y=\"2.9921875\"",
"         x=\"0\"",
"         height=\"170\"",
"         width=\"170\"",
"         id=\"rect3304\"",
"         style=\"fill:#0000ff;fill-opacity:0.1;stroke:#000000;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1\" />",
"      <text",
"         sodipodi:linespacing=\"100%\"",
"         id=\"text3306\"",
"         y=\"42.992188\"",
"         x=\"40\"",
"         style=\"font-size:48px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#000000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"42.992188\"",
"           x=\"40\"",
"           id=\"tspan3308\"",
"           sodipodi:role=\"line\">4</tspan></text>",
"      <text",
"         id=\"text3310\"",
"         y=\"60\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#000000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"60\"",
"           x=\"5\"",
"           id=\"tspan3312\"",
"           sodipodi:role=\"line\">key:4</tspan></text>",
"      <text",
"         id=\"text3314\"",
"         y=\"75\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#ff0000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           id=\"tspan3316\"",
"           y=\"75\"",
"           x=\"5\"",
"           sodipodi:role=\"line\">key:&lt;Primary&gt;4</tspan></text>",
"      <text",
"         id=\"text3318\"",
"         y=\"90\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#008000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"90\"",
"           x=\"5\"",
"           id=\"tspan3320\"",
"           sodipodi:role=\"line\">key:&lt;Shift&gt;dollar</tspan></text>",
"      <text",
"         id=\"text3322\"",
"         y=\"105\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#0000ff;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"105\"",
"           x=\"5\"",
"           id=\"tspan3324\"",
"           sodipodi:role=\"line\">key:&lt;Alt&gt;4</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-style:normal;font-weight:normal;font-size:12.00000048px;font-family:'Bitstream Vera Sans';fill:#464646;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;-inkscape-font-specification:'Bitstream Vera Sans, Normal';font-stretch:normal;font-variant:normal;text-anchor:start;text-align:start;writing-mode:lr;line-height:125%;\"",
"         x=\"5\"",
"         y=\"120\"",
"         id=\"text6345\"",
"         sodipodi:linespacing=\"125%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan6349\">key:&lt;Super&gt;4</tspan></text>",
"      <text",
"         sodipodi:linespacing=\"100%\"",
"         id=\"text3326\"",
"         y=\"135\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#808000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"135\"",
"           x=\"5\"",
"           id=\"tspan3328\"",
"           sodipodi:role=\"line\">key:&lt;Primary&gt;&lt;Shift&gt;dollar</tspan></text>",
"      <text",
"         sodipodi:linespacing=\"100%\"",
"         id=\"text3330\"",
"         y=\"150\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#ff00ff;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"150\"",
"           x=\"5\"",
"           id=\"tspan3332\"",
"           sodipodi:role=\"line\">key:&lt;Primary&gt;&lt;Alt&gt;4</tspan></text>",
"      <text",
"         sodipodi:linespacing=\"100%\"",
"         id=\"text3334\"",
"         y=\"165\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#008080;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"165\"",
"           x=\"5\"",
"           id=\"tspan3336\"",
"           sodipodi:role=\"line\">key:&lt;Shift&gt;&lt;Alt&gt;dollar</tspan></text>",
"    </g>",
"    <g",
"       id=\"g3338\"",
"       transform=\"translate(1940,710)\">",
"      <rect",
"         y=\"2.9921875\"",
"         x=\"0\"",
"         height=\"170\"",
"         width=\"170\"",
"         id=\"rect3340\"",
"         style=\"fill:#0000ff;fill-opacity:0.1;stroke:#000000;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1\" />",
"      <text",
"         sodipodi:linespacing=\"100%\"",
"         id=\"text3342\"",
"         y=\"42.992188\"",
"         x=\"40\"",
"         style=\"font-size:48px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#000000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"42.992188\"",
"           x=\"40\"",
"           id=\"tspan3344\"",
"           sodipodi:role=\"line\">L</tspan></text>",
"      <text",
"         id=\"text3346\"",
"         y=\"60\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#000000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"60\"",
"           x=\"5\"",
"           id=\"tspan3348\"",
"           sodipodi:role=\"line\">key:L</tspan></text>",
"      <text",
"         id=\"text3350\"",
"         y=\"75\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#ff0000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           id=\"tspan3352\"",
"           y=\"75\"",
"           x=\"5\"",
"           sodipodi:role=\"line\">key:&lt;Primary&gt;L</tspan></text>",
"      <text",
"         id=\"text3354\"",
"         y=\"90\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#008000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"90\"",
"           x=\"5\"",
"           id=\"tspan3356\"",
"           sodipodi:role=\"line\">key:&lt;Shift&gt;L</tspan></text>",
"      <text",
"         id=\"text3358\"",
"         y=\"105\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#0000ff;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"105\"",
"           x=\"5\"",
"           id=\"tspan3360\"",
"           sodipodi:role=\"line\">key:&lt;Alt&gt;L</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-style:normal;font-weight:normal;font-size:12.00000048px;font-family:'Bitstream Vera Sans';fill:#464646;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;-inkscape-font-specification:'Bitstream Vera Sans, Normal';font-stretch:normal;font-variant:normal;text-anchor:start;text-align:start;writing-mode:lr;line-height:125%;\"",
"         x=\"5\"",
"         y=\"120\"",
"         id=\"text6607\"",
"         sodipodi:linespacing=\"125%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan6611\">key:&lt;Super&gt;L</tspan></text>",
"      <text",
"         sodipodi:linespacing=\"100%\"",
"         id=\"text3362\"",
"         y=\"135\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#808000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"135\"",
"           x=\"5\"",
"           id=\"tspan3364\"",
"           sodipodi:role=\"line\">key:&lt;Primary&gt;&lt;Shift&gt;L</tspan></text>",
"      <text",
"         sodipodi:linespacing=\"100%\"",
"         id=\"text3366\"",
"         y=\"150\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#ff00ff;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"150\"",
"           x=\"5\"",
"           id=\"tspan3368\"",
"           sodipodi:role=\"line\">key:&lt;Primary&gt;&lt;Alt&gt;L</tspan></text>",
"      <text",
"         sodipodi:linespacing=\"100%\"",
"         id=\"text3370\"",
"         y=\"165\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#008080;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"165\"",
"           x=\"5\"",
"           id=\"tspan3372\"",
"           sodipodi:role=\"line\">key:&lt;Shift&gt;&lt;Alt&gt;L</tspan></text>",
"    </g>",
"    <g",
"       transform=\"translate(1760,710)\"",
"       id=\"g3374\">",
"      <rect",
"         style=\"fill:#0000ff;fill-opacity:0.1;stroke:#000000;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1\"",
"         id=\"rect3376\"",
"         width=\"170\"",
"         height=\"170\"",
"         x=\"0\"",
"         y=\"2.9921875\" />",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:48px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#000000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         x=\"40\"",
"         y=\"42.992188\"",
"         id=\"text3378\"",
"         sodipodi:linespacing=\"100%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan3380\"",
"           x=\"40\"",
"           y=\"42.992188\">K</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#000000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"60\"",
"         id=\"text3382\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan3384\"",
"           x=\"5\"",
"           y=\"60\">key:K</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#ff0000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"75\"",
"         id=\"text3386\"><tspan",
"           sodipodi:role=\"line\"",
"           x=\"5\"",
"           y=\"75\"",
"           id=\"tspan3388\">key:&lt;Primary&gt;K</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#008000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"90\"",
"         id=\"text3390\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan3392\"",
"           x=\"5\"",
"           y=\"90\">key:&lt;Shift&gt;K</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#0000ff;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"105\"",
"         id=\"text3394\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan3396\"",
"           x=\"5\"",
"           y=\"105\">key:&lt;Alt&gt;K</tspan></text>",
"      <text",
"         id=\"text6601\"",
"         y=\"120\"",
"         x=\"5\"",
"         style=\"font-style:normal;font-weight:normal;font-size:12.00000048px;font-family:'Bitstream Vera Sans';fill:#464646;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;-inkscape-font-specification:'Bitstream Vera Sans, Normal';font-stretch:normal;font-variant:normal;text-anchor:start;text-align:start;writing-mode:lr;line-height:125%;\"",
"         xml:space=\"preserve\"",
"         sodipodi:linespacing=\"125%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan6605\">key:&lt;Super&gt;K</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#808000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"135\"",
"         id=\"text3398\"",
"         sodipodi:linespacing=\"100%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan3400\"",
"           x=\"5\"",
"           y=\"135\">key:&lt;Primary&gt;&lt;Shift&gt;K</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#ff00ff;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"150\"",
"         id=\"text3402\"",
"         sodipodi:linespacing=\"100%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan3404\"",
"           x=\"5\"",
"           y=\"150\">key:&lt;Primary&gt;&lt;Alt&gt;K</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#008080;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"165\"",
"         id=\"text3406\"",
"         sodipodi:linespacing=\"100%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan3408\"",
"           x=\"5\"",
"           y=\"165\">key:&lt;Shift&gt;&lt;Alt&gt;K</tspan></text>",
"    </g>",
"    <g",
"       id=\"g3410\"",
"       transform=\"translate(1890,530)\">",
"      <rect",
"         y=\"2.9921875\"",
"         x=\"0\"",
"         height=\"170\"",
"         width=\"170\"",
"         id=\"rect3412\"",
"         style=\"fill:#0000ff;fill-opacity:0.1;stroke:#000000;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1\" />",
"      <text",
"         sodipodi:linespacing=\"100%\"",
"         id=\"text3414\"",
"         y=\"42.992188\"",
"         x=\"40\"",
"         style=\"font-size:48px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#000000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"42.992188\"",
"           x=\"40\"",
"           id=\"tspan3416\"",
"           sodipodi:role=\"line\">O</tspan></text>",
"      <text",
"         id=\"text3418\"",
"         y=\"60\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#000000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"60\"",
"           x=\"5\"",
"           id=\"tspan3420\"",
"           sodipodi:role=\"line\">key:O</tspan></text>",
"      <text",
"         id=\"text3422\"",
"         y=\"75\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#ff0000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           id=\"tspan3424\"",
"           y=\"75\"",
"           x=\"5\"",
"           sodipodi:role=\"line\">key:&lt;Primary&gt;O</tspan></text>",
"      <text",
"         id=\"text3426\"",
"         y=\"90\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#008000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"90\"",
"           x=\"5\"",
"           id=\"tspan3428\"",
"           sodipodi:role=\"line\">key:&lt;Shift&gt;O</tspan></text>",
"      <text",
"         id=\"text3430\"",
"         y=\"105\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#0000ff;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"105\"",
"           x=\"5\"",
"           id=\"tspan3432\"",
"           sodipodi:role=\"line\">key:&lt;Alt&gt;O</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-style:normal;font-weight:normal;font-size:12.00000048px;font-family:'Bitstream Vera Sans';fill:#464646;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;-inkscape-font-specification:'Bitstream Vera Sans, Normal';font-stretch:normal;font-variant:normal;text-anchor:start;text-align:start;writing-mode:lr;line-height:125%;\"",
"         x=\"5\"",
"         y=\"120\"",
"         id=\"text6537\"",
"         sodipodi:linespacing=\"125%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan6541\">key:&lt;Super&gt;O</tspan></text>",
"      <text",
"         sodipodi:linespacing=\"100%\"",
"         id=\"text3434\"",
"         y=\"135\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#808000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"135\"",
"           x=\"5\"",
"           id=\"tspan3436\"",
"           sodipodi:role=\"line\">key:&lt;Primary&gt;&lt;Shift&gt;O</tspan></text>",
"      <text",
"         sodipodi:linespacing=\"100%\"",
"         id=\"text3438\"",
"         y=\"150\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#ff00ff;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"150\"",
"           x=\"5\"",
"           id=\"tspan3440\"",
"           sodipodi:role=\"line\">key:&lt;Primary&gt;&lt;Alt&gt;O</tspan></text>",
"      <text",
"         sodipodi:linespacing=\"100%\"",
"         id=\"text3442\"",
"         y=\"165\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#008080;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"165\"",
"           x=\"5\"",
"           id=\"tspan3444\"",
"           sodipodi:role=\"line\">key:&lt;Shift&gt;&lt;Alt&gt;O</tspan></text>",
"    </g>",
"    <g",
"       transform=\"translate(1710,530)\"",
"       id=\"g3446\">",
"      <rect",
"         style=\"fill:#0000ff;fill-opacity:0.1;stroke:#000000;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1\"",
"         id=\"rect3448\"",
"         width=\"170\"",
"         height=\"170\"",
"         x=\"0\"",
"         y=\"2.9921875\" />",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:48px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#000000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         x=\"40\"",
"         y=\"42.992188\"",
"         id=\"text3450\"",
"         sodipodi:linespacing=\"100%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan3452\"",
"           x=\"40\"",
"           y=\"42.992188\">I</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#000000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"60\"",
"         id=\"text3454\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan3456\"",
"           x=\"5\"",
"           y=\"60\">key:I</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#ff0000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"75\"",
"         id=\"text3458\"><tspan",
"           sodipodi:role=\"line\"",
"           x=\"5\"",
"           y=\"75\"",
"           id=\"tspan3460\">key:&lt;Primary&gt;I</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#008000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"90\"",
"         id=\"text3462\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan3464\"",
"           x=\"5\"",
"           y=\"90\">key:&lt;Shift&gt;I</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#0000ff;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"105\"",
"         id=\"text3466\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan3468\"",
"           x=\"5\"",
"           y=\"105\">key:&lt;Alt&gt;I</tspan></text>",
"      <text",
"         id=\"text6531\"",
"         y=\"120\"",
"         x=\"5\"",
"         style=\"font-style:normal;font-weight:normal;font-size:12.00000048px;font-family:'Bitstream Vera Sans';fill:#464646;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;-inkscape-font-specification:'Bitstream Vera Sans, Normal';font-stretch:normal;font-variant:normal;text-anchor:start;text-align:start;writing-mode:lr;line-height:125%;\"",
"         xml:space=\"preserve\"",
"         sodipodi:linespacing=\"125%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan6535\">key:&lt;Super&gt;I</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#808000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"135\"",
"         id=\"text3470\"",
"         sodipodi:linespacing=\"100%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan3472\"",
"           x=\"5\"",
"           y=\"135\">key:&lt;Primary&gt;&lt;Shift&gt;I</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#ff00ff;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"150\"",
"         id=\"text3474\"",
"         sodipodi:linespacing=\"100%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan3476\"",
"           x=\"5\"",
"           y=\"150\">key:&lt;Primary&gt;&lt;Alt&gt;I</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#008080;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"165\"",
"         id=\"text3478\"",
"         sodipodi:linespacing=\"100%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan3480\"",
"           x=\"5\"",
"           y=\"165\">key:&lt;Shift&gt;&lt;Alt&gt;I</tspan></text>",
"    </g>",
"    <g",
"       id=\"g3482\"",
"       transform=\"translate(2030,890)\">",
"      <rect",
"         y=\"2.9921875\"",
"         x=\"0\"",
"         height=\"170\"",
"         width=\"170\"",
"         id=\"rect3484\"",
"         style=\"fill:#0000ff;fill-opacity:0.1;stroke:#000000;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1\" />",
"      <text",
"         sodipodi:linespacing=\"100%\"",
"         id=\"text3486\"",
"         y=\"42.992188\"",
"         x=\"40\"",
"         style=\"font-size:48px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#000000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"42.992188\"",
"           x=\"40\"",
"           id=\"tspan3488\"",
"           sodipodi:role=\"line\">.</tspan></text>",
"      <text",
"         id=\"text3490\"",
"         y=\"60\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#000000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"60\"",
"           x=\"5\"",
"           id=\"tspan3492\"",
"           sodipodi:role=\"line\">key:period</tspan></text>",
"      <text",
"         id=\"text3494\"",
"         y=\"75\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#ff0000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           id=\"tspan3496\"",
"           y=\"75\"",
"           x=\"5\"",
"           sodipodi:role=\"line\">key:&lt;Primary&gt;period</tspan></text>",
"      <text",
"         id=\"text3498\"",
"         y=\"90\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#008000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"90\"",
"           x=\"5\"",
"           id=\"tspan3500\"",
"           sodipodi:role=\"line\">key:&lt;Shift&gt;greater</tspan></text>",
"      <text",
"         id=\"text3502\"",
"         y=\"105\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#0000ff;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"105\"",
"           x=\"5\"",
"           id=\"tspan3504\"",
"           sodipodi:role=\"line\">key:&lt;Alt&gt;period</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-style:normal;font-weight:normal;font-size:12.00000048px;font-family:'Bitstream Vera Sans';fill:#464646;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;-inkscape-font-specification:'Bitstream Vera Sans, Normal';font-stretch:normal;font-variant:normal;text-anchor:start;text-align:start;writing-mode:lr;line-height:125%;\"",
"         x=\"5\"",
"         y=\"120\"",
"         id=\"text6685\"",
"         sodipodi:linespacing=\"125%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan6689\">key:&lt;Super&gt;period</tspan></text>",
"      <text",
"         sodipodi:linespacing=\"100%\"",
"         id=\"text3506\"",
"         y=\"135\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#808000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"135\"",
"           x=\"5\"",
"           id=\"tspan3508\"",
"           sodipodi:role=\"line\">key:&lt;Primary&gt;&lt;Shift&gt;greater</tspan></text>",
"      <text",
"         sodipodi:linespacing=\"100%\"",
"         id=\"text3510\"",
"         y=\"150\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#ff00ff;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"150\"",
"           x=\"5\"",
"           id=\"tspan3512\"",
"           sodipodi:role=\"line\">key:&lt;Primary&gt;&lt;Alt&gt;period</tspan></text>",
"      <text",
"         sodipodi:linespacing=\"100%\"",
"         id=\"text3514\"",
"         y=\"165\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#008080;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"165\"",
"           x=\"5\"",
"           id=\"tspan3516\"",
"           sodipodi:role=\"line\">key:&lt;Shift&gt;&lt;Alt&gt;greater</tspan></text>",
"    </g>",
"    <g",
"       transform=\"translate(1850,890)\"",
"       id=\"g3518\">",
"      <rect",
"         style=\"fill:#0000ff;fill-opacity:0.1;stroke:#000000;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1\"",
"         id=\"rect3520\"",
"         width=\"170\"",
"         height=\"170\"",
"         x=\"0\"",
"         y=\"2.9921875\" />",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:48px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#000000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         x=\"40\"",
"         y=\"42.992188\"",
"         id=\"text3522\"",
"         sodipodi:linespacing=\"100%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan3524\"",
"           x=\"40\"",
"           y=\"42.992188\">,</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#000000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"60\"",
"         id=\"text3526\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan3528\"",
"           x=\"5\"",
"           y=\"60\">key:comma</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#ff0000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"75\"",
"         id=\"text3530\"><tspan",
"           sodipodi:role=\"line\"",
"           x=\"5\"",
"           y=\"75\"",
"           id=\"tspan3532\">key:&lt;Primary&gt;comma</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#008000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"90\"",
"         id=\"text3534\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan3536\"",
"           x=\"5\"",
"           y=\"90\">key:&lt;Shift&gt;less</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#0000ff;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"105\"",
"         id=\"text3538\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan3540\"",
"           x=\"5\"",
"           y=\"105\">key:&lt;Alt&gt;comma</tspan></text>",
"      <text",
"         id=\"text6679\"",
"         y=\"120\"",
"         x=\"5\"",
"         style=\"font-style:normal;font-weight:normal;font-size:12.00000048px;font-family:'Bitstream Vera Sans';fill:#464646;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;-inkscape-font-specification:'Bitstream Vera Sans, Normal';font-stretch:normal;font-variant:normal;text-anchor:start;text-align:start;writing-mode:lr;line-height:125%;\"",
"         xml:space=\"preserve\"",
"         sodipodi:linespacing=\"125%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan6683\">key:&lt;Super&gt;comma</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#808000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"135\"",
"         id=\"text3542\"",
"         sodipodi:linespacing=\"100%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan3544\"",
"           x=\"5\"",
"           y=\"135\">key:&lt;Primary&gt;&lt;Shift&gt;less</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#ff00ff;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"150\"",
"         id=\"text3546\"",
"         sodipodi:linespacing=\"100%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan3548\"",
"           x=\"5\"",
"           y=\"150\">key:&lt;Primary&gt;&lt;Alt&gt;comma</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#008080;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"165\"",
"         id=\"text3550\"",
"         sodipodi:linespacing=\"100%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan3552\"",
"           x=\"5\"",
"           y=\"165\">key:&lt;Shift&gt;&lt;Alt&gt;less</tspan></text>",
"    </g>",
"    <g",
"       id=\"g3554\"",
"       transform=\"translate(1670,890)\">",
"      <rect",
"         y=\"2.9921875\"",
"         x=\"0\"",
"         height=\"170\"",
"         width=\"170\"",
"         id=\"rect3556\"",
"         style=\"fill:#0000ff;fill-opacity:0.1;stroke:#000000;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1\" />",
"      <text",
"         sodipodi:linespacing=\"100%\"",
"         id=\"text3558\"",
"         y=\"42.992188\"",
"         x=\"40\"",
"         style=\"font-size:48px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#000000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"42.992188\"",
"           x=\"40\"",
"           id=\"tspan3560\"",
"           sodipodi:role=\"line\">M</tspan></text>",
"      <text",
"         id=\"text3562\"",
"         y=\"60\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#000000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"60\"",
"           x=\"5\"",
"           id=\"tspan3564\"",
"           sodipodi:role=\"line\">key:M</tspan></text>",
"      <text",
"         id=\"text3566\"",
"         y=\"75\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#ff0000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           id=\"tspan3568\"",
"           y=\"75\"",
"           x=\"5\"",
"           sodipodi:role=\"line\">key:&lt;Primary&gt;M</tspan></text>",
"      <text",
"         id=\"text3570\"",
"         y=\"90\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#008000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"90\"",
"           x=\"5\"",
"           id=\"tspan3572\"",
"           sodipodi:role=\"line\">key:&lt;Shift&gt;M</tspan></text>",
"      <text",
"         id=\"text3574\"",
"         y=\"105\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#0000ff;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"105\"",
"           x=\"5\"",
"           id=\"tspan3576\"",
"           sodipodi:role=\"line\">key:&lt;Alt&gt;M</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-style:normal;font-weight:normal;font-size:12.00000048px;font-family:'Bitstream Vera Sans';fill:#464646;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;-inkscape-font-specification:'Bitstream Vera Sans, Normal';font-stretch:normal;font-variant:normal;text-anchor:start;text-align:start;writing-mode:lr;line-height:125%;\"",
"         x=\"5\"",
"         y=\"120\"",
"         id=\"text6673\"",
"         sodipodi:linespacing=\"125%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan6677\">key:&lt;Super&gt;M</tspan></text>",
"      <text",
"         sodipodi:linespacing=\"100%\"",
"         id=\"text3578\"",
"         y=\"135\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#808000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"135\"",
"           x=\"5\"",
"           id=\"tspan3580\"",
"           sodipodi:role=\"line\">key:&lt;Primary&gt;&lt;Shift&gt;M</tspan></text>",
"      <text",
"         sodipodi:linespacing=\"100%\"",
"         id=\"text3582\"",
"         y=\"150\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#ff00ff;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"150\"",
"           x=\"5\"",
"           id=\"tspan3584\"",
"           sodipodi:role=\"line\">key:&lt;Primary&gt;&lt;Alt&gt;M</tspan></text>",
"      <text",
"         sodipodi:linespacing=\"100%\"",
"         id=\"text3586\"",
"         y=\"165\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#008080;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"165\"",
"           x=\"5\"",
"           id=\"tspan3588\"",
"           sodipodi:role=\"line\">key:&lt;Shift&gt;&lt;Alt&gt;M</tspan></text>",
"    </g>",
"    <g",
"       transform=\"translate(1490,890)\"",
"       id=\"g3590\">",
"      <rect",
"         style=\"fill:#0000ff;fill-opacity:0.1;stroke:#000000;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1\"",
"         id=\"rect3592\"",
"         width=\"170\"",
"         height=\"170\"",
"         x=\"0\"",
"         y=\"2.9921875\" />",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:48px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#000000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         x=\"40\"",
"         y=\"42.992188\"",
"         id=\"text3594\"",
"         sodipodi:linespacing=\"100%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan3596\"",
"           x=\"40\"",
"           y=\"42.992188\">N</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#000000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"60\"",
"         id=\"text3598\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan3600\"",
"           x=\"5\"",
"           y=\"60\">key:N</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#ff0000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"75\"",
"         id=\"text3602\"><tspan",
"           sodipodi:role=\"line\"",
"           x=\"5\"",
"           y=\"75\"",
"           id=\"tspan3604\">key:&lt;Primary&gt;N</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#008000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"90\"",
"         id=\"text3606\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan3608\"",
"           x=\"5\"",
"           y=\"90\">key:&lt;Shift&gt;N</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#0000ff;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"105\"",
"         id=\"text3610\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan3612\"",
"           x=\"5\"",
"           y=\"105\">key:&lt;Alt&gt;N</tspan></text>",
"      <text",
"         id=\"text6667\"",
"         y=\"120\"",
"         x=\"5\"",
"         style=\"font-style:normal;font-weight:normal;font-size:12.00000048px;font-family:'Bitstream Vera Sans';fill:#464646;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;-inkscape-font-specification:'Bitstream Vera Sans, Normal';font-stretch:normal;font-variant:normal;text-anchor:start;text-align:start;writing-mode:lr;line-height:125%;\"",
"         xml:space=\"preserve\"",
"         sodipodi:linespacing=\"125%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan6671\">key:&lt;Super&gt;N</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#808000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"135\"",
"         id=\"text3614\"",
"         sodipodi:linespacing=\"100%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan3616\"",
"           x=\"5\"",
"           y=\"135\">key:&lt;Primary&gt;&lt;Shift&gt;N</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#ff00ff;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"150\"",
"         id=\"text3618\"",
"         sodipodi:linespacing=\"100%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan3620\"",
"           x=\"5\"",
"           y=\"150\">key:&lt;Primary&gt;&lt;Alt&gt;N</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#008080;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"165\"",
"         id=\"text3622\"",
"         sodipodi:linespacing=\"100%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan3624\"",
"           x=\"5\"",
"           y=\"165\">key:&lt;Shift&gt;&lt;Alt&gt;N</tspan></text>",
"    </g>",
"    <g",
"       id=\"g3626\"",
"       transform=\"translate(1310,890)\">",
"      <rect",
"         y=\"2.9921875\"",
"         x=\"0\"",
"         height=\"170\"",
"         width=\"170\"",
"         id=\"rect3628\"",
"         style=\"fill:#0000ff;fill-opacity:0.1;stroke:#000000;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1\" />",
"      <text",
"         sodipodi:linespacing=\"100%\"",
"         id=\"text3630\"",
"         y=\"42.992188\"",
"         x=\"40\"",
"         style=\"font-size:48px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#000000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"42.992188\"",
"           x=\"40\"",
"           id=\"tspan3632\"",
"           sodipodi:role=\"line\">B</tspan></text>",
"      <text",
"         id=\"text3634\"",
"         y=\"60\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#000000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"60\"",
"           x=\"5\"",
"           id=\"tspan3636\"",
"           sodipodi:role=\"line\">key:B</tspan></text>",
"      <text",
"         id=\"text3638\"",
"         y=\"75\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#ff0000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           id=\"tspan3640\"",
"           y=\"75\"",
"           x=\"5\"",
"           sodipodi:role=\"line\">key:&lt;Primary&gt;B</tspan></text>",
"      <text",
"         id=\"text3642\"",
"         y=\"90\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#008000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"90\"",
"           x=\"5\"",
"           id=\"tspan3644\"",
"           sodipodi:role=\"line\">key:&lt;Shift&gt;B</tspan></text>",
"      <text",
"         id=\"text3646\"",
"         y=\"105\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#0000ff;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"105\"",
"           x=\"5\"",
"           id=\"tspan3648\"",
"           sodipodi:role=\"line\">key:&lt;Alt&gt;B</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-style:normal;font-weight:normal;font-size:12.00000048px;font-family:'Bitstream Vera Sans';fill:#464646;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;-inkscape-font-specification:'Bitstream Vera Sans, Normal';font-stretch:normal;font-variant:normal;text-anchor:start;text-align:start;writing-mode:lr;line-height:125%;\"",
"         x=\"5\"",
"         y=\"120\"",
"         id=\"text6661\"",
"         sodipodi:linespacing=\"125%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan6665\">key:&lt;Super&gt;B</tspan></text>",
"      <text",
"         sodipodi:linespacing=\"100%\"",
"         id=\"text3650\"",
"         y=\"135\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#808000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"135\"",
"           x=\"5\"",
"           id=\"tspan3652\"",
"           sodipodi:role=\"line\">key:&lt;Primary&gt;&lt;Shift&gt;B</tspan></text>",
"      <text",
"         sodipodi:linespacing=\"100%\"",
"         id=\"text3654\"",
"         y=\"150\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#ff00ff;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"150\"",
"           x=\"5\"",
"           id=\"tspan3656\"",
"           sodipodi:role=\"line\">key:&lt;Primary&gt;&lt;Alt&gt;B</tspan></text>",
"      <text",
"         sodipodi:linespacing=\"100%\"",
"         id=\"text3658\"",
"         y=\"165\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#008080;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"165\"",
"           x=\"5\"",
"           id=\"tspan3660\"",
"           sodipodi:role=\"line\">key:&lt;Shift&gt;&lt;Alt&gt;B</tspan></text>",
"    </g>",
"    <g",
"       id=\"g3662\"",
"       transform=\"translate(2500,350)\">",
"      <rect",
"         y=\"2.9921875\"",
"         x=\"0\"",
"         height=\"170\"",
"         width=\"170\"",
"         id=\"rect3664\"",
"         style=\"fill:#0000ff;fill-opacity:0.1;stroke:#000000;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1\" />",
"      <text",
"         sodipodi:linespacing=\"100%\"",
"         id=\"text3666\"",
"         y=\"42.992188\"",
"         x=\"40\"",
"         style=\"font-size:48px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#000000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"42.992188\"",
"           x=\"40\"",
"           id=\"tspan3668\"",
"           sodipodi:role=\"line\">\\</tspan></text>",
"      <text",
"         id=\"text3670\"",
"         y=\"60\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#000000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"60\"",
"           x=\"5\"",
"           id=\"tspan3672\"",
"           sodipodi:role=\"line\">key:backslash</tspan></text>",
"      <text",
"         id=\"text3674\"",
"         y=\"75\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#ff0000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           id=\"tspan3676\"",
"           y=\"75\"",
"           x=\"5\"",
"           sodipodi:role=\"line\">key:&lt;Primary&gt;backslash</tspan></text>",
"      <text",
"         id=\"text3678\"",
"         y=\"90\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#008000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"90\"",
"           x=\"5\"",
"           id=\"tspan3680\"",
"           sodipodi:role=\"line\">key:&lt;Shift&gt;bar</tspan></text>",
"      <text",
"         id=\"text3682\"",
"         y=\"105\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#0000ff;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"105\"",
"           x=\"5\"",
"           id=\"tspan3684\"",
"           sodipodi:role=\"line\">key:&lt;Alt&gt;backslash</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-style:normal;font-weight:normal;font-size:12.00000048px;font-family:'Bitstream Vera Sans';fill:#464646;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;-inkscape-font-specification:'Bitstream Vera Sans, Normal';font-stretch:normal;font-variant:normal;text-anchor:start;text-align:start;writing-mode:lr;line-height:125%;\"",
"         x=\"5\"",
"         y=\"120\"",
"         id=\"text6383\"",
"         sodipodi:linespacing=\"125%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan6485\">key:&lt;Super&gt;backslash</tspan></text>",
"      <text",
"         sodipodi:linespacing=\"100%\"",
"         id=\"text3686\"",
"         y=\"135\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#808000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"135\"",
"           x=\"5\"",
"           id=\"tspan3688\"",
"           sodipodi:role=\"line\">key:&lt;Primary&gt;&lt;Shift&gt;bar</tspan></text>",
"      <text",
"         sodipodi:linespacing=\"100%\"",
"         id=\"text3690\"",
"         y=\"150\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#ff00ff;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"150\"",
"           x=\"5\"",
"           id=\"tspan3692\"",
"           sodipodi:role=\"line\">key:&lt;Primary&gt;&lt;Alt&gt;backslash</tspan></text>",
"      <text",
"         sodipodi:linespacing=\"100%\"",
"         id=\"text3694\"",
"         y=\"165\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#008080;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"165\"",
"           x=\"5\"",
"           id=\"tspan3696\"",
"           sodipodi:role=\"line\">key:&lt;Shift&gt;&lt;Alt&gt;bar</tspan></text>",
"    </g>",
"    <g",
"       transform=\"translate(2320,350)\"",
"       id=\"g3698\">",
"      <rect",
"         style=\"fill:#0000ff;fill-opacity:0.1;stroke:#000000;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1\"",
"         id=\"rect3700\"",
"         width=\"170\"",
"         height=\"170\"",
"         x=\"0\"",
"         y=\"2.9921875\" />",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:48px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#000000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         x=\"40\"",
"         y=\"42.992188\"",
"         id=\"text3702\"",
"         sodipodi:linespacing=\"100%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan3704\"",
"           x=\"40\"",
"           y=\"42.992188\">=</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#000000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"60\"",
"         id=\"text3706\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan3708\"",
"           x=\"5\"",
"           y=\"60\">key:Equal</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#ff0000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"75\"",
"         id=\"text3710\"><tspan",
"           sodipodi:role=\"line\"",
"           x=\"5\"",
"           y=\"75\"",
"           id=\"tspan3712\">key:&lt;Primary&gt;Equal</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12.00000048px;font-style:normal;font-weight:normal;fill:#008000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:'Bitstream Vera Sans';-inkscape-font-specification:'Bitstream Vera Sans, Normal';font-stretch:normal;font-variant:normal;text-anchor:start;text-align:start;writing-mode:lr;line-height:125%;\"",
"         x=\"5\"",
"         y=\"90\"",
"         id=\"text3714\"",
"         sodipodi:linespacing=\"125%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan11561\">key:&lt;Shift&gt;Equal</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#0000ff;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"105\"",
"         id=\"text3718\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan3720\"",
"           x=\"5\"",
"           y=\"105\">key:&lt;Alt&gt;Equal</tspan></text>",
"      <text",
"         id=\"text6379\"",
"         y=\"120\"",
"         x=\"5\"",
"         style=\"font-style:normal;font-weight:normal;font-size:12.00000048px;font-family:'Bitstream Vera Sans';fill:#464646;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;-inkscape-font-specification:'Bitstream Vera Sans, Normal';font-stretch:normal;font-variant:normal;text-anchor:start;text-align:start;writing-mode:lr;line-height:125%;\"",
"         xml:space=\"preserve\"",
"         sodipodi:linespacing=\"125%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan6483\">key:&lt;Super&gt;Equal</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#808000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"135\"",
"         id=\"text3722\"",
"         sodipodi:linespacing=\"100%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan3724\"",
"           x=\"5\"",
"           y=\"135\">key:&lt;Primary&gt;&lt;Shift&gt;Equal</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#ff00ff;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"150\"",
"         id=\"text3726\"",
"         sodipodi:linespacing=\"100%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan3728\"",
"           x=\"5\"",
"           y=\"150\">key:&lt;Primary&gt;&lt;Alt&gt;Equal</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#008080;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"165\"",
"         id=\"text3730\"",
"         sodipodi:linespacing=\"100%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan3732\"",
"           x=\"5\"",
"           y=\"165\">key:&lt;Shift&gt;&lt;Alt&gt;Equal</tspan></text>",
"    </g>",
"    <g",
"       id=\"g3734\"",
"       transform=\"translate(2140,350)\">",
"      <rect",
"         y=\"2.9921875\"",
"         x=\"0\"",
"         height=\"170\"",
"         width=\"170\"",
"         id=\"rect3736\"",
"         style=\"fill:#0000ff;fill-opacity:0.1;stroke:#000000;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1\" />",
"      <text",
"         sodipodi:linespacing=\"100%\"",
"         id=\"text3738\"",
"         y=\"42.992188\"",
"         x=\"40\"",
"         style=\"font-size:48px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#000000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"42.992188\"",
"           x=\"40\"",
"           id=\"tspan3740\"",
"           sodipodi:role=\"line\">-</tspan></text>",
"      <text",
"         id=\"text3742\"",
"         y=\"60\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#000000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"60\"",
"           x=\"5\"",
"           id=\"tspan3744\"",
"           sodipodi:role=\"line\">key:Minus</tspan></text>",
"      <text",
"         id=\"text3746\"",
"         y=\"75\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#ff0000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           id=\"tspan3748\"",
"           y=\"75\"",
"           x=\"5\"",
"           sodipodi:role=\"line\">key:&lt;Primary&gt;Minus</tspan></text>",
"      <text",
"         id=\"text3750\"",
"         y=\"90\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#008000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"90\"",
"           x=\"5\"",
"           id=\"tspan3752\"",
"           sodipodi:role=\"line\">key:&lt;Shift&gt;Minus</tspan></text>",
"      <text",
"         id=\"text3754\"",
"         y=\"105\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#0000ff;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"105\"",
"           x=\"5\"",
"           id=\"tspan3756\"",
"           sodipodi:role=\"line\">key:&lt;Alt&gt;Minus</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-style:normal;font-weight:normal;font-size:12.00000048px;font-family:'Bitstream Vera Sans';fill:#464646;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;-inkscape-font-specification:'Bitstream Vera Sans, Normal';font-stretch:normal;font-variant:normal;text-anchor:start;text-align:start;writing-mode:lr;line-height:125%;\"",
"         x=\"5\"",
"         y=\"120\"",
"         id=\"text6363\"",
"         sodipodi:linespacing=\"125%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan6481\">key:&lt;Super&gt;Minus</tspan></text>",
"      <text",
"         sodipodi:linespacing=\"100%\"",
"         id=\"text3758\"",
"         y=\"135\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#808000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"135\"",
"           x=\"5\"",
"           id=\"tspan3760\"",
"           sodipodi:role=\"line\">key:&lt;Primary&gt;&lt;Shift&gt;Minus</tspan></text>",
"      <text",
"         sodipodi:linespacing=\"100%\"",
"         id=\"text3762\"",
"         y=\"150\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#ff00ff;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"150\"",
"           x=\"5\"",
"           id=\"tspan3764\"",
"           sodipodi:role=\"line\">key:&lt;Primary&gt;&lt;Alt&gt;Minus</tspan></text>",
"      <text",
"         sodipodi:linespacing=\"100%\"",
"         id=\"text3766\"",
"         y=\"165\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#008080;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"165\"",
"           x=\"5\"",
"           id=\"tspan3768\"",
"           sodipodi:role=\"line\">key:&lt;Shift&gt;&lt;Alt&gt;Minus</tspan></text>",
"    </g>",
"    <g",
"       transform=\"translate(1960,350)\"",
"       id=\"g3770\">",
"      <rect",
"         style=\"fill:#0000ff;fill-opacity:0.1;stroke:#000000;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1\"",
"         id=\"rect3772\"",
"         width=\"170\"",
"         height=\"170\"",
"         x=\"0\"",
"         y=\"2.9921875\" />",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:48px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#000000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         x=\"40\"",
"         y=\"42.992188\"",
"         id=\"text3774\"",
"         sodipodi:linespacing=\"100%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan3776\"",
"           x=\"40\"",
"           y=\"42.992188\">0</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#000000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"60\"",
"         id=\"text3778\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan3780\"",
"           x=\"5\"",
"           y=\"60\">key:0</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#ff0000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"75\"",
"         id=\"text3782\"><tspan",
"           sodipodi:role=\"line\"",
"           x=\"5\"",
"           y=\"75\"",
"           id=\"tspan3784\">key:&lt;Primary&gt;0</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#008000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"90\"",
"         id=\"text3786\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan3788\"",
"           x=\"5\"",
"           y=\"90\">key:&lt;Shift&gt;parenright</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#0000ff;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"105\"",
"         id=\"text3790\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan3792\"",
"           x=\"5\"",
"           y=\"105\">key:&lt;Alt&gt;0</tspan></text>",
"      <text",
"         id=\"text6359\"",
"         y=\"120\"",
"         x=\"5\"",
"         style=\"font-style:normal;font-weight:normal;font-size:12.00000048px;font-family:'Bitstream Vera Sans';fill:#464646;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;-inkscape-font-specification:'Bitstream Vera Sans, Normal';font-stretch:normal;font-variant:normal;text-anchor:start;text-align:start;writing-mode:lr;line-height:125%;\"",
"         xml:space=\"preserve\"",
"         sodipodi:linespacing=\"125%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan6479\">key:&lt;Super&gt;0</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#808000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"135\"",
"         id=\"text3794\"",
"         sodipodi:linespacing=\"100%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan3796\"",
"           x=\"5\"",
"           y=\"135\">key:&lt;Primary&gt;&lt;Shift&gt;parenright</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#ff00ff;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"150\"",
"         id=\"text3798\"",
"         sodipodi:linespacing=\"100%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan3800\"",
"           x=\"5\"",
"           y=\"150\">key:&lt;Primary&gt;&lt;Alt&gt;0</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#008080;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"165\"",
"         id=\"text3802\"",
"         sodipodi:linespacing=\"100%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan3804\"",
"           x=\"5\"",
"           y=\"165\">key:&lt;Shift&gt;&lt;Alt&gt;parenright</tspan></text>",
"    </g>",
"    <g",
"       transform=\"translate(2430,530)\"",
"       id=\"g3806\">",
"      <rect",
"         style=\"fill:#0000ff;fill-opacity:0.1;stroke:#000000;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1\"",
"         id=\"rect3808\"",
"         width=\"170\"",
"         height=\"170\"",
"         x=\"0\"",
"         y=\"2.9921875\" />",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:48px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#000000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         x=\"40\"",
"         y=\"42.992188\"",
"         id=\"text3810\"",
"         sodipodi:linespacing=\"100%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan3812\"",
"           x=\"40\"",
"           y=\"42.992188\">]</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#000000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"60\"",
"         id=\"text3814\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan3816\"",
"           x=\"5\"",
"           y=\"60\">key:BracketRight</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#ff0000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"75\"",
"         id=\"text3818\"><tspan",
"           sodipodi:role=\"line\"",
"           x=\"5\"",
"           y=\"75\"",
"           id=\"tspan3820\">key:&lt;Primary&gt;BracketRight</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#008000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"90\"",
"         id=\"text3822\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan3824\"",
"           x=\"5\"",
"           y=\"90\">key:&lt;Shift&gt;BracketRight</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#0000ff;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"105\"",
"         id=\"text3826\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan3828\"",
"           x=\"5\"",
"           y=\"105\">key:&lt;Alt&gt;BracketRight</tspan></text>",
"      <text",
"         id=\"text6555\"",
"         y=\"120\"",
"         x=\"5\"",
"         style=\"font-style:normal;font-weight:normal;font-size:12.00000048px;font-family:'Bitstream Vera Sans';fill:#464646;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;-inkscape-font-specification:'Bitstream Vera Sans, Normal';font-stretch:normal;font-variant:normal;text-anchor:start;text-align:start;writing-mode:lr;line-height:125%;\"",
"         xml:space=\"preserve\"",
"         sodipodi:linespacing=\"125%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan6559\">key:&lt;Super&gt;BracketRight</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#808000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"135\"",
"         id=\"text3830\"",
"         sodipodi:linespacing=\"100%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan3832\"",
"           x=\"5\"",
"           y=\"135\">key:&lt;Primary&gt;&lt;Shift&gt;BracketRight</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#ff00ff;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"150\"",
"         id=\"text3834\"",
"         sodipodi:linespacing=\"100%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan3836\"",
"           x=\"5\"",
"           y=\"150\">key:&lt;Primary&gt;&lt;Alt&gt;BracketRight</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#008080;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"165\"",
"         id=\"text3838\"",
"         sodipodi:linespacing=\"100%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan3840\"",
"           x=\"5\"",
"           y=\"165\">key:&lt;Shift&gt;&lt;Alt&gt;BracketRight</tspan></text>",
"    </g>",
"    <g",
"       id=\"g3842\"",
"       transform=\"translate(2250,530)\">",
"      <rect",
"         y=\"2.9921875\"",
"         x=\"0\"",
"         height=\"170\"",
"         width=\"170\"",
"         id=\"rect3844\"",
"         style=\"fill:#0000ff;fill-opacity:0.1;stroke:#000000;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1\" />",
"      <text",
"         sodipodi:linespacing=\"100%\"",
"         id=\"text3846\"",
"         y=\"42.992188\"",
"         x=\"40\"",
"         style=\"font-size:48px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#000000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"42.992188\"",
"           x=\"40\"",
"           id=\"tspan3848\"",
"           sodipodi:role=\"line\">[</tspan></text>",
"      <text",
"         id=\"text3850\"",
"         y=\"60\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#000000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"60\"",
"           x=\"5\"",
"           id=\"tspan3852\"",
"           sodipodi:role=\"line\">key:BracketLeft</tspan></text>",
"      <text",
"         id=\"text3854\"",
"         y=\"75\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#ff0000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           id=\"tspan3856\"",
"           y=\"75\"",
"           x=\"5\"",
"           sodipodi:role=\"line\">key:&lt;Primary&gt;BracketLeft</tspan></text>",
"      <text",
"         id=\"text3858\"",
"         y=\"90\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#008000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"90\"",
"           x=\"5\"",
"           id=\"tspan3860\"",
"           sodipodi:role=\"line\">key:&lt;Shift&gt;BracketLeft</tspan></text>",
"      <text",
"         id=\"text3862\"",
"         y=\"105\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#0000ff;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"105\"",
"           x=\"5\"",
"           id=\"tspan3864\"",
"           sodipodi:role=\"line\">key:&lt;Alt&gt;BracketLeft</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-style:normal;font-weight:normal;font-size:12.00000048px;font-family:'Bitstream Vera Sans';fill:#464646;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;-inkscape-font-specification:'Bitstream Vera Sans, Normal';font-stretch:normal;font-variant:normal;text-anchor:start;text-align:start;writing-mode:lr;line-height:125%;\"",
"         x=\"5\"",
"         y=\"120\"",
"         id=\"text6549\"",
"         sodipodi:linespacing=\"125%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan6553\">key:&lt;Super&gt;BracketLeft</tspan></text>",
"      <text",
"         sodipodi:linespacing=\"100%\"",
"         id=\"text3866\"",
"         y=\"135\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#808000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"135\"",
"           x=\"5\"",
"           id=\"tspan3868\"",
"           sodipodi:role=\"line\">key:&lt;Primary&gt;&lt;Shift&gt;BracketLeft</tspan></text>",
"      <text",
"         sodipodi:linespacing=\"100%\"",
"         id=\"text3870\"",
"         y=\"150\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#ff00ff;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"150\"",
"           x=\"5\"",
"           id=\"tspan3872\"",
"           sodipodi:role=\"line\">key:&lt;Primary&gt;&lt;Alt&gt;BracketLeft</tspan></text>",
"      <text",
"         sodipodi:linespacing=\"100%\"",
"         id=\"text3874\"",
"         y=\"165\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#008080;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"165\"",
"           x=\"5\"",
"           id=\"tspan3876\"",
"           sodipodi:role=\"line\">key:&lt;Shift&gt;&lt;Alt&gt;BracketLeft</tspan></text>",
"    </g>",
"    <g",
"       transform=\"translate(2070,530)\"",
"       id=\"g3878\">",
"      <rect",
"         style=\"fill:#0000ff;fill-opacity:0.1;stroke:#000000;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1\"",
"         id=\"rect3880\"",
"         width=\"170\"",
"         height=\"170\"",
"         x=\"0\"",
"         y=\"2.9921875\" />",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:48px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#000000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         x=\"40\"",
"         y=\"42.992188\"",
"         id=\"text3882\"",
"         sodipodi:linespacing=\"100%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan3884\"",
"           x=\"40\"",
"           y=\"42.992188\">P</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#000000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"60\"",
"         id=\"text3886\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan3888\"",
"           x=\"5\"",
"           y=\"60\">key:P</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#ff0000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"75\"",
"         id=\"text3890\"><tspan",
"           sodipodi:role=\"line\"",
"           x=\"5\"",
"           y=\"75\"",
"           id=\"tspan3892\">key:&lt;Primary&gt;P</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#008000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"90\"",
"         id=\"text3894\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan3896\"",
"           x=\"5\"",
"           y=\"90\">key:&lt;Shift&gt;P</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#0000ff;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"105\"",
"         id=\"text3898\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan3900\"",
"           x=\"5\"",
"           y=\"105\">key:&lt;Alt&gt;P</tspan></text>",
"      <text",
"         id=\"text6543\"",
"         y=\"120\"",
"         x=\"5\"",
"         style=\"font-style:normal;font-weight:normal;font-size:12.00000048px;font-family:'Bitstream Vera Sans';fill:#464646;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;-inkscape-font-specification:'Bitstream Vera Sans, Normal';font-stretch:normal;font-variant:normal;text-anchor:start;text-align:start;writing-mode:lr;line-height:125%;\"",
"         xml:space=\"preserve\"",
"         sodipodi:linespacing=\"125%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan6547\">key:&lt;Super&gt;P</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#808000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"135\"",
"         id=\"text3902\"",
"         sodipodi:linespacing=\"100%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan3904\"",
"           x=\"5\"",
"           y=\"135\">key:&lt;Primary&gt;&lt;Shift&gt;P</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#ff00ff;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"150\"",
"         id=\"text3906\"",
"         sodipodi:linespacing=\"100%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan3908\"",
"           x=\"5\"",
"           y=\"150\">key:&lt;Primary&gt;&lt;Alt&gt;P</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#008080;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"165\"",
"         id=\"text3910\"",
"         sodipodi:linespacing=\"100%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan3912\"",
"           x=\"5\"",
"           y=\"165\">key:&lt;Shift&gt;&lt;Alt&gt;P</tspan></text>",
"    </g>",
"    <g",
"       transform=\"translate(2680,350)\"",
"       id=\"g3914\">",
"      <rect",
"         style=\"fill:#0000ff;fill-opacity:0.1;stroke:#000000;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1\"",
"         id=\"rect3916\"",
"         width=\"170\"",
"         height=\"170\"",
"         x=\"0\"",
"         y=\"2.9921875\" />",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:48px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#000000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         x=\"40\"",
"         y=\"42.992188\"",
"         id=\"text3918\"",
"         sodipodi:linespacing=\"100%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan3920\"",
"           x=\"40\"",
"           y=\"42.992188\">BS</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#000000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"60\"",
"         id=\"text3922\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan3924\"",
"           x=\"5\"",
"           y=\"60\">key:Backspace</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#ff0000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"75\"",
"         id=\"text3926\"><tspan",
"           sodipodi:role=\"line\"",
"           x=\"5\"",
"           y=\"75\"",
"           id=\"tspan3928\">key:&lt;Primary&gt;Backspace</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#008000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"90\"",
"         id=\"text3930\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan3932\"",
"           x=\"5\"",
"           y=\"90\">key:&lt;Shift&gt;Backspace</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#0000ff;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"105\"",
"         id=\"text3934\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan3936\"",
"           x=\"5\"",
"           y=\"105\">key:&lt;Alt&gt;Backspace</tspan></text>",
"      <text",
"         id=\"text6387\"",
"         y=\"120\"",
"         x=\"5\"",
"         style=\"font-style:normal;font-weight:normal;font-size:12.00000048px;font-family:'Bitstream Vera Sans';fill:#464646;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;-inkscape-font-specification:'Bitstream Vera Sans, Normal';font-stretch:normal;font-variant:normal;text-anchor:start;text-align:start;writing-mode:lr;line-height:125%;\"",
"         xml:space=\"preserve\"",
"         sodipodi:linespacing=\"125%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan6487\">key:&lt;Super&gt;Backspace</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#808000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"135\"",
"         id=\"text3938\"",
"         sodipodi:linespacing=\"100%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan3940\"",
"           x=\"5\"",
"           y=\"135\">key:&lt;Primary&gt;&lt;Shift&gt;Backspace</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#ff00ff;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"150\"",
"         id=\"text3942\"",
"         sodipodi:linespacing=\"100%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan3944\"",
"           x=\"5\"",
"           y=\"150\">key:&lt;Primary&gt;&lt;Alt&gt;Backspace</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#008080;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"165\"",
"         id=\"text3946\"",
"         sodipodi:linespacing=\"100%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan3948\"",
"           x=\"5\"",
"           y=\"165\">key:&lt;Shift&gt;&lt;Alt&gt;Backspace</tspan></text>",
"    </g>",
"    <g",
"       transform=\"translate(2300,710)\"",
"       id=\"g3986\">",
"      <rect",
"         style=\"fill:#0000ff;fill-opacity:0.1;stroke:#000000;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1\"",
"         id=\"rect3988\"",
"         width=\"170\"",
"         height=\"170\"",
"         x=\"0\"",
"         y=\"2.9921875\" />",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:48px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#000000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         x=\"40\"",
"         y=\"42.992188\"",
"         id=\"text3990\"",
"         sodipodi:linespacing=\"100%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan3992\"",
"           x=\"40\"",
"           y=\"42.992188\">'</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#000000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"60\"",
"         id=\"text3994\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan3996\"",
"           x=\"5\"",
"           y=\"60\">key:apostrophe</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#ff0000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"75\"",
"         id=\"text3998\"><tspan",
"           sodipodi:role=\"line\"",
"           x=\"5\"",
"           y=\"75\"",
"           id=\"tspan4000\">key:&lt;Primary&gt;apostrophe</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#008000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"90\"",
"         id=\"text4002\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan4004\"",
"           x=\"5\"",
"           y=\"90\">key:&lt;Shift&gt;at</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#0000ff;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"105\"",
"         id=\"text4006\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan4008\"",
"           x=\"5\"",
"           y=\"105\">key:&lt;Alt&gt;apostrophe</tspan></text>",
"      <text",
"         id=\"text6619\"",
"         y=\"120\"",
"         x=\"5\"",
"         style=\"font-style:normal;font-weight:normal;font-size:12.00000048px;font-family:'Bitstream Vera Sans';fill:#464646;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;-inkscape-font-specification:'Bitstream Vera Sans, Normal';font-stretch:normal;font-variant:normal;text-anchor:start;text-align:start;writing-mode:lr;line-height:125%;\"",
"         xml:space=\"preserve\"",
"         sodipodi:linespacing=\"125%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan6623\">key:&lt;Super&gt;apostrophe</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#808000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"135\"",
"         id=\"text4010\"",
"         sodipodi:linespacing=\"100%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan4012\"",
"           x=\"5\"",
"           y=\"135\">key:&lt;Primary&gt;&lt;Shift&gt;at</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#ff00ff;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"150\"",
"         id=\"text4014\"",
"         sodipodi:linespacing=\"100%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan4016\"",
"           x=\"5\"",
"           y=\"150\">key:&lt;Primary&gt;&lt;Alt&gt;apostrophe</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#008080;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"165\"",
"         id=\"text4018\"",
"         sodipodi:linespacing=\"100%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan4020\"",
"           x=\"5\"",
"           y=\"165\">key:&lt;Shift&gt;&lt;Alt&gt;at</tspan></text>",
"    </g>",
"    <g",
"       id=\"g4022\"",
"       transform=\"translate(2120,710)\">",
"      <rect",
"         y=\"2.9921875\"",
"         x=\"0\"",
"         height=\"170\"",
"         width=\"170\"",
"         id=\"rect4024\"",
"         style=\"fill:#0000ff;fill-opacity:0.1;stroke:#000000;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1\" />",
"      <text",
"         sodipodi:linespacing=\"100%\"",
"         id=\"text4026\"",
"         y=\"42.992188\"",
"         x=\"40\"",
"         style=\"font-size:48px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#000000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"42.992188\"",
"           x=\"40\"",
"           id=\"tspan4028\"",
"           sodipodi:role=\"line\">;</tspan></text>",
"      <text",
"         id=\"text4030\"",
"         y=\"60\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#000000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"60\"",
"           x=\"5\"",
"           id=\"tspan4032\"",
"           sodipodi:role=\"line\">key:semicolon</tspan></text>",
"      <text",
"         id=\"text4034\"",
"         y=\"75\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#ff0000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           id=\"tspan4036\"",
"           y=\"75\"",
"           x=\"5\"",
"           sodipodi:role=\"line\">key:&lt;Primary&gt;semicolon</tspan></text>",
"      <text",
"         id=\"text4038\"",
"         y=\"90\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#008000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"90\"",
"           x=\"5\"",
"           id=\"tspan4040\"",
"           sodipodi:role=\"line\">key:&lt;Shift&gt;colon</tspan></text>",
"      <text",
"         id=\"text4042\"",
"         y=\"105\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#0000ff;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"105\"",
"           x=\"5\"",
"           id=\"tspan4044\"",
"           sodipodi:role=\"line\">key:&lt;Alt&gt;semicolon</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-style:normal;font-weight:normal;font-size:12.00000048px;font-family:'Bitstream Vera Sans';fill:#464646;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;-inkscape-font-specification:'Bitstream Vera Sans, Normal';font-stretch:normal;font-variant:normal;text-anchor:start;text-align:start;writing-mode:lr;line-height:125%;\"",
"         x=\"5\"",
"         y=\"120\"",
"         id=\"text6613\"",
"         sodipodi:linespacing=\"125%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan6617\">key:&lt;Super&gt;semicolon</tspan></text>",
"      <text",
"         sodipodi:linespacing=\"100%\"",
"         id=\"text4046\"",
"         y=\"135\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#808000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"135\"",
"           x=\"5\"",
"           id=\"tspan4048\"",
"           sodipodi:role=\"line\">key:&lt;Primary&gt;&lt;Shift&gt;colon</tspan></text>",
"      <text",
"         sodipodi:linespacing=\"100%\"",
"         id=\"text4050\"",
"         y=\"150\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#ff00ff;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"150\"",
"           x=\"5\"",
"           id=\"tspan4052\"",
"           sodipodi:role=\"line\">key:&lt;Primary&gt;&lt;Alt&gt;semicolon</tspan></text>",
"      <text",
"         sodipodi:linespacing=\"100%\"",
"         id=\"text4054\"",
"         y=\"165\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#008080;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"165\"",
"           x=\"5\"",
"           id=\"tspan4056\"",
"           sodipodi:role=\"line\">key:&lt;Shift&gt;&lt;Alt&gt;colon</tspan></text>",
"    </g>",
"    <g",
"       id=\"g4058\"",
"       transform=\"translate(2210,890)\">",
"      <rect",
"         y=\"2.9921875\"",
"         x=\"0\"",
"         height=\"170\"",
"         width=\"170\"",
"         id=\"rect4060\"",
"         style=\"fill:#0000ff;fill-opacity:0.1;stroke:#000000;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1\" />",
"      <text",
"         sodipodi:linespacing=\"100%\"",
"         id=\"text4062\"",
"         y=\"42.992188\"",
"         x=\"40\"",
"         style=\"font-size:48px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#000000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"42.992188\"",
"           x=\"40\"",
"           id=\"tspan4064\"",
"           sodipodi:role=\"line\">/</tspan></text>",
"      <text",
"         id=\"text4066\"",
"         y=\"60\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#000000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"60\"",
"           x=\"5\"",
"           id=\"tspan4068\"",
"           sodipodi:role=\"line\">key:slash</tspan></text>",
"      <text",
"         id=\"text4070\"",
"         y=\"75\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#ff0000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           id=\"tspan4072\"",
"           y=\"75\"",
"           x=\"5\"",
"           sodipodi:role=\"line\">key:&lt;Primary&gt;slash</tspan></text>",
"      <text",
"         id=\"text4074\"",
"         y=\"90\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#008000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"90\"",
"           x=\"5\"",
"           id=\"tspan4076\"",
"           sodipodi:role=\"line\">key:&lt;Shift&gt;question</tspan></text>",
"      <text",
"         id=\"text4078\"",
"         y=\"105\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#0000ff;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"105\"",
"           x=\"5\"",
"           id=\"tspan4080\"",
"           sodipodi:role=\"line\">key:&lt;Alt&gt;slash</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-style:normal;font-weight:normal;font-size:12.00000048px;font-family:'Bitstream Vera Sans';fill:#464646;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;-inkscape-font-specification:'Bitstream Vera Sans, Normal';font-stretch:normal;font-variant:normal;text-anchor:start;text-align:start;writing-mode:lr;line-height:125%;\"",
"         x=\"5\"",
"         y=\"120\"",
"         id=\"text6691\"",
"         sodipodi:linespacing=\"125%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan6953\">key:&lt;Super&gt;slash</tspan></text>",
"      <text",
"         sodipodi:linespacing=\"100%\"",
"         id=\"text4082\"",
"         y=\"135\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#808000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"135\"",
"           x=\"5\"",
"           id=\"tspan4084\"",
"           sodipodi:role=\"line\">key:&lt;Primary&gt;&lt;Shift&gt;question</tspan></text>",
"      <text",
"         sodipodi:linespacing=\"100%\"",
"         id=\"text4086\"",
"         y=\"150\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#ff00ff;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"150\"",
"           x=\"5\"",
"           id=\"tspan4088\"",
"           sodipodi:role=\"line\">key:&lt;Primary&gt;&lt;Alt&gt;slash</tspan></text>",
"      <text",
"         sodipodi:linespacing=\"100%\"",
"         id=\"text4090\"",
"         y=\"165\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#008080;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"165\"",
"           x=\"5\"",
"           id=\"tspan4092\"",
"           sodipodi:role=\"line\">key:&lt;Shift&gt;&lt;Alt&gt;question</tspan></text>",
"    </g>",
"    <g",
"       transform=\"translate(160,350)\"",
"       id=\"g4094\">",
"      <rect",
"         style=\"fill:#0000ff;fill-opacity:0.1;stroke:#000000;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1\"",
"         id=\"rect4096\"",
"         width=\"170\"",
"         height=\"170\"",
"         x=\"0\"",
"         y=\"2.9921875\" />",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:48px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#000000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         x=\"40\"",
"         y=\"42.992188\"",
"         id=\"text4098\"",
"         sodipodi:linespacing=\"100%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan4100\"",
"           x=\"40\"",
"           y=\"42.992188\">`</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#000000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"60\"",
"         id=\"text4102\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan4104\"",
"           x=\"5\"",
"           y=\"60\">key:grave</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#ff0000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"75\"",
"         id=\"text4106\"><tspan",
"           sodipodi:role=\"line\"",
"           x=\"5\"",
"           y=\"75\"",
"           id=\"tspan4108\">key:&lt;Primary&gt;grave</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#008000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"90\"",
"         id=\"text4110\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan4112\"",
"           x=\"5\"",
"           y=\"90\">key:&lt;Shift&gt;notsign</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#0000ff;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"105\"",
"         id=\"text4114\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan4116\"",
"           x=\"5\"",
"           y=\"105\">key:&lt;Alt&gt;grave</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#808000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"135\"",
"         id=\"text4118\"",
"         sodipodi:linespacing=\"100%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan4120\"",
"           x=\"5\"",
"           y=\"135\">key:&lt;Primary&gt;&lt;Shift&gt;notsign</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#ff00ff;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"150\"",
"         id=\"text4122\"",
"         sodipodi:linespacing=\"100%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan4124\"",
"           x=\"5\"",
"           y=\"150\">key:&lt;Primary&gt;&lt;Alt&gt;grave</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#008080;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"165\"",
"         id=\"text4126\"",
"         sodipodi:linespacing=\"100%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan4128\"",
"           x=\"5\"",
"           y=\"165\">key:&lt;Shift&gt;&lt;Alt&gt;notsign</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-style:normal;font-weight:normal;font-size:12px;font-family:'Bitstream Vera Sans';fill:#464646;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1\"",
"         x=\"5\"",
"         y=\"120\"",
"         id=\"text4114-0\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan4116-5\"",
"           x=\"5\"",
"           y=\"120\">key:&lt;Super&gt;grave</tspan></text>",
"    </g>",
"    <g",
"       id=\"g4130\"",
"       transform=\"translate(160,130)\">",
"      <rect",
"         y=\"2.9921875\"",
"         x=\"0\"",
"         height=\"170\"",
"         width=\"170\"",
"         id=\"rect4132\"",
"         style=\"fill:#0000ff;fill-opacity:0.1;stroke:#000000;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1\" />",
"      <text",
"         sodipodi:linespacing=\"100%\"",
"         id=\"text4134\"",
"         y=\"42.992188\"",
"         x=\"40\"",
"         style=\"font-size:48px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#000000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"42.992188\"",
"           x=\"40\"",
"           id=\"tspan4136\"",
"           sodipodi:role=\"line\">ESC</tspan></text>",
"      <text",
"         id=\"text4138\"",
"         y=\"60\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#000000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"60\"",
"           x=\"5\"",
"           id=\"tspan4140\"",
"           sodipodi:role=\"line\">key:ESCAPE</tspan></text>",
"      <text",
"         id=\"text4142\"",
"         y=\"75\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#ff0000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           id=\"tspan4144\"",
"           y=\"75\"",
"           x=\"5\"",
"           sodipodi:role=\"line\">key:&lt;Primary&gt;ESCAPE</tspan></text>",
"      <text",
"         id=\"text4146\"",
"         y=\"90\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#008000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"90\"",
"           x=\"5\"",
"           id=\"tspan4148\"",
"           sodipodi:role=\"line\">key:&lt;Shift&gt;ESCAPE</tspan></text>",
"      <text",
"         id=\"text4150\"",
"         y=\"105\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#0000ff;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"105\"",
"           x=\"5\"",
"           id=\"tspan4152\"",
"           sodipodi:role=\"line\">key:&lt;Alt&gt;ESCAPE</tspan></text>",
"      <text",
"         sodipodi:linespacing=\"100%\"",
"         id=\"text4154\"",
"         y=\"135\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#808000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"135\"",
"           x=\"5\"",
"           id=\"tspan4156\"",
"           sodipodi:role=\"line\">key:&lt;Primary&gt;&lt;Shift&gt;ESCAPE</tspan></text>",
"      <text",
"         sodipodi:linespacing=\"100%\"",
"         id=\"text4158\"",
"         y=\"150\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#ff00ff;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"150\"",
"           x=\"5\"",
"           id=\"tspan4160\"",
"           sodipodi:role=\"line\">key:&lt;Primary&gt;&lt;Alt&gt;ESCAPE</tspan></text>",
"      <text",
"         sodipodi:linespacing=\"100%\"",
"         id=\"text4162\"",
"         y=\"165\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#008080;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"165\"",
"           x=\"5\"",
"           id=\"tspan4164\"",
"           sodipodi:role=\"line\">key:&lt;Shift&gt;&lt;Alt&gt;ESCAPE</tspan></text>",
"    </g>",
"    <g",
"       transform=\"translate(1060,130)\"",
"       id=\"g4166\">",
"      <rect",
"         y=\"2.9921875\"",
"         x=\"0\"",
"         height=\"170\"",
"         width=\"170\"",
"         id=\"rect4168\"",
"         style=\"fill:#0000ff;fill-opacity:0.1;stroke:#000000;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1\" />",
"      <text",
"         sodipodi:linespacing=\"100%\"",
"         id=\"text4170\"",
"         y=\"42.992188\"",
"         x=\"40\"",
"         style=\"font-size:48px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#000000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"42.992188\"",
"           x=\"40\"",
"           id=\"tspan4172\"",
"           sodipodi:role=\"line\">F4</tspan></text>",
"      <text",
"         id=\"text4174\"",
"         y=\"60\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#000000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"60\"",
"           x=\"5\"",
"           id=\"tspan4176\"",
"           sodipodi:role=\"line\">key:F4</tspan></text>",
"      <text",
"         id=\"text4178\"",
"         y=\"75\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#ff0000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           id=\"tspan4180\"",
"           y=\"75\"",
"           x=\"5\"",
"           sodipodi:role=\"line\">key:&lt;Primary&gt;F4</tspan></text>",
"      <text",
"         id=\"text4182\"",
"         y=\"90\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#008000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"90\"",
"           x=\"5\"",
"           id=\"tspan4184\"",
"           sodipodi:role=\"line\">key:&lt;Shift&gt;F4</tspan></text>",
"      <text",
"         id=\"text4186\"",
"         y=\"105\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#0000ff;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"105\"",
"           x=\"5\"",
"           id=\"tspan4188\"",
"           sodipodi:role=\"line\">key:&lt;Alt&gt;F4</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-style:normal;font-weight:normal;font-size:12.00000048px;font-family:'Bitstream Vera Sans';fill:#464646;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;-inkscape-font-specification:'Bitstream Vera Sans, Normal';font-stretch:normal;font-variant:normal;text-anchor:start;text-align:start;writing-mode:lr;line-height:125%;\"",
"         x=\"5\"",
"         y=\"120\"",
"         id=\"text6415\"",
"         sodipodi:linespacing=\"125%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan6419\">key:&lt;Super&gt;F4</tspan></text>",
"      <text",
"         sodipodi:linespacing=\"100%\"",
"         id=\"text4190\"",
"         y=\"135\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#808000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"135\"",
"           x=\"5\"",
"           id=\"tspan4192\"",
"           sodipodi:role=\"line\">key:&lt;Primary&gt;&lt;Shift&gt;F4</tspan></text>",
"      <text",
"         sodipodi:linespacing=\"100%\"",
"         id=\"text4194\"",
"         y=\"150\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#ff00ff;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"150\"",
"           x=\"5\"",
"           id=\"tspan4196\"",
"           sodipodi:role=\"line\">key:&lt;Primary&gt;&lt;Alt&gt;F4</tspan></text>",
"      <text",
"         sodipodi:linespacing=\"100%\"",
"         id=\"text4198\"",
"         y=\"165\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#008080;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"165\"",
"           x=\"5\"",
"           id=\"tspan4200\"",
"           sodipodi:role=\"line\">key:&lt;Shift&gt;&lt;Alt&gt;F4</tspan></text>",
"    </g>",
"    <g",
"       id=\"g4202\"",
"       transform=\"translate(880,130)\">",
"      <rect",
"         style=\"fill:#0000ff;fill-opacity:0.1;stroke:#000000;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1\"",
"         id=\"rect4204\"",
"         width=\"170\"",
"         height=\"170\"",
"         x=\"0\"",
"         y=\"2.9921875\" />",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:48px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#000000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         x=\"40\"",
"         y=\"42.992188\"",
"         id=\"text4206\"",
"         sodipodi:linespacing=\"100%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan4208\"",
"           x=\"40\"",
"           y=\"42.992188\">F3</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#000000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"60\"",
"         id=\"text4210\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan4212\"",
"           x=\"5\"",
"           y=\"60\">key:F3</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#ff0000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"75\"",
"         id=\"text4214\"><tspan",
"           sodipodi:role=\"line\"",
"           x=\"5\"",
"           y=\"75\"",
"           id=\"tspan4216\">key:&lt;Primary&gt;F3</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#008000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"90\"",
"         id=\"text4218\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan4220\"",
"           x=\"5\"",
"           y=\"90\">key:&lt;Shift&gt;F3</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#0000ff;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"105\"",
"         id=\"text4222\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan4224\"",
"           x=\"5\"",
"           y=\"105\">key:&lt;Alt&gt;F3</tspan></text>",
"      <text",
"         id=\"text6409\"",
"         y=\"120\"",
"         x=\"5\"",
"         style=\"font-style:normal;font-weight:normal;font-size:12.00000048px;font-family:'Bitstream Vera Sans';fill:#464646;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;-inkscape-font-specification:'Bitstream Vera Sans, Normal';font-stretch:normal;font-variant:normal;text-anchor:start;text-align:start;writing-mode:lr;line-height:125%;\"",
"         xml:space=\"preserve\"",
"         sodipodi:linespacing=\"125%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan6413\">key:&lt;Super&gt;F3</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#808000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"135\"",
"         id=\"text4226\"",
"         sodipodi:linespacing=\"100%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan4228\"",
"           x=\"5\"",
"           y=\"135\">key:&lt;Primary&gt;&lt;Shift&gt;F3</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#ff00ff;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"150\"",
"         id=\"text4230\"",
"         sodipodi:linespacing=\"100%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan4232\"",
"           x=\"5\"",
"           y=\"150\">key:&lt;Primary&gt;&lt;Alt&gt;F3</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#008080;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"165\"",
"         id=\"text4234\"",
"         sodipodi:linespacing=\"100%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan4236\"",
"           x=\"5\"",
"           y=\"165\">key:&lt;Shift&gt;&lt;Alt&gt;F3</tspan></text>",
"    </g>",
"    <g",
"       transform=\"translate(700,130)\"",
"       id=\"g4238\">",
"      <rect",
"         y=\"2.9921875\"",
"         x=\"0\"",
"         height=\"170\"",
"         width=\"170\"",
"         id=\"rect4240\"",
"         style=\"fill:#0000ff;fill-opacity:0.1;stroke:#000000;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1\" />",
"      <text",
"         sodipodi:linespacing=\"100%\"",
"         id=\"text4242\"",
"         y=\"42.992188\"",
"         x=\"40\"",
"         style=\"font-size:48px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#000000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"42.992188\"",
"           x=\"40\"",
"           id=\"tspan4244\"",
"           sodipodi:role=\"line\">F2</tspan></text>",
"      <text",
"         id=\"text4246\"",
"         y=\"60\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#000000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"60\"",
"           x=\"5\"",
"           id=\"tspan4248\"",
"           sodipodi:role=\"line\">key:F2</tspan></text>",
"      <text",
"         id=\"text4250\"",
"         y=\"75\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#ff0000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           id=\"tspan4252\"",
"           y=\"75\"",
"           x=\"5\"",
"           sodipodi:role=\"line\">key:&lt;Primary&gt;F2</tspan></text>",
"      <text",
"         id=\"text4254\"",
"         y=\"90\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#008000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"90\"",
"           x=\"5\"",
"           id=\"tspan4256\"",
"           sodipodi:role=\"line\">key:&lt;Shift&gt;F2</tspan></text>",
"      <text",
"         id=\"text4258\"",
"         y=\"105\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#0000ff;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"105\"",
"           x=\"5\"",
"           id=\"tspan4260\"",
"           sodipodi:role=\"line\">key:&lt;Alt&gt;F2</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-style:normal;font-weight:normal;font-size:12.00000048px;font-family:'Bitstream Vera Sans';fill:#464646;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;-inkscape-font-specification:'Bitstream Vera Sans, Normal';font-stretch:normal;font-variant:normal;text-anchor:start;text-align:start;writing-mode:lr;line-height:125%;\"",
"         x=\"5\"",
"         y=\"120\"",
"         id=\"text6401\"",
"         sodipodi:linespacing=\"125%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan6405\">key:&lt;Super&gt;F2</tspan></text>",
"      <text",
"         sodipodi:linespacing=\"100%\"",
"         id=\"text4262\"",
"         y=\"135\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#808000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"135\"",
"           x=\"5\"",
"           id=\"tspan4264\"",
"           sodipodi:role=\"line\">key:&lt;Primary&gt;&lt;Shift&gt;F2</tspan></text>",
"      <text",
"         sodipodi:linespacing=\"100%\"",
"         id=\"text4266\"",
"         y=\"150\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#ff00ff;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"150\"",
"           x=\"5\"",
"           id=\"tspan4268\"",
"           sodipodi:role=\"line\">key:&lt;Primary&gt;&lt;Alt&gt;F2</tspan></text>",
"      <text",
"         sodipodi:linespacing=\"100%\"",
"         id=\"text4270\"",
"         y=\"165\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#008080;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"165\"",
"           x=\"5\"",
"           id=\"tspan4272\"",
"           sodipodi:role=\"line\">key:&lt;Shift&gt;&lt;Alt&gt;F2</tspan></text>",
"    </g>",
"    <g",
"       id=\"g4274\"",
"       transform=\"translate(520,130)\">",
"      <rect",
"         style=\"fill:#0000ff;fill-opacity:0.1;stroke:#000000;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1\"",
"         id=\"rect4276\"",
"         width=\"170\"",
"         height=\"170\"",
"         x=\"0\"",
"         y=\"2.9921875\" />",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:48px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#000000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         x=\"40\"",
"         y=\"42.992188\"",
"         id=\"text4278\"",
"         sodipodi:linespacing=\"100%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan4280\"",
"           x=\"40\"",
"           y=\"42.992188\">F1</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#000000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"60\"",
"         id=\"text4282\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan4284\"",
"           x=\"5\"",
"           y=\"60\">key:F1</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#ff0000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"75\"",
"         id=\"text4286\"><tspan",
"           sodipodi:role=\"line\"",
"           x=\"5\"",
"           y=\"75\"",
"           id=\"tspan4288\">key:&lt;Primary&gt;F1</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#008000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"90\"",
"         id=\"text4290\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan4292\"",
"           x=\"5\"",
"           y=\"90\">key:&lt;Shift&gt;F1</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#0000ff;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"105\"",
"         id=\"text4294\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan4296\"",
"           x=\"5\"",
"           y=\"105\">key:&lt;Alt&gt;F1</tspan></text>",
"      <text",
"         id=\"text6397\"",
"         y=\"120\"",
"         x=\"5\"",
"         style=\"font-style:normal;font-weight:normal;font-size:12.00000048px;font-family:'Bitstream Vera Sans';fill:#464646;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;-inkscape-font-specification:'Bitstream Vera Sans, Normal';font-stretch:normal;font-variant:normal;text-anchor:start;text-align:start;writing-mode:lr;line-height:125%;\"",
"         xml:space=\"preserve\"",
"         sodipodi:linespacing=\"125%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan6407\">key:&lt;Super&gt;F1</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#808000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"135\"",
"         id=\"text4298\"",
"         sodipodi:linespacing=\"100%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan4300\"",
"           x=\"5\"",
"           y=\"135\">key:&lt;Primary&gt;&lt;Shift&gt;F1</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#ff00ff;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"150\"",
"         id=\"text4302\"",
"         sodipodi:linespacing=\"100%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan4304\"",
"           x=\"5\"",
"           y=\"150\">key:&lt;Primary&gt;&lt;Alt&gt;F1</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#008080;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"165\"",
"         id=\"text4306\"",
"         sodipodi:linespacing=\"100%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan4308\"",
"           x=\"5\"",
"           y=\"165\">key:&lt;Shift&gt;&lt;Alt&gt;F1</tspan></text>",
"    </g>",
"    <g",
"       id=\"g4310\"",
"       transform=\"translate(1880,130)\">",
"      <rect",
"         style=\"fill:#0000ff;fill-opacity:0.1;stroke:#000000;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1\"",
"         id=\"rect4312\"",
"         width=\"170\"",
"         height=\"170\"",
"         x=\"0\"",
"         y=\"2.9921875\" />",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:48px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#000000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         x=\"40\"",
"         y=\"42.992188\"",
"         id=\"text4314\"",
"         sodipodi:linespacing=\"100%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan4316\"",
"           x=\"40\"",
"           y=\"42.992188\">F8</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#000000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"60\"",
"         id=\"text4318\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan4320\"",
"           x=\"5\"",
"           y=\"60\">key:F8</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#ff0000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"75\"",
"         id=\"text4322\"><tspan",
"           sodipodi:role=\"line\"",
"           x=\"5\"",
"           y=\"75\"",
"           id=\"tspan4324\">key:&lt;Primary&gt;F8</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#008000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"90\"",
"         id=\"text4326\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan4328\"",
"           x=\"5\"",
"           y=\"90\">key:&lt;Shift&gt;F8</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#0000ff;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"105\"",
"         id=\"text4330\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan4332\"",
"           x=\"5\"",
"           y=\"105\">key:&lt;Alt&gt;F8</tspan></text>",
"      <text",
"         id=\"text6439\"",
"         y=\"120\"",
"         x=\"5\"",
"         style=\"font-style:normal;font-weight:normal;font-size:12.00000048px;font-family:'Bitstream Vera Sans';fill:#464646;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;-inkscape-font-specification:'Bitstream Vera Sans, Normal';font-stretch:normal;font-variant:normal;text-anchor:start;text-align:start;writing-mode:lr;line-height:125%;\"",
"         xml:space=\"preserve\"",
"         sodipodi:linespacing=\"125%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan6443\">key:&lt;Super&gt;F8</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#808000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"135\"",
"         id=\"text4334\"",
"         sodipodi:linespacing=\"100%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan4336\"",
"           x=\"5\"",
"           y=\"135\">key:&lt;Primary&gt;&lt;Shift&gt;F8</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#ff00ff;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"150\"",
"         id=\"text4338\"",
"         sodipodi:linespacing=\"100%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan4340\"",
"           x=\"5\"",
"           y=\"150\">key:&lt;Primary&gt;&lt;Alt&gt;F8</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#008080;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"165\"",
"         id=\"text4342\"",
"         sodipodi:linespacing=\"100%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan4344\"",
"           x=\"5\"",
"           y=\"165\">key:&lt;Shift&gt;&lt;Alt&gt;F8</tspan></text>",
"    </g>",
"    <g",
"       transform=\"translate(1700,130)\"",
"       id=\"g4346\">",
"      <rect",
"         y=\"2.9921875\"",
"         x=\"0\"",
"         height=\"170\"",
"         width=\"170\"",
"         id=\"rect4348\"",
"         style=\"fill:#0000ff;fill-opacity:0.1;stroke:#000000;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1\" />",
"      <text",
"         sodipodi:linespacing=\"100%\"",
"         id=\"text4350\"",
"         y=\"42.992188\"",
"         x=\"40\"",
"         style=\"font-size:48px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#000000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"42.992188\"",
"           x=\"40\"",
"           id=\"tspan4352\"",
"           sodipodi:role=\"line\">F7</tspan></text>",
"      <text",
"         id=\"text4354\"",
"         y=\"60\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#000000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"60\"",
"           x=\"5\"",
"           id=\"tspan4356\"",
"           sodipodi:role=\"line\">key:F7</tspan></text>",
"      <text",
"         id=\"text4358\"",
"         y=\"75\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#ff0000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           id=\"tspan4360\"",
"           y=\"75\"",
"           x=\"5\"",
"           sodipodi:role=\"line\">key:&lt;Primary&gt;F7</tspan></text>",
"      <text",
"         id=\"text4362\"",
"         y=\"90\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#008000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"90\"",
"           x=\"5\"",
"           id=\"tspan4364\"",
"           sodipodi:role=\"line\">key:&lt;Shift&gt;F7</tspan></text>",
"      <text",
"         id=\"text4366\"",
"         y=\"105\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#0000ff;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"105\"",
"           x=\"5\"",
"           id=\"tspan4368\"",
"           sodipodi:role=\"line\">key:&lt;Alt&gt;F7</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-style:normal;font-weight:normal;font-size:12.00000048px;font-family:'Bitstream Vera Sans';fill:#464646;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;-inkscape-font-specification:'Bitstream Vera Sans, Normal';font-stretch:normal;font-variant:normal;text-anchor:start;text-align:start;writing-mode:lr;line-height:125%;\"",
"         x=\"5\"",
"         y=\"120\"",
"         id=\"text6433\"",
"         sodipodi:linespacing=\"125%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan6437\">key:&lt;Super&gt;F7</tspan></text>",
"      <text",
"         sodipodi:linespacing=\"100%\"",
"         id=\"text4370\"",
"         y=\"135\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#808000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"135\"",
"           x=\"5\"",
"           id=\"tspan4372\"",
"           sodipodi:role=\"line\">key:&lt;Primary&gt;&lt;Shift&gt;F7</tspan></text>",
"      <text",
"         sodipodi:linespacing=\"100%\"",
"         id=\"text4374\"",
"         y=\"150\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#ff00ff;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"150\"",
"           x=\"5\"",
"           id=\"tspan4376\"",
"           sodipodi:role=\"line\">key:&lt;Primary&gt;&lt;Alt&gt;F7</tspan></text>",
"      <text",
"         sodipodi:linespacing=\"100%\"",
"         id=\"text4378\"",
"         y=\"165\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#008080;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"165\"",
"           x=\"5\"",
"           id=\"tspan4380\"",
"           sodipodi:role=\"line\">key:&lt;Shift&gt;&lt;Alt&gt;F7</tspan></text>",
"    </g>",
"    <g",
"       id=\"g4382\"",
"       transform=\"translate(1520,130)\">",
"      <rect",
"         style=\"fill:#0000ff;fill-opacity:0.1;stroke:#000000;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1\"",
"         id=\"rect4384\"",
"         width=\"170\"",
"         height=\"170\"",
"         x=\"0\"",
"         y=\"2.9921875\" />",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:48px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#000000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         x=\"40\"",
"         y=\"42.992188\"",
"         id=\"text4386\"",
"         sodipodi:linespacing=\"100%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan4388\"",
"           x=\"40\"",
"           y=\"42.992188\">F6</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#000000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"60\"",
"         id=\"text4390\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan4392\"",
"           x=\"5\"",
"           y=\"60\">key:F6</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#ff0000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"75\"",
"         id=\"text4394\"><tspan",
"           sodipodi:role=\"line\"",
"           x=\"5\"",
"           y=\"75\"",
"           id=\"tspan4396\">key:&lt;Primary&gt;F6</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#008000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"90\"",
"         id=\"text4398\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan4400\"",
"           x=\"5\"",
"           y=\"90\">key:&lt;Shift&gt;F6</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#0000ff;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"105\"",
"         id=\"text4402\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan4404\"",
"           x=\"5\"",
"           y=\"105\">key:&lt;Alt&gt;F6</tspan></text>",
"      <text",
"         id=\"text6427\"",
"         y=\"120\"",
"         x=\"5\"",
"         style=\"font-style:normal;font-weight:normal;font-size:12.00000048px;font-family:'Bitstream Vera Sans';fill:#464646;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;-inkscape-font-specification:'Bitstream Vera Sans, Normal';font-stretch:normal;font-variant:normal;text-anchor:start;text-align:start;writing-mode:lr;line-height:125%;\"",
"         xml:space=\"preserve\"",
"         sodipodi:linespacing=\"125%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan6431\">key:&lt;Super&gt;F6</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#808000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"135\"",
"         id=\"text4406\"",
"         sodipodi:linespacing=\"100%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan4408\"",
"           x=\"5\"",
"           y=\"135\">key:&lt;Primary&gt;&lt;Shift&gt;F6</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#ff00ff;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"150\"",
"         id=\"text4410\"",
"         sodipodi:linespacing=\"100%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan4412\"",
"           x=\"5\"",
"           y=\"150\">key:&lt;Primary&gt;&lt;Alt&gt;F6</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#008080;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"165\"",
"         id=\"text4414\"",
"         sodipodi:linespacing=\"100%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan4416\"",
"           x=\"5\"",
"           y=\"165\">key:&lt;Shift&gt;&lt;Alt&gt;F6</tspan></text>",
"    </g>",
"    <g",
"       transform=\"translate(1340,130)\"",
"       id=\"g4418\">",
"      <rect",
"         y=\"2.9921875\"",
"         x=\"0\"",
"         height=\"170\"",
"         width=\"170\"",
"         id=\"rect4420\"",
"         style=\"fill:#0000ff;fill-opacity:0.1;stroke:#000000;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1\" />",
"      <text",
"         sodipodi:linespacing=\"100%\"",
"         id=\"text4422\"",
"         y=\"42.992188\"",
"         x=\"40\"",
"         style=\"font-size:48px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#000000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"42.992188\"",
"           x=\"40\"",
"           id=\"tspan4424\"",
"           sodipodi:role=\"line\">F5</tspan></text>",
"      <text",
"         id=\"text4426\"",
"         y=\"60\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#000000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"60\"",
"           x=\"5\"",
"           id=\"tspan4428\"",
"           sodipodi:role=\"line\">key:F5</tspan></text>",
"      <text",
"         id=\"text4430\"",
"         y=\"75\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#ff0000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           id=\"tspan4432\"",
"           y=\"75\"",
"           x=\"5\"",
"           sodipodi:role=\"line\">key:&lt;Primary&gt;F5</tspan></text>",
"      <text",
"         id=\"text4434\"",
"         y=\"90\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#008000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"90\"",
"           x=\"5\"",
"           id=\"tspan4436\"",
"           sodipodi:role=\"line\">key:&lt;Shift&gt;F5</tspan></text>",
"      <text",
"         id=\"text4438\"",
"         y=\"105\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#0000ff;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"105\"",
"           x=\"5\"",
"           id=\"tspan4440\"",
"           sodipodi:role=\"line\">key:&lt;Alt&gt;F5</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-style:normal;font-weight:normal;font-size:12.00000048px;font-family:'Bitstream Vera Sans';fill:#464646;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;-inkscape-font-specification:'Bitstream Vera Sans, Normal';font-stretch:normal;font-variant:normal;text-anchor:start;text-align:start;writing-mode:lr;line-height:125%;\"",
"         x=\"5\"",
"         y=\"120\"",
"         id=\"text6421\"",
"         sodipodi:linespacing=\"125%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan6425\">key:&lt;Super&gt;F5</tspan></text>",
"      <text",
"         sodipodi:linespacing=\"100%\"",
"         id=\"text4442\"",
"         y=\"135\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#808000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"135\"",
"           x=\"5\"",
"           id=\"tspan4444\"",
"           sodipodi:role=\"line\">key:&lt;Primary&gt;&lt;Shift&gt;F5</tspan></text>",
"      <text",
"         sodipodi:linespacing=\"100%\"",
"         id=\"text4446\"",
"         y=\"150\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#ff00ff;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"150\"",
"           x=\"5\"",
"           id=\"tspan4448\"",
"           sodipodi:role=\"line\">key:&lt;Primary&gt;&lt;Alt&gt;F5</tspan></text>",
"      <text",
"         sodipodi:linespacing=\"100%\"",
"         id=\"text4450\"",
"         y=\"165\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#008080;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"165\"",
"           x=\"5\"",
"           id=\"tspan4452\"",
"           sodipodi:role=\"line\">key:&lt;Shift&gt;&lt;Alt&gt;F5</tspan></text>",
"    </g>",
"    <g",
"       transform=\"translate(2140,130)\"",
"       id=\"g4454\">",
"      <rect",
"         y=\"2.9921875\"",
"         x=\"0\"",
"         height=\"170\"",
"         width=\"170\"",
"         id=\"rect4456\"",
"         style=\"fill:#0000ff;fill-opacity:0.1;stroke:#000000;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1\" />",
"      <text",
"         sodipodi:linespacing=\"100%\"",
"         id=\"text4458\"",
"         y=\"42.992188\"",
"         x=\"40\"",
"         style=\"font-size:48px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#000000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"42.992188\"",
"           x=\"40\"",
"           id=\"tspan4460\"",
"           sodipodi:role=\"line\">F9</tspan></text>",
"      <text",
"         id=\"text4462\"",
"         y=\"60\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#000000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"60\"",
"           x=\"5\"",
"           id=\"tspan4464\"",
"           sodipodi:role=\"line\">key:F9</tspan></text>",
"      <text",
"         id=\"text4466\"",
"         y=\"75\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#ff0000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           id=\"tspan4468\"",
"           y=\"75\"",
"           x=\"5\"",
"           sodipodi:role=\"line\">key:&lt;Primary&gt;F9</tspan></text>",
"      <text",
"         id=\"text4470\"",
"         y=\"90\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#008000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"90\"",
"           x=\"5\"",
"           id=\"tspan4472\"",
"           sodipodi:role=\"line\">key:&lt;Shift&gt;F9</tspan></text>",
"      <text",
"         id=\"text4474\"",
"         y=\"105\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#0000ff;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"105\"",
"           x=\"5\"",
"           id=\"tspan4476\"",
"           sodipodi:role=\"line\">key:&lt;Alt&gt;F9</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-style:normal;font-weight:normal;font-size:12.00000048px;font-family:'Bitstream Vera Sans';fill:#464646;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;-inkscape-font-specification:'Bitstream Vera Sans, Normal';font-stretch:normal;font-variant:normal;text-anchor:start;text-align:start;writing-mode:lr;line-height:125%;\"",
"         x=\"5\"",
"         y=\"120\"",
"         id=\"text6445\"",
"         sodipodi:linespacing=\"125%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan6449\">key:&lt;Super&gt;F9</tspan></text>",
"      <text",
"         sodipodi:linespacing=\"100%\"",
"         id=\"text4478\"",
"         y=\"135\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#808000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"135\"",
"           x=\"5\"",
"           id=\"tspan4480\"",
"           sodipodi:role=\"line\">key:&lt;Primary&gt;&lt;Shift&gt;F9</tspan></text>",
"      <text",
"         sodipodi:linespacing=\"100%\"",
"         id=\"text4482\"",
"         y=\"150\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#ff00ff;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"150\"",
"           x=\"5\"",
"           id=\"tspan4484\"",
"           sodipodi:role=\"line\">key:&lt;Primary&gt;&lt;Alt&gt;F9</tspan></text>",
"      <text",
"         sodipodi:linespacing=\"100%\"",
"         id=\"text4486\"",
"         y=\"165\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#008080;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"165\"",
"           x=\"5\"",
"           id=\"tspan4488\"",
"           sodipodi:role=\"line\">key:&lt;Shift&gt;&lt;Alt&gt;F9</tspan></text>",
"    </g>",
"    <g",
"       id=\"g4490\"",
"       transform=\"translate(2320,130)\">",
"      <rect",
"         style=\"fill:#0000ff;fill-opacity:0.1;stroke:#000000;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1\"",
"         id=\"rect4492\"",
"         width=\"170\"",
"         height=\"170\"",
"         x=\"0\"",
"         y=\"2.9921875\" />",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:48px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#000000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         x=\"40\"",
"         y=\"42.992188\"",
"         id=\"text4494\"",
"         sodipodi:linespacing=\"100%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan4496\"",
"           x=\"40\"",
"           y=\"42.992188\">F10</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#000000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"60\"",
"         id=\"text4498\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan4500\"",
"           x=\"5\"",
"           y=\"60\">key:F10</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#ff0000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"75\"",
"         id=\"text4502\"><tspan",
"           sodipodi:role=\"line\"",
"           x=\"5\"",
"           y=\"75\"",
"           id=\"tspan4504\">key:&lt;Primary&gt;F10</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#008000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"90\"",
"         id=\"text4506\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan4508\"",
"           x=\"5\"",
"           y=\"90\">key:&lt;Shift&gt;F10</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#0000ff;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"105\"",
"         id=\"text4510\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan4512\"",
"           x=\"5\"",
"           y=\"105\">key:&lt;Alt&gt;F10</tspan></text>",
"      <text",
"         id=\"text6451\"",
"         y=\"120\"",
"         x=\"5\"",
"         style=\"font-style:normal;font-weight:normal;font-size:12.00000048px;font-family:'Bitstream Vera Sans';fill:#464646;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;-inkscape-font-specification:'Bitstream Vera Sans, Normal';font-stretch:normal;font-variant:normal;text-anchor:start;text-align:start;writing-mode:lr;line-height:125%;\"",
"         xml:space=\"preserve\"",
"         sodipodi:linespacing=\"125%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan6455\">key:&lt;Super&gt;F10</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#808000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"135\"",
"         id=\"text4514\"",
"         sodipodi:linespacing=\"100%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan4516\"",
"           x=\"5\"",
"           y=\"135\">key:&lt;Primary&gt;&lt;Shift&gt;F10</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#ff00ff;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"150\"",
"         id=\"text4518\"",
"         sodipodi:linespacing=\"100%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan4520\"",
"           x=\"5\"",
"           y=\"150\">key:&lt;Primary&gt;&lt;Alt&gt;F10</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#008080;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"165\"",
"         id=\"text4522\"",
"         sodipodi:linespacing=\"100%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan4524\"",
"           x=\"5\"",
"           y=\"165\">key:&lt;Shift&gt;&lt;Alt&gt;F10</tspan></text>",
"    </g>",
"    <g",
"       transform=\"translate(2500,130)\"",
"       id=\"g4526\">",
"      <rect",
"         y=\"2.9921875\"",
"         x=\"0\"",
"         height=\"170\"",
"         width=\"170\"",
"         id=\"rect4528\"",
"         style=\"fill:#0000ff;fill-opacity:0.1;stroke:#000000;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1\" />",
"      <text",
"         sodipodi:linespacing=\"100%\"",
"         id=\"text4530\"",
"         y=\"42.992188\"",
"         x=\"40\"",
"         style=\"font-size:48px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#000000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"42.992188\"",
"           x=\"40\"",
"           id=\"tspan4532\"",
"           sodipodi:role=\"line\">F11</tspan></text>",
"      <text",
"         id=\"text4534\"",
"         y=\"60\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#000000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"60\"",
"           x=\"5\"",
"           id=\"tspan4536\"",
"           sodipodi:role=\"line\">key:F11</tspan></text>",
"      <text",
"         id=\"text4538\"",
"         y=\"75\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#ff0000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           id=\"tspan4540\"",
"           y=\"75\"",
"           x=\"5\"",
"           sodipodi:role=\"line\">key:&lt;Primary&gt;F11</tspan></text>",
"      <text",
"         id=\"text4542\"",
"         y=\"90\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#008000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"90\"",
"           x=\"5\"",
"           id=\"tspan4544\"",
"           sodipodi:role=\"line\">key:&lt;Shift&gt;F11</tspan></text>",
"      <text",
"         id=\"text4546\"",
"         y=\"105\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#0000ff;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"105\"",
"           x=\"5\"",
"           id=\"tspan4548\"",
"           sodipodi:role=\"line\">key:&lt;Alt&gt;F11</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-style:normal;font-weight:normal;font-size:12.00000048px;font-family:'Bitstream Vera Sans';fill:#464646;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;-inkscape-font-specification:'Bitstream Vera Sans, Normal';font-stretch:normal;font-variant:normal;text-anchor:start;text-align:start;writing-mode:lr;line-height:125%;\"",
"         x=\"5\"",
"         y=\"120\"",
"         id=\"text6457\"",
"         sodipodi:linespacing=\"125%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan6461\">key:&lt;Super&gt;F11</tspan></text>",
"      <text",
"         sodipodi:linespacing=\"100%\"",
"         id=\"text4550\"",
"         y=\"135\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#808000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"135\"",
"           x=\"5\"",
"           id=\"tspan4552\"",
"           sodipodi:role=\"line\">key:&lt;Primary&gt;&lt;Shift&gt;F11</tspan></text>",
"      <text",
"         sodipodi:linespacing=\"100%\"",
"         id=\"text4554\"",
"         y=\"150\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#ff00ff;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"150\"",
"           x=\"5\"",
"           id=\"tspan4556\"",
"           sodipodi:role=\"line\">key:&lt;Primary&gt;&lt;Alt&gt;F11</tspan></text>",
"      <text",
"         sodipodi:linespacing=\"100%\"",
"         id=\"text4558\"",
"         y=\"165\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#008080;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"165\"",
"           x=\"5\"",
"           id=\"tspan4560\"",
"           sodipodi:role=\"line\">key:&lt;Shift&gt;&lt;Alt&gt;F11</tspan></text>",
"    </g>",
"    <g",
"       id=\"g4562\"",
"       transform=\"translate(2680,130)\">",
"      <rect",
"         style=\"fill:#0000ff;fill-opacity:0.1;stroke:#000000;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1\"",
"         id=\"rect4564\"",
"         width=\"170\"",
"         height=\"170\"",
"         x=\"0\"",
"         y=\"2.9921875\" />",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:48px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#000000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         x=\"40\"",
"         y=\"42.992188\"",
"         id=\"text4566\"",
"         sodipodi:linespacing=\"100%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan4568\"",
"           x=\"40\"",
"           y=\"42.992188\">F12</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#000000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"60\"",
"         id=\"text4570\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan4572\"",
"           x=\"5\"",
"           y=\"60\">key:F12</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#ff0000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"75\"",
"         id=\"text4574\"><tspan",
"           sodipodi:role=\"line\"",
"           x=\"5\"",
"           y=\"75\"",
"           id=\"tspan4576\">key:&lt;Primary&gt;F12</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#008000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"90\"",
"         id=\"text4578\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan4580\"",
"           x=\"5\"",
"           y=\"90\">key:&lt;Shift&gt;F12</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#0000ff;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"105\"",
"         id=\"text4582\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan4584\"",
"           x=\"5\"",
"           y=\"105\">key:&lt;Alt&gt;F12</tspan></text>",
"      <text",
"         id=\"text6463\"",
"         y=\"120\"",
"         x=\"5\"",
"         style=\"font-style:normal;font-weight:normal;font-size:12.00000048px;font-family:'Bitstream Vera Sans';fill:#4646ff;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;-inkscape-font-specification:'Bitstream Vera Sans, Normal';font-stretch:normal;font-variant:normal;text-anchor:start;text-align:start;writing-mode:lr;line-height:125%;\"",
"         xml:space=\"preserve\"",
"         sodipodi:linespacing=\"125%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan6467\">key:&lt;Super&gt;F12</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#808000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"135\"",
"         id=\"text4586\"",
"         sodipodi:linespacing=\"100%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan4588\"",
"           x=\"5\"",
"           y=\"135\">key:&lt;Primary&gt;&lt;Shift&gt;F12</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#ff00ff;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"150\"",
"         id=\"text4590\"",
"         sodipodi:linespacing=\"100%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan4592\"",
"           x=\"5\"",
"           y=\"150\">key:&lt;Primary&gt;&lt;Alt&gt;F12</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#008080;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"165\"",
"         id=\"text4594\"",
"         sodipodi:linespacing=\"100%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan4596\"",
"           x=\"5\"",
"           y=\"165\">key:&lt;Shift&gt;&lt;Alt&gt;F12</tspan></text>",
"    </g>",
"    <g",
"       id=\"g4598\"",
"       transform=\"translate(160,530)\">",
"      <rect",
"         y=\"2.9921875\"",
"         x=\"0\"",
"         height=\"170\"",
"         width=\"280\"",
"         id=\"rect4600\"",
"         style=\"fill:#0000ff;fill-opacity:0.1;stroke:#000000;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1\" />",
"      <text",
"         sodipodi:linespacing=\"100%\"",
"         id=\"text4602\"",
"         y=\"42.992188\"",
"         x=\"40\"",
"         style=\"font-size:48px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#000000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"42.992188\"",
"           x=\"40\"",
"           id=\"tspan4604\"",
"           sodipodi:role=\"line\">TAB</tspan></text>",
"      <text",
"         id=\"text4606\"",
"         y=\"60\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#000000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"60\"",
"           x=\"5\"",
"           id=\"tspan4608\"",
"           sodipodi:role=\"line\">key:TAB</tspan></text>",
"      <text",
"         id=\"text4610\"",
"         y=\"75\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#ff0000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           id=\"tspan4612\"",
"           y=\"75\"",
"           x=\"5\"",
"           sodipodi:role=\"line\">key:&lt;Primary&gt;TAB</tspan></text>",
"      <text",
"         id=\"text4614\"",
"         y=\"90\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#008000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"90\"",
"           x=\"5\"",
"           id=\"tspan4616\"",
"           sodipodi:role=\"line\">key:&lt;Shift&gt;TAB</tspan></text>",
"      <text",
"         id=\"text4618\"",
"         y=\"105\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#0000ff;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"105\"",
"           x=\"5\"",
"           id=\"tspan4620\"",
"           sodipodi:role=\"line\">key:&lt;Alt&gt;TAB</tspan></text>",
"      <text",
"         sodipodi:linespacing=\"100%\"",
"         id=\"text4622\"",
"         y=\"135\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#808000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"135\"",
"           x=\"5\"",
"           id=\"tspan4624\"",
"           sodipodi:role=\"line\">key:&lt;Primary&gt;&lt;Shift&gt;TAB</tspan></text>",
"      <text",
"         sodipodi:linespacing=\"100%\"",
"         id=\"text4626\"",
"         y=\"150\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#ff00ff;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"150\"",
"           x=\"5\"",
"           id=\"tspan4628\"",
"           sodipodi:role=\"line\">key:&lt;Primary&gt;&lt;Alt&gt;TAB</tspan></text>",
"      <text",
"         sodipodi:linespacing=\"100%\"",
"         id=\"text4630\"",
"         y=\"165\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#008080;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"165\"",
"           x=\"5\"",
"           id=\"tspan4632\"",
"           sodipodi:role=\"line\">key:&lt;Shift&gt;&lt;Alt&gt;TAB</tspan></text>",
"    </g>",
"    <g",
"       id=\"g4670\"",
"       transform=\"translate(160,710)\">",
"      <rect",
"         y=\"2.9921875\"",
"         x=\"0\"",
"         height=\"170\"",
"         width=\"330\"",
"         id=\"rect4672\"",
"         style=\"fill:#0000ff;fill-opacity:0.1;stroke:#000000;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1\" />",
"      <text",
"         sodipodi:linespacing=\"100%\"",
"         id=\"text4674\"",
"         y=\"42.992188\"",
"         x=\"40\"",
"         style=\"font-size:48px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#000000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"42.992188\"",
"           x=\"40\"",
"           id=\"tspan4676\"",
"           sodipodi:role=\"line\">Caps Lock</tspan></text>",
"      <text",
"         id=\"text4678\"",
"         y=\"60\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#000000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"60\"",
"           x=\"5\"",
"           id=\"tspan4680\"",
"           sodipodi:role=\"line\">key:Caps Lock</tspan></text>",
"      <text",
"         id=\"text4682\"",
"         y=\"75\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#ff0000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           id=\"tspan4684\"",
"           y=\"75\"",
"           x=\"5\"",
"           sodipodi:role=\"line\">key:&lt;Primary&gt;Caps Lock</tspan></text>",
"      <text",
"         id=\"text4686\"",
"         y=\"90\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#008000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"90\"",
"           x=\"5\"",
"           id=\"tspan4688\"",
"           sodipodi:role=\"line\">key:&lt;Shift&gt;Caps Lock</tspan></text>",
"      <text",
"         id=\"text4690\"",
"         y=\"105\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#0000ff;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"105\"",
"           x=\"5\"",
"           id=\"tspan4692\"",
"           sodipodi:role=\"line\">key:&lt;Alt&gt;Caps Lock</tspan></text>",
"      <text",
"         sodipodi:linespacing=\"100%\"",
"         id=\"text4694\"",
"         y=\"135\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#808000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"135\"",
"           x=\"5\"",
"           id=\"tspan4696\"",
"           sodipodi:role=\"line\">key:&lt;Primary&gt;&lt;Shift&gt;Caps Lock</tspan></text>",
"      <text",
"         sodipodi:linespacing=\"100%\"",
"         id=\"text4698\"",
"         y=\"150\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#ff00ff;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"150\"",
"           x=\"5\"",
"           id=\"tspan4700\"",
"           sodipodi:role=\"line\">key:&lt;Primary&gt;&lt;Alt&gt;Caps Lock</tspan></text>",
"      <text",
"         sodipodi:linespacing=\"100%\"",
"         id=\"text4702\"",
"         y=\"165\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#008080;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"165\"",
"           x=\"5\"",
"           id=\"tspan4704\"",
"           sodipodi:role=\"line\">key:&lt;Shift&gt;&lt;Alt&gt;Caps Lock</tspan></text>",
"    </g>",
"    <g",
"       transform=\"matrix(0.55182768,0,0,1.0018861,159.83047,889.83404)\"",
"       id=\"g4706\">",
"      <rect",
"         style=\"fill:#0000ff;fill-opacity:0.1;stroke:#000000;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1\"",
"         id=\"rect4708\"",
"         width=\"420\"",
"         height=\"170\"",
"         x=\"0\"",
"         y=\"2.9921875\" />",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;font-size:62.31797791px;line-height:100%;font-family:'Bitstream Vera Sans';-inkscape-font-specification:'Bitstream Vera Sans';text-align:start;writing-mode:lr-tb;text-anchor:start;fill:#008000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1\"",
"         x=\"29.138952\"",
"         y=\"55.81638\"",
"         id=\"text4710\"",
"         sodipodi:linespacing=\"100%\"",
"         transform=\"scale(1.2982912,0.77024322)\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan4712\"",
"           x=\"29.138952\"",
"           y=\"55.81638\"",
"           style=\"font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;font-size:62.31797791px;line-height:100%;font-family:'Bitstream Vera Sans';-inkscape-font-specification:'Bitstream Vera Sans';text-align:start;writing-mode:lr-tb;text-anchor:start;fill:#008000\">Shift</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-style:normal;font-weight:normal;font-size:16.04530907px;font-family:'Bitstream Vera Sans';fill:#000000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1\"",
"         x=\"3.0972457\"",
"         y=\"80.22654\"",
"         id=\"text4714\"",
"         transform=\"scale(1.337109,0.74788218)\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan4716\"",
"           x=\"3.0972457\"",
"           y=\"80.22654\">key:Shift</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-style:normal;font-weight:normal;font-size:15.77950764px;font-family:'Bitstream Vera Sans';fill:#ff0000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1\"",
"         x=\"3.1981044\"",
"         y=\"98.621925\"",
"         id=\"text4718\"",
"         transform=\"scale(1.314959,0.76047999)\"><tspan",
"           sodipodi:role=\"line\"",
"           x=\"3.1981044\"",
"           y=\"98.621925\"",
"           id=\"tspan4720\">key:&lt;Primary&gt;Shift</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-style:normal;font-weight:normal;font-size:15.77000427px;font-family:'Bitstream Vera Sans';fill:#008000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1\"",
"         x=\"3.2017586\"",
"         y=\"118.27503\"",
"         id=\"text4722\"",
"         transform=\"scale(1.314167,0.76093829)\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan4724\"",
"           x=\"3.2017586\"",
"           y=\"118.27503\">key:&lt;Shift&gt;Shift</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-style:normal;font-weight:normal;font-size:15.68608284px;font-family:'Bitstream Vera Sans';fill:#0000ff;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1\"",
"         x=\"3.2341726\"",
"         y=\"137.25322\"",
"         id=\"text4726\"",
"         transform=\"scale(1.3071735,0.76500937)\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan4728\"",
"           x=\"3.2341726\"",
"           y=\"137.25322\">key:&lt;Alt&gt;Shift</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;font-size:15.11343098px;line-height:100%;font-family:'Bitstream Vera Sans';-inkscape-font-specification:'Bitstream Vera Sans';text-align:start;writing-mode:lr-tb;text-anchor:start;fill:#808000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1\"",
"         x=\"3.4627037\"",
"         y=\"151.13431\"",
"         id=\"text4730\"",
"         sodipodi:linespacing=\"100%\"",
"         transform=\"scale(1.2594525,0.79399578)\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan4732\"",
"           x=\"3.4627037\"",
"           y=\"151.13431\">key:&lt;Primary&gt;&lt;Shift&gt;Shift</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;font-size:15.32491875px;line-height:100%;font-family:'Bitstream Vera Sans';-inkscape-font-specification:'Bitstream Vera Sans';text-align:start;writing-mode:lr-tb;text-anchor:start;fill:#ff00ff;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1\"",
"         x=\"3.3767674\"",
"         y=\"172.40533\"",
"         id=\"text4734\"",
"         sodipodi:linespacing=\"100%\"",
"         transform=\"scale(1.2770766,0.78303839)\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan4736\"",
"           x=\"3.3767674\"",
"           y=\"172.40533\">key:&lt;Primary&gt;&lt;Alt&gt;Shift</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;font-size:15.51035213px;line-height:100%;font-family:'Bitstream Vera Sans';-inkscape-font-specification:'Bitstream Vera Sans';text-align:start;writing-mode:lr-tb;text-anchor:start;fill:#008080;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1\"",
"         x=\"3.3029158\"",
"         y=\"193.87941\"",
"         id=\"text4738\"",
"         sodipodi:linespacing=\"100%\"",
"         transform=\"scale(1.2925294,0.77367682)\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan4740\"",
"           x=\"3.3029158\"",
"           y=\"193.87941\">key:&lt;Shift&gt;&lt;Alt&gt;Shift</tspan></text>",
"    </g>",
"    <g",
"       transform=\"translate(740,1070)\"",
"       id=\"g4742\">",
"      <rect",
"         style=\"fill:#0000ff;fill-opacity:0.1;stroke:#000000;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1\"",
"         id=\"rect4744\"",
"         width=\"280\"",
"         height=\"170\"",
"         x=\"0\"",
"         y=\"2.9921875\" />",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:48px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#0000ff;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         x=\"40\"",
"         y=\"42.992188\"",
"         id=\"text4746\"",
"         sodipodi:linespacing=\"100%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan4748\"",
"           x=\"40\"",
"           y=\"42.992188\"",
"           style=\"font-size:48px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#0000ff;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\">Alt</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#000000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"60\"",
"         id=\"text4750\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan4752\"",
"           x=\"5\"",
"           y=\"60\">key:Alt</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#ff0000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"75\"",
"         id=\"text4754\"><tspan",
"           sodipodi:role=\"line\"",
"           x=\"5\"",
"           y=\"75\"",
"           id=\"tspan4756\">key:&lt;Primary&gt;Alt</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#008000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"90\"",
"         id=\"text4758\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan4760\"",
"           x=\"5\"",
"           y=\"90\">key:&lt;Shift&gt;Alt</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#0000ff;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"105\"",
"         id=\"text4762\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan4764\"",
"           x=\"5\"",
"           y=\"105\">key:&lt;Alt&gt;Alt</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#808000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"135\"",
"         id=\"text4766\"",
"         sodipodi:linespacing=\"100%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan4768\"",
"           x=\"5\"",
"           y=\"135\">key:&lt;Primary&gt;&lt;Shift&gt;Alt</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#ff00ff;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"150\"",
"         id=\"text4770\"",
"         sodipodi:linespacing=\"100%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan4772\"",
"           x=\"5\"",
"           y=\"150\">key:&lt;Primary&gt;&lt;Alt&gt;Alt</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#008080;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"165\"",
"         id=\"text4774\"",
"         sodipodi:linespacing=\"100%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan4776\"",
"           x=\"5\"",
"           y=\"165\">key:&lt;Shift&gt;&lt;Alt&gt;Alt</tspan></text>",
"    </g>",
"    <g",
"       id=\"g4778\"",
"       transform=\"translate(450,1070)\">",
"      <rect",
"         y=\"2.9921875\"",
"         x=\"0\"",
"         height=\"170\"",
"         width=\"280\"",
"         id=\"rect4780\"",
"         style=\"fill:#0000ff;fill-opacity:0.1;stroke:#000000;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1\" />",
"      <text",
"         sodipodi:linespacing=\"100%\"",
"         id=\"text4782\"",
"         y=\"42.992188\"",
"         x=\"40\"",
"         style=\"font-size:48px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#464646;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans;\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"42.992188\"",
"           x=\"40\"",
"           id=\"tspan4784\"",
"           sodipodi:role=\"line\">Win</tspan></text>",
"      <text",
"         id=\"text4786\"",
"         y=\"60\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#000000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"60\"",
"           x=\"5\"",
"           id=\"tspan4788\"",
"           sodipodi:role=\"line\">key:Win</tspan></text>",
"      <text",
"         id=\"text4790\"",
"         y=\"75\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#ff0000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           id=\"tspan4792\"",
"           y=\"75\"",
"           x=\"5\"",
"           sodipodi:role=\"line\">key:&lt;Primary&gt;Win</tspan></text>",
"      <text",
"         id=\"text4794\"",
"         y=\"90\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#008000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"90\"",
"           x=\"5\"",
"           id=\"tspan4796\"",
"           sodipodi:role=\"line\">key:&lt;Shift&gt;Win</tspan></text>",
"      <text",
"         id=\"text4798\"",
"         y=\"105\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#0000ff;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"105\"",
"           x=\"5\"",
"           id=\"tspan4800\"",
"           sodipodi:role=\"line\">key:&lt;Alt&gt;Win</tspan></text>",
"      <text",
"         sodipodi:linespacing=\"100%\"",
"         id=\"text4802\"",
"         y=\"135\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#808000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"135\"",
"           x=\"5\"",
"           id=\"tspan4804\"",
"           sodipodi:role=\"line\">key:&lt;Primary&gt;&lt;Shift&gt;Win</tspan></text>",
"      <text",
"         sodipodi:linespacing=\"100%\"",
"         id=\"text4806\"",
"         y=\"150\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#ff00ff;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"150\"",
"           x=\"5\"",
"           id=\"tspan4808\"",
"           sodipodi:role=\"line\">key:&lt;Primary&gt;&lt;Alt&gt;Win</tspan></text>",
"      <text",
"         sodipodi:linespacing=\"100%\"",
"         id=\"text4810\"",
"         y=\"165\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#008080;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"165\"",
"           x=\"5\"",
"           id=\"tspan4812\"",
"           sodipodi:role=\"line\">key:&lt;Shift&gt;&lt;Alt&gt;Win</tspan></text>",
"    </g>",
"    <g",
"       transform=\"translate(160,1070)\"",
"       id=\"g4814\">",
"      <rect",
"         style=\"fill:#0000ff;fill-opacity:0.1;stroke:#000000;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1\"",
"         id=\"rect4816\"",
"         width=\"280\"",
"         height=\"170\"",
"         x=\"0\"",
"         y=\"2.9921875\" />",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:48px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#ff0000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         x=\"40\"",
"         y=\"42.992188\"",
"         id=\"text4818\"",
"         sodipodi:linespacing=\"100%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan4820\"",
"           x=\"40\"",
"           y=\"42.992188\"",
"           style=\"font-size:48px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#ff0000;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\">Ctrl</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#000000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"60\"",
"         id=\"text4822\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan4824\"",
"           x=\"5\"",
"           y=\"60\">key:Ctrl</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#ff0000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"75\"",
"         id=\"text4826\"><tspan",
"           sodipodi:role=\"line\"",
"           x=\"5\"",
"           y=\"75\"",
"           id=\"tspan4828\">key:&lt;Primary&gt;Ctrl</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#008000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"90\"",
"         id=\"text4830\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan4832\"",
"           x=\"5\"",
"           y=\"90\">key:&lt;Shift&gt;Ctrl</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#0000ff;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"105\"",
"         id=\"text4834\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan4836\"",
"           x=\"5\"",
"           y=\"105\">key:&lt;Alt&gt;Ctrl</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#808000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"135\"",
"         id=\"text4838\"",
"         sodipodi:linespacing=\"100%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan4840\"",
"           x=\"5\"",
"           y=\"135\">key:&lt;Primary&gt;&lt;Shift&gt;Ctrl</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#ff00ff;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"150\"",
"         id=\"text4842\"",
"         sodipodi:linespacing=\"100%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan4844\"",
"           x=\"5\"",
"           y=\"150\">key:&lt;Primary&gt;&lt;Alt&gt;Ctrl</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#008080;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"165\"",
"         id=\"text4846\"",
"         sodipodi:linespacing=\"100%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan4848\"",
"           x=\"5\"",
"           y=\"165\">key:&lt;Shift&gt;&lt;Alt&gt;Ctrl</tspan></text>",
"    </g>",
"    <g",
"       id=\"g4850\"",
"       transform=\"translate(2390,890)\">",
"      <rect",
"         y=\"2.9921875\"",
"         x=\"0\"",
"         height=\"170\"",
"         width=\"460\"",
"         id=\"rect4852\"",
"         style=\"fill:#0000ff;fill-opacity:0.1;stroke:#000000;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1\"",
"         ry=\"0\" />",
"      <text",
"         sodipodi:linespacing=\"100%\"",
"         id=\"text4854\"",
"         y=\"42.992188\"",
"         x=\"40\"",
"         style=\"font-size:48px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#008000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           style=\"font-size:48px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#008000;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"           y=\"42.992188\"",
"           x=\"40\"",
"           id=\"tspan4856\"",
"           sodipodi:role=\"line\">Shift</tspan></text>",
"      <text",
"         id=\"text4858\"",
"         y=\"60\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#000000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"60\"",
"           x=\"5\"",
"           id=\"tspan4860\"",
"           sodipodi:role=\"line\">key:Shift</tspan></text>",
"      <text",
"         id=\"text4862\"",
"         y=\"75\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#ff0000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           id=\"tspan4864\"",
"           y=\"75\"",
"           x=\"5\"",
"           sodipodi:role=\"line\">key:&lt;Primary&gt;Shift</tspan></text>",
"      <text",
"         id=\"text4866\"",
"         y=\"90\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#008000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"90\"",
"           x=\"5\"",
"           id=\"tspan4868\"",
"           sodipodi:role=\"line\">key:&lt;Shift&gt;Shift</tspan></text>",
"      <text",
"         id=\"text4870\"",
"         y=\"105\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#0000ff;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"105\"",
"           x=\"5\"",
"           id=\"tspan4872\"",
"           sodipodi:role=\"line\">key:&lt;Alt&gt;Shift</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-style:normal;font-weight:normal;font-size:12.00000048px;font-family:'Bitstream Vera Sans';fill:#464646;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;-inkscape-font-specification:'Bitstream Vera Sans, Normal';font-stretch:normal;font-variant:normal;text-anchor:start;text-align:start;writing-mode:lr;line-height:125%;\"",
"         x=\"5\"",
"         y=\"120\"",
"         id=\"text6695\"",
"         sodipodi:linespacing=\"125%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan6699\">key:&lt;Super&gt;Shift</tspan></text>",
"      <text",
"         sodipodi:linespacing=\"100%\"",
"         id=\"text4874\"",
"         y=\"135\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#808000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"135\"",
"           x=\"5\"",
"           id=\"tspan4876\"",
"           sodipodi:role=\"line\">key:&lt;Primary&gt;&lt;Shift&gt;Shift</tspan></text>",
"      <text",
"         sodipodi:linespacing=\"100%\"",
"         id=\"text4878\"",
"         y=\"150\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#ff00ff;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"150\"",
"           x=\"5\"",
"           id=\"tspan4880\"",
"           sodipodi:role=\"line\">key:&lt;Primary&gt;&lt;Alt&gt;Shift</tspan></text>",
"      <text",
"         sodipodi:linespacing=\"100%\"",
"         id=\"text4882\"",
"         y=\"165\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#008080;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"165\"",
"           x=\"5\"",
"           id=\"tspan4884\"",
"           sodipodi:role=\"line\">key:&lt;Shift&gt;&lt;Alt&gt;Shift</tspan></text>",
"    </g>",
"    <g",
"       id=\"g4888\"",
"       transform=\"translate(2570,1070)\">",
"      <rect",
"         y=\"2.9921875\"",
"         x=\"0\"",
"         height=\"170\"",
"         width=\"280\"",
"         id=\"rect4890\"",
"         style=\"fill:#0000ff;fill-opacity:0.1;stroke:#000000;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1\" />",
"      <text",
"         sodipodi:linespacing=\"100%\"",
"         id=\"text4892\"",
"         y=\"42.992188\"",
"         x=\"40\"",
"         style=\"font-size:48px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#ff0000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           style=\"font-size:48px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#ff0000;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"           y=\"42.992188\"",
"           x=\"40\"",
"           id=\"tspan4894\"",
"           sodipodi:role=\"line\">Ctrl</tspan></text>",
"      <text",
"         id=\"text4896\"",
"         y=\"60\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#000000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"60\"",
"           x=\"5\"",
"           id=\"tspan4898\"",
"           sodipodi:role=\"line\">key:Ctrl</tspan></text>",
"      <text",
"         id=\"text4900\"",
"         y=\"75\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#ff0000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           id=\"tspan4902\"",
"           y=\"75\"",
"           x=\"5\"",
"           sodipodi:role=\"line\">key:&lt;Primary&gt;Ctrl</tspan></text>",
"      <text",
"         id=\"text4904\"",
"         y=\"90\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#008000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"90\"",
"           x=\"5\"",
"           id=\"tspan4906\"",
"           sodipodi:role=\"line\">key:&lt;Shift&gt;Ctrl</tspan></text>",
"      <text",
"         id=\"text4908\"",
"         y=\"105\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#0000ff;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"105\"",
"           x=\"5\"",
"           id=\"tspan4910\"",
"           sodipodi:role=\"line\">key:&lt;Alt&gt;Ctrl</tspan></text>",
"      <text",
"         sodipodi:linespacing=\"100%\"",
"         id=\"text4912\"",
"         y=\"135\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#808000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"135\"",
"           x=\"5\"",
"           id=\"tspan4914\"",
"           sodipodi:role=\"line\">key:&lt;Primary&gt;&lt;Shift&gt;Ctrl</tspan></text>",
"      <text",
"         sodipodi:linespacing=\"100%\"",
"         id=\"text4916\"",
"         y=\"150\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#ff00ff;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"150\"",
"           x=\"5\"",
"           id=\"tspan4918\"",
"           sodipodi:role=\"line\">key:&lt;Primary&gt;&lt;Alt&gt;Ctrl</tspan></text>",
"      <text",
"         sodipodi:linespacing=\"100%\"",
"         id=\"text4920\"",
"         y=\"165\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#008080;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"165\"",
"           x=\"5\"",
"           id=\"tspan4922\"",
"           sodipodi:role=\"line\">key:&lt;Shift&gt;&lt;Alt&gt;Ctrl</tspan></text>",
"    </g>",
"    <g",
"       transform=\"translate(1700,1070)\"",
"       id=\"g4924\">",
"      <rect",
"         style=\"fill:#0000ff;fill-opacity:0.1;stroke:#000000;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1\"",
"         id=\"rect4926\"",
"         width=\"280\"",
"         height=\"170\"",
"         x=\"0\"",
"         y=\"2.9921875\" />",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:48px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#0000ff;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         x=\"40\"",
"         y=\"42.992188\"",
"         id=\"text4928\"",
"         sodipodi:linespacing=\"100%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan4930\"",
"           x=\"40\"",
"           y=\"42.992188\"",
"           style=\"font-size:48px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#0000ff;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\">Alt</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#000000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"60\"",
"         id=\"text4932\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan4934\"",
"           x=\"5\"",
"           y=\"60\">key:Alt</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#ff0000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"75\"",
"         id=\"text4936\"><tspan",
"           sodipodi:role=\"line\"",
"           x=\"5\"",
"           y=\"75\"",
"           id=\"tspan4938\">key:&lt;Primary&gt;Alt</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#008000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"90\"",
"         id=\"text4940\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan4942\"",
"           x=\"5\"",
"           y=\"90\">key:&lt;Shift&gt;Alt</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#0000ff;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"105\"",
"         id=\"text4944\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan4946\"",
"           x=\"5\"",
"           y=\"105\">key:&lt;Alt&gt;Alt</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#808000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"135\"",
"         id=\"text4948\"",
"         sodipodi:linespacing=\"100%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan4950\"",
"           x=\"5\"",
"           y=\"135\">key:&lt;Primary&gt;&lt;Shift&gt;Alt</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#ff00ff;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"150\"",
"         id=\"text4952\"",
"         sodipodi:linespacing=\"100%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan4954\"",
"           x=\"5\"",
"           y=\"150\">key:&lt;Primary&gt;&lt;Alt&gt;Alt</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#008080;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"165\"",
"         id=\"text4956\"",
"         sodipodi:linespacing=\"100%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan4958\"",
"           x=\"5\"",
"           y=\"165\">key:&lt;Shift&gt;&lt;Alt&gt;Alt</tspan></text>",
"    </g>",
"    <g",
"       id=\"g4960\"",
"       transform=\"translate(1990,1070)\">",
"      <rect",
"         y=\"2.9921875\"",
"         x=\"0\"",
"         height=\"170\"",
"         width=\"280\"",
"         id=\"rect4962\"",
"         style=\"fill:#0000ff;fill-opacity:0.1;stroke:#000000;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1\" />",
"      <text",
"         sodipodi:linespacing=\"100%\"",
"         id=\"text4964\"",
"         y=\"42.992188\"",
"         x=\"40\"",
"         style=\"font-size:48px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#464646;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans;\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"42.992188\"",
"           x=\"40\"",
"           id=\"tspan4966\"",
"           sodipodi:role=\"line\">Win</tspan></text>",
"      <text",
"         id=\"text4968\"",
"         y=\"60\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#000000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"60\"",
"           x=\"5\"",
"           id=\"tspan4970\"",
"           sodipodi:role=\"line\">key:Win</tspan></text>",
"      <text",
"         id=\"text4972\"",
"         y=\"75\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#ff0000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           id=\"tspan4974\"",
"           y=\"75\"",
"           x=\"5\"",
"           sodipodi:role=\"line\">key:&lt;Primary&gt;Win</tspan></text>",
"      <text",
"         id=\"text4976\"",
"         y=\"90\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#008000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"90\"",
"           x=\"5\"",
"           id=\"tspan4978\"",
"           sodipodi:role=\"line\">key:&lt;Shift&gt;Win</tspan></text>",
"      <text",
"         id=\"text4980\"",
"         y=\"105\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#0000ff;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"105\"",
"           x=\"5\"",
"           id=\"tspan4982\"",
"           sodipodi:role=\"line\">key:&lt;Alt&gt;Win</tspan></text>",
"      <text",
"         sodipodi:linespacing=\"100%\"",
"         id=\"text4984\"",
"         y=\"135\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#808000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"135\"",
"           x=\"5\"",
"           id=\"tspan4986\"",
"           sodipodi:role=\"line\">key:&lt;Primary&gt;&lt;Shift&gt;Win</tspan></text>",
"      <text",
"         sodipodi:linespacing=\"100%\"",
"         id=\"text4988\"",
"         y=\"150\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#ff00ff;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"150\"",
"           x=\"5\"",
"           id=\"tspan4990\"",
"           sodipodi:role=\"line\">key:&lt;Primary&gt;&lt;Alt&gt;Win</tspan></text>",
"      <text",
"         sodipodi:linespacing=\"100%\"",
"         id=\"text4992\"",
"         y=\"165\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#008080;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"165\"",
"           x=\"5\"",
"           id=\"tspan4994\"",
"           sodipodi:role=\"line\">key:&lt;Shift&gt;&lt;Alt&gt;Win</tspan></text>",
"    </g>",
"    <g",
"       transform=\"translate(2280,1070)\"",
"       id=\"g4996\">",
"      <rect",
"         style=\"fill:#0000ff;fill-opacity:0.1;stroke:#000000;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1\"",
"         id=\"rect4998\"",
"         width=\"280\"",
"         height=\"170\"",
"         x=\"0\"",
"         y=\"2.9921875\" />",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:48px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#000000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         x=\"40\"",
"         y=\"42.992188\"",
"         id=\"text5000\"",
"         sodipodi:linespacing=\"100%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan5002\"",
"           x=\"40\"",
"           y=\"42.992188\">Menu</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#000000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"60\"",
"         id=\"text5004\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan5006\"",
"           x=\"5\"",
"           y=\"60\">key:Menu</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#ff0000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"75\"",
"         id=\"text5008\"><tspan",
"           sodipodi:role=\"line\"",
"           x=\"5\"",
"           y=\"75\"",
"           id=\"tspan5010\">key:&lt;Primary&gt;Menu</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#008000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"90\"",
"         id=\"text5012\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan5014\"",
"           x=\"5\"",
"           y=\"90\">key:&lt;Shift&gt;Menu</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#0000ff;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"105\"",
"         id=\"text5016\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan5018\"",
"           x=\"5\"",
"           y=\"105\">key:&lt;Alt&gt;Menu</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#808000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"135\"",
"         id=\"text5020\"",
"         sodipodi:linespacing=\"100%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan5022\"",
"           x=\"5\"",
"           y=\"135\">key:&lt;Primary&gt;&lt;Shift&gt;Menu</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#ff00ff;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"150\"",
"         id=\"text5024\"",
"         sodipodi:linespacing=\"100%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan5026\"",
"           x=\"5\"",
"           y=\"150\">key:&lt;Primary&gt;&lt;Alt&gt;Menu</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#008080;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"165\"",
"         id=\"text5028\"",
"         sodipodi:linespacing=\"100%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan5030\"",
"           x=\"5\"",
"           y=\"165\">key:&lt;Shift&gt;&lt;Alt&gt;Menu</tspan></text>",
"    </g>",
"    <g",
"       id=\"g5032\"",
"       transform=\"translate(1030,1070)\">",
"      <rect",
"         y=\"2.9921875\"",
"         x=\"0\"",
"         height=\"170\"",
"         width=\"660\"",
"         id=\"rect5034\"",
"         style=\"fill:#0000ff;fill-opacity:0.1;stroke:#000000;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1\" />",
"      <text",
"         sodipodi:linespacing=\"100%\"",
"         id=\"text5036\"",
"         y=\"42.992188\"",
"         x=\"40\"",
"         style=\"font-size:48px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#000000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"42.992188\"",
"           x=\"40\"",
"           id=\"tspan5038\"",
"           sodipodi:role=\"line\">Space</tspan></text>",
"      <text",
"         id=\"text5040\"",
"         y=\"60\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#000000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"60\"",
"           x=\"5\"",
"           id=\"tspan5042\"",
"           sodipodi:role=\"line\">key:Space</tspan></text>",
"      <text",
"         id=\"text5044\"",
"         y=\"75\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#ff0000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           id=\"tspan5046\"",
"           y=\"75\"",
"           x=\"5\"",
"           sodipodi:role=\"line\">key:&lt;Primary&gt;Space</tspan></text>",
"      <text",
"         id=\"text5048\"",
"         y=\"90\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#008000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"90\"",
"           x=\"5\"",
"           id=\"tspan5050\"",
"           sodipodi:role=\"line\">key:&lt;Shift&gt;Space</tspan></text>",
"      <text",
"         id=\"text5052\"",
"         y=\"105\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#0000ff;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"105\"",
"           x=\"5\"",
"           id=\"tspan5054\"",
"           sodipodi:role=\"line\">key:&lt;Alt&gt;Space</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-style:normal;font-weight:normal;font-size:12.00000048px;font-family:'Bitstream Vera Sans';fill:#464646;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;-inkscape-font-specification:'Bitstream Vera Sans, Normal';font-stretch:normal;font-variant:normal;text-anchor:start;text-align:start;writing-mode:lr;line-height:125%;\"",
"         x=\"5\"",
"         y=\"120\"",
"         id=\"text6701\"",
"         sodipodi:linespacing=\"125%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan6705\">key:&lt;Super&gt;Space</tspan></text>",
"      <text",
"         sodipodi:linespacing=\"100%\"",
"         id=\"text5056\"",
"         y=\"135\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#808000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"135\"",
"           x=\"5\"",
"           id=\"tspan5058\"",
"           sodipodi:role=\"line\">key:&lt;Primary&gt;&lt;Shift&gt;Space</tspan></text>",
"      <text",
"         sodipodi:linespacing=\"100%\"",
"         id=\"text5060\"",
"         y=\"150\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#ff00ff;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"150\"",
"           x=\"5\"",
"           id=\"tspan5062\"",
"           sodipodi:role=\"line\">key:&lt;Primary&gt;&lt;Alt&gt;Space</tspan></text>",
"      <text",
"         sodipodi:linespacing=\"100%\"",
"         id=\"text5064\"",
"         y=\"165\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#008080;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"165\"",
"           x=\"5\"",
"           id=\"tspan5066\"",
"           sodipodi:role=\"line\">key:&lt;Shift&gt;&lt;Alt&gt;Space</tspan></text>",
"    </g>",
"    <g",
"       id=\"g5068\"",
"       transform=\"matrix(0.52641739,0,0,1.0009796,2654.9902,709.64482)\">",
"      <path",
"         style=\"fill:#0000ff;fill-opacity:0.1;stroke:#000000;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1\"",
"         d=\"M 0,-5.9276459 -81.673421,-6.2844393 -83.030302,-175.58064 370,-177.00781 l 0,350 -370,0 z\"",
"         id=\"rect5070\"",
"         sodipodi:nodetypes=\"ccccccc\"",
"         inkscape:connector-curvature=\"0\" />",
"      <text",
"         sodipodi:linespacing=\"100%\"",
"         id=\"text5072\"",
"         y=\"-131.57762\"",
"         x=\"85.018333\"",
"         style=\"font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;font-size:48px;line-height:100%;font-family:'Bitstream Vera Sans';-inkscape-font-specification:'Bitstream Vera Sans';text-align:start;writing-mode:lr-tb;text-anchor:start;fill:#000000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"-131.57762\"",
"           x=\"85.018333\"",
"           id=\"tspan5074\"",
"           sodipodi:role=\"line\">Enter</tspan></text>",
"      <text",
"         id=\"text5076\"",
"         y=\"-101.5424\"",
"         x=\"12.855455\"",
"         style=\"font-style:normal;font-weight:normal;font-size:18.25518608px;font-family:'Bitstream Vera Sans';fill:#000000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1\"",
"         xml:space=\"preserve\"",
"         transform=\"scale(1.2142803,0.82353311)\"><tspan",
"           y=\"-101.5424\"",
"           x=\"12.855455\"",
"           id=\"tspan5078\"",
"           sodipodi:role=\"line\">key:Enter</tspan></text>",
"      <text",
"         id=\"text5080\"",
"         y=\"-78.592957\"",
"         x=\"12.817163\"",
"         style=\"font-style:normal;font-weight:normal;font-size:18.41451836px;font-family:'Bitstream Vera Sans';fill:#ff0000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1\"",
"         xml:space=\"preserve\"",
"         transform=\"scale(1.2162767,0.82218135)\"><tspan",
"           id=\"tspan5082\"",
"           y=\"-78.592957\"",
"           x=\"12.817163\"",
"           sodipodi:role=\"line\">key:&lt;Primary&gt;Enter</tspan></text>",
"      <text",
"         id=\"text5084\"",
"         y=\"-55.57481\"",
"         x=\"12.817163\"",
"         style=\"font-style:normal;font-weight:normal;font-size:18.41451836px;font-family:'Bitstream Vera Sans';fill:#008000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1\"",
"         xml:space=\"preserve\"",
"         transform=\"scale(1.2162767,0.82218135)\"><tspan",
"           y=\"-55.57481\"",
"           x=\"12.817163\"",
"           id=\"tspan5086\"",
"           sodipodi:role=\"line\">key:&lt;Shift&gt;Enter</tspan></text>",
"      <text",
"         id=\"text5088\"",
"         y=\"-32.556656\"",
"         x=\"12.817163\"",
"         style=\"font-style:normal;font-weight:normal;font-size:18.41451836px;font-family:'Bitstream Vera Sans';fill:#0000ff;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1\"",
"         xml:space=\"preserve\"",
"         transform=\"scale(1.2162767,0.82218135)\"><tspan",
"           y=\"-32.556656\"",
"           x=\"12.817163\"",
"           id=\"tspan5090\"",
"           sodipodi:role=\"line\">key:&lt;Alt&gt;Enter</tspan></text>",
"      <text",
"         sodipodi:linespacing=\"100%\"",
"         id=\"text5092\"",
"         y=\"-9.5385094\"",
"         x=\"12.817163\"",
"         style=\"font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;font-size:18.41451836px;line-height:100%;font-family:'Bitstream Vera Sans';-inkscape-font-specification:'Bitstream Vera Sans';text-align:start;writing-mode:lr-tb;text-anchor:start;fill:#808000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1\"",
"         xml:space=\"preserve\"",
"         transform=\"scale(1.2162767,0.82218135)\"><tspan",
"           y=\"-9.5385094\"",
"           x=\"12.817163\"",
"           id=\"tspan5094\"",
"           sodipodi:role=\"line\">key:&lt;Primary&gt;&lt;Shift&gt;Enter</tspan></text>",
"      <text",
"         sodipodi:linespacing=\"100%\"",
"         id=\"text5096\"",
"         y=\"13.479638\"",
"         x=\"12.817163\"",
"         style=\"font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;font-size:18.41451836px;line-height:100%;font-family:'Bitstream Vera Sans';-inkscape-font-specification:'Bitstream Vera Sans';text-align:start;writing-mode:lr-tb;text-anchor:start;fill:#ff00ff;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1\"",
"         xml:space=\"preserve\"",
"         transform=\"scale(1.2162767,0.82218135)\"><tspan",
"           y=\"13.479638\"",
"           x=\"12.817163\"",
"           id=\"tspan5098\"",
"           sodipodi:role=\"line\">key:&lt;Primary&gt;&lt;Alt&gt;Enter</tspan></text>",
"      <text",
"         sodipodi:linespacing=\"100%\"",
"         id=\"text5100\"",
"         y=\"35.371498\"",
"         x=\"12.855455\"",
"         style=\"font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;font-size:18.25518608px;line-height:100%;font-family:'Bitstream Vera Sans';-inkscape-font-specification:'Bitstream Vera Sans';text-align:start;writing-mode:lr-tb;text-anchor:start;fill:#008080;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1\"",
"         xml:space=\"preserve\"",
"         transform=\"scale(1.2142803,0.82353311)\"><tspan",
"           y=\"35.371498\"",
"           x=\"12.855455\"",
"           id=\"tspan5102\"",
"           sodipodi:role=\"line\">key:&lt;Shift&gt;&lt;Alt&gt;Enter</tspan></text>",
"      <flowRoot",
"         xml:space=\"preserve\"",
"         id=\"flowRoot12379\"",
"         style=\"fill:black;stroke:none;stroke-opacity:1;stroke-width:1px;stroke-linejoin:miter;stroke-linecap:butt;fill-opacity:1;font-family:sans-serif;font-style:normal;font-weight:normal;font-size:40px;line-height:125%;letter-spacing:0px;word-spacing:0px\"><flowRegion",
"           id=\"flowRegion12381\"><rect",
"             id=\"rect12383\"",
"             width=\"80\"",
"             height=\"50\"",
"             x=\"2740\"",
"             y=\"610\" /></flowRegion><flowPara",
"           id=\"flowPara12385\" /></flowRoot>      <flowRoot",
"         xml:space=\"preserve\"",
"         id=\"flowRoot12387\"",
"         style=\"fill:black;stroke:none;stroke-opacity:1;stroke-width:1px;stroke-linejoin:miter;stroke-linecap:butt;fill-opacity:1;font-family:sans-serif;font-style:normal;font-weight:normal;font-size:40px;line-height:125%;letter-spacing:0px;word-spacing:0px\"><flowRegion",
"           id=\"flowRegion12389\"><rect",
"             id=\"rect12391\"",
"             width=\"100\"",
"             height=\"70\"",
"             x=\"2730\"",
"             y=\"610\" /></flowRegion><flowPara",
"           id=\"flowPara12393\" /></flowRoot>      <flowRoot",
"         xml:space=\"preserve\"",
"         id=\"flowRoot12395\"",
"         style=\"fill:black;stroke:none;stroke-opacity:1;stroke-width:1px;stroke-linejoin:miter;stroke-linecap:butt;fill-opacity:1;font-family:sans-serif;font-style:normal;font-weight:normal;font-size:40px;line-height:125%;letter-spacing:0px;word-spacing:0px\"><flowRegion",
"           id=\"flowRegion12397\"><rect",
"             id=\"rect12399\"",
"             width=\"80\"",
"             height=\"70\"",
"             x=\"2660\"",
"             y=\"750\" /></flowRegion><flowPara",
"           id=\"flowPara12401\" /></flowRoot>      <flowRoot",
"         xml:space=\"preserve\"",
"         id=\"flowRoot12403\"",
"         style=\"fill:black;stroke:none;stroke-opacity:1;stroke-width:1px;stroke-linejoin:miter;stroke-linecap:butt;fill-opacity:1;font-family:sans-serif;font-style:normal;font-weight:normal;font-size:40px;line-height:125%;letter-spacing:0px;word-spacing:0px\"><flowRegion",
"           id=\"flowRegion12405\"><rect",
"             id=\"rect12407\"",
"             width=\"80\"",
"             height=\"110\"",
"             x=\"2660\"",
"             y=\"750\" /></flowRegion><flowPara",
"           id=\"flowPara12409\" /></flowRoot>      <flowRoot",
"         xml:space=\"preserve\"",
"         id=\"flowRoot12411\"",
"         style=\"fill:black;stroke:none;stroke-opacity:1;stroke-width:1px;stroke-linejoin:miter;stroke-linecap:butt;fill-opacity:1;font-family:sans-serif;font-style:normal;font-weight:normal;font-size:40px;line-height:125%;letter-spacing:0px;word-spacing:0px\"><flowRegion",
"           id=\"flowRegion12413\"><rect",
"             id=\"rect12415\"",
"             width=\"90\"",
"             height=\"100\"",
"             x=\"2660\"",
"             y=\"760\" /></flowRegion><flowPara",
"           id=\"flowPara12417\" /></flowRoot>    </g>",
"    <g",
"       transform=\"translate(1560,1450)\"",
"       id=\"g3712\">",
"      <rect",
"         style=\"fill:#0000ff;fill-opacity:0.1;stroke:#000000;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1\"",
"         id=\"rect3714\"",
"         width=\"170\"",
"         height=\"170\"",
"         x=\"0\"",
"         y=\"2.9921875\" />",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:48px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#000000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         x=\"20\"",
"         y=\"50\"",
"         id=\"text3716\"",
"         sodipodi:linespacing=\"100%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan3718\"",
"           x=\"20\"",
"           y=\"50\">PrtScr</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12.00000048px;font-style:normal;font-weight:normal;fill:#000000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:'Bitstream Vera Sans';-inkscape-font-specification:'Bitstream Vera Sans, Normal';font-stretch:normal;font-variant:normal;text-anchor:start;text-align:start;writing-mode:lr;line-height:125%;\"",
"         x=\"5\"",
"         y=\"60\"",
"         id=\"text3720\"",
"         sodipodi:linespacing=\"125%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan6375\">key:Print</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12.00000048px;font-style:normal;font-weight:normal;fill:#ff0000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:'Bitstream Vera Sans';-inkscape-font-specification:'Bitstream Vera Sans, Normal';font-stretch:normal;font-variant:normal;text-anchor:start;text-align:start;writing-mode:lr;line-height:125%;\"",
"         x=\"5\"",
"         y=\"75\"",
"         id=\"text3724\"",
"         sodipodi:linespacing=\"125%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan6377\">key:&lt;Primary&gt;Print</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12.00000048px;font-style:normal;font-weight:normal;fill:#008000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:'Bitstream Vera Sans';-inkscape-font-specification:'Bitstream Vera Sans, Normal';font-stretch:normal;font-variant:normal;text-anchor:start;text-align:start;writing-mode:lr;line-height:125%;\"",
"         x=\"5\"",
"         y=\"90\"",
"         id=\"text3728\"",
"         sodipodi:linespacing=\"125%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan6379\">key:&lt;Shift&gt;Print</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12.00000048px;font-style:normal;font-weight:normal;fill:#0000ff;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:'Bitstream Vera Sans';-inkscape-font-specification:'Bitstream Vera Sans, Normal';font-stretch:normal;font-variant:normal;text-anchor:start;text-align:start;writing-mode:lr;line-height:125%;\"",
"         x=\"5\"",
"         y=\"105\"",
"         id=\"text3732\"",
"         sodipodi:linespacing=\"125%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan6381\">key:&lt;Alt&gt;Print</tspan></text>",
"      <text",
"         id=\"text6735\"",
"         y=\"120\"",
"         x=\"5\"",
"         style=\"font-style:normal;font-weight:normal;font-size:12.00000048px;font-family:'Bitstream Vera Sans';fill:#464646;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;-inkscape-font-specification:'Bitstream Vera Sans, Normal';font-stretch:normal;font-variant:normal;text-anchor:start;text-align:start;writing-mode:lr;line-height:125%;\"",
"         xml:space=\"preserve\"",
"         sodipodi:linespacing=\"125%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan6383\">key:&lt;Super&gt;Print</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12.00000048px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr;text-anchor:start;fill:#808000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:'Bitstream Vera Sans';-inkscape-font-specification:'Bitstream Vera Sans, Normal';\"",
"         x=\"5\"",
"         y=\"135\"",
"         id=\"text3736\"",
"         sodipodi:linespacing=\"100%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan6385\">key:&lt;Primary&gt;&lt;Shift&gt;Print</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12.00000048px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr;text-anchor:start;fill:#ff00ff;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:'Bitstream Vera Sans';-inkscape-font-specification:'Bitstream Vera Sans, Normal';\"",
"         x=\"5\"",
"         y=\"150\"",
"         id=\"text3740\"",
"         sodipodi:linespacing=\"100%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan6387\">key:&lt;Primary&gt;&lt;Alt&gt;Print</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12.00000048px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr;text-anchor:start;fill:#008080;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:'Bitstream Vera Sans';-inkscape-font-specification:'Bitstream Vera Sans, Normal';\"",
"         x=\"5\"",
"         y=\"165\"",
"         id=\"text3744\"",
"         sodipodi:linespacing=\"100%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan6389\">key:&lt;Shift&gt;&lt;Alt&gt;Print</tspan></text>",
"    </g>",
"    <g",
"       id=\"g3748\"",
"       transform=\"translate(1560,1670)\">",
"      <rect",
"         y=\"2.9921875\"",
"         x=\"0\"",
"         height=\"170\"",
"         width=\"170\"",
"         id=\"rect3750\"",
"         style=\"fill:#0000ff;fill-opacity:0.1;stroke:#000000;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1\" />",
"      <text",
"         sodipodi:linespacing=\"100%\"",
"         id=\"text3752\"",
"         y=\"42.992188\"",
"         x=\"40\"",
"         style=\"font-size:48px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#000000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"42.992188\"",
"           x=\"40\"",
"           id=\"tspan3754\"",
"           sodipodi:role=\"line\">Insert</tspan></text>",
"      <text",
"         id=\"text3756\"",
"         y=\"60\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#000000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"60\"",
"           x=\"5\"",
"           id=\"tspan3758\"",
"           sodipodi:role=\"line\">key:Insert</tspan></text>",
"      <text",
"         id=\"text3760\"",
"         y=\"75\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#ff0000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           id=\"tspan3762\"",
"           y=\"75\"",
"           x=\"5\"",
"           sodipodi:role=\"line\">key:&lt;Primary&gt;Insert</tspan></text>",
"      <text",
"         id=\"text3764\"",
"         y=\"90\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#008000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"90\"",
"           x=\"5\"",
"           id=\"tspan3766\"",
"           sodipodi:role=\"line\">key:&lt;Shift&gt;Insert</tspan></text>",
"      <text",
"         id=\"text3768\"",
"         y=\"105\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#0000ff;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"105\"",
"           x=\"5\"",
"           id=\"tspan3770\"",
"           sodipodi:role=\"line\">key:&lt;Alt&gt;Insert</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-style:normal;font-weight:normal;font-size:12.00000048px;font-family:'Bitstream Vera Sans';fill:#464646;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;-inkscape-font-specification:'Bitstream Vera Sans, Normal';font-stretch:normal;font-variant:normal;text-anchor:start;text-align:start;writing-mode:lr;line-height:125%;\"",
"         x=\"5\"",
"         y=\"120\"",
"         id=\"text6753\"",
"         sodipodi:linespacing=\"125%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan6757\">key:&lt;Super&gt;Insert</tspan></text>",
"      <text",
"         sodipodi:linespacing=\"100%\"",
"         id=\"text3772\"",
"         y=\"135\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#808000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"135\"",
"           x=\"5\"",
"           id=\"tspan3774\"",
"           sodipodi:role=\"line\">key:&lt;Primary&gt;&lt;Shift&gt;Insert</tspan></text>",
"      <text",
"         sodipodi:linespacing=\"100%\"",
"         id=\"text3776\"",
"         y=\"150\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#ff00ff;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"150\"",
"           x=\"5\"",
"           id=\"tspan3778\"",
"           sodipodi:role=\"line\">key:&lt;Primary&gt;&lt;Alt&gt;Insert</tspan></text>",
"      <text",
"         sodipodi:linespacing=\"100%\"",
"         id=\"text3780\"",
"         y=\"165\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#008080;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"165\"",
"           x=\"5\"",
"           id=\"tspan3782\"",
"           sodipodi:role=\"line\">key:&lt;Shift&gt;&lt;Alt&gt;Insert</tspan></text>",
"    </g>",
"    <g",
"       id=\"g3784\"",
"       transform=\"translate(1920,1670)\">",
"      <rect",
"         y=\"2.9921875\"",
"         x=\"0\"",
"         height=\"170\"",
"         width=\"170\"",
"         id=\"rect3786\"",
"         style=\"fill:#0000ff;fill-opacity:0.1;stroke:#000000;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1\" />",
"      <text",
"         sodipodi:linespacing=\"100%\"",
"         id=\"text3788\"",
"         y=\"42.992188\"",
"         x=\"40\"",
"         style=\"font-size:48px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#000000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"42.992188\"",
"           x=\"40\"",
"           id=\"tspan3790\"",
"           sodipodi:role=\"line\">PgUp</tspan></text>",
"      <text",
"         id=\"text3792\"",
"         y=\"60\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#000000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"60\"",
"           x=\"5\"",
"           id=\"tspan3794\"",
"           sodipodi:role=\"line\">key:Page_Up</tspan></text>",
"      <text",
"         id=\"text3796\"",
"         y=\"75\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#ff0000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           id=\"tspan3798\"",
"           y=\"75\"",
"           x=\"5\"",
"           sodipodi:role=\"line\">key:&lt;Primary&gt;Page_Up</tspan></text>",
"      <text",
"         id=\"text3800\"",
"         y=\"90\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#008000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"90\"",
"           x=\"5\"",
"           id=\"tspan3802\"",
"           sodipodi:role=\"line\">key:&lt;Shift&gt;Page_Up</tspan></text>",
"      <text",
"         id=\"text3804\"",
"         y=\"105\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#0000ff;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"105\"",
"           x=\"5\"",
"           id=\"tspan3806\"",
"           sodipodi:role=\"line\">key:&lt;Alt&gt;Page_Up</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-style:normal;font-weight:normal;font-size:12.00000048px;font-family:'Bitstream Vera Sans';fill:#464646;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;-inkscape-font-specification:'Bitstream Vera Sans, Normal';font-stretch:normal;font-variant:normal;text-anchor:start;text-align:start;writing-mode:lr;line-height:125%;\"",
"         x=\"5\"",
"         y=\"120\"",
"         id=\"text6765\"",
"         sodipodi:linespacing=\"125%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan6769\">key:&lt;Super&gt;Page_Up</tspan></text>",
"      <text",
"         sodipodi:linespacing=\"100%\"",
"         id=\"text3808\"",
"         y=\"135\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#808000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"135\"",
"           x=\"5\"",
"           id=\"tspan3810\"",
"           sodipodi:role=\"line\">key:&lt;Primary&gt;&lt;Shift&gt;Page_Up</tspan></text>",
"      <text",
"         sodipodi:linespacing=\"100%\"",
"         id=\"text3812\"",
"         y=\"150\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#ff00ff;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"150\"",
"           x=\"5\"",
"           id=\"tspan3814\"",
"           sodipodi:role=\"line\">key:&lt;Primary&gt;&lt;Alt&gt;Page_Up</tspan></text>",
"      <text",
"         sodipodi:linespacing=\"100%\"",
"         id=\"text3816\"",
"         y=\"165\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#008080;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"165\"",
"           x=\"5\"",
"           id=\"tspan3818\"",
"           sodipodi:role=\"line\">key:&lt;Shift&gt;&lt;Alt&gt;Page_Up</tspan></text>",
"    </g>",
"    <g",
"       transform=\"translate(1740,1670)\"",
"       id=\"g3820\">",
"      <rect",
"         style=\"fill:#0000ff;fill-opacity:0.1;stroke:#000000;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1\"",
"         id=\"rect3822\"",
"         width=\"170\"",
"         height=\"170\"",
"         x=\"0\"",
"         y=\"2.9921875\" />",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:48px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#000000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         x=\"40\"",
"         y=\"42.992188\"",
"         id=\"text3824\"",
"         sodipodi:linespacing=\"100%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan3826\"",
"           x=\"40\"",
"           y=\"42.992188\">Home</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#000000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"60\"",
"         id=\"text3828\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan3830\"",
"           x=\"5\"",
"           y=\"60\">key:Home</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#ff0000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"75\"",
"         id=\"text3832\"><tspan",
"           sodipodi:role=\"line\"",
"           x=\"5\"",
"           y=\"75\"",
"           id=\"tspan3834\">key:&lt;Primary&gt;Home</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#008000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"90\"",
"         id=\"text3836\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan3838\"",
"           x=\"5\"",
"           y=\"90\">key:&lt;Shift&gt;Home</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#0000ff;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"105\"",
"         id=\"text3840\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan3842\"",
"           x=\"5\"",
"           y=\"105\">key:&lt;Alt&gt;Home</tspan></text>",
"      <text",
"         id=\"text6759\"",
"         y=\"120\"",
"         x=\"5\"",
"         style=\"font-style:normal;font-weight:normal;font-size:12.00000048px;font-family:'Bitstream Vera Sans';fill:#464646;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;-inkscape-font-specification:'Bitstream Vera Sans, Normal';font-stretch:normal;font-variant:normal;text-anchor:start;text-align:start;writing-mode:lr;line-height:125%;\"",
"         xml:space=\"preserve\"",
"         sodipodi:linespacing=\"125%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan6763\">key:&lt;Super&gt;Home</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#808000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"135\"",
"         id=\"text3844\"",
"         sodipodi:linespacing=\"100%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan3846\"",
"           x=\"5\"",
"           y=\"135\">key:&lt;Primary&gt;&lt;Shift&gt;Home</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#ff00ff;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"150\"",
"         id=\"text3848\"",
"         sodipodi:linespacing=\"100%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan3850\"",
"           x=\"5\"",
"           y=\"150\">key:&lt;Primary&gt;&lt;Alt&gt;Home</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#008080;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"165\"",
"         id=\"text3852\"",
"         sodipodi:linespacing=\"100%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan3854\"",
"           x=\"5\"",
"           y=\"165\">key:&lt;Shift&gt;&lt;Alt&gt;Home</tspan></text>",
"    </g>",
"    <g",
"       id=\"g3856\"",
"       transform=\"translate(1920,1450)\">",
"      <rect",
"         y=\"2.9921875\"",
"         x=\"0\"",
"         height=\"170\"",
"         width=\"170\"",
"         id=\"rect3858\"",
"         style=\"fill:#0000ff;fill-opacity:0.1;stroke:#000000;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1\" />",
"      <text",
"         sodipodi:linespacing=\"100%\"",
"         id=\"text3860\"",
"         y=\"50\"",
"         x=\"20\"",
"         style=\"font-size:48px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#000000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"50\"",
"           x=\"20\"",
"           id=\"tspan3862\"",
"           sodipodi:role=\"line\">Pause</tspan></text>",
"      <text",
"         id=\"text3864\"",
"         y=\"60\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#000000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"60\"",
"           x=\"5\"",
"           id=\"tspan3866\"",
"           sodipodi:role=\"line\">key:Pause</tspan></text>",
"      <text",
"         id=\"text3868\"",
"         y=\"75\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#ff0000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           id=\"tspan3870\"",
"           y=\"75\"",
"           x=\"5\"",
"           sodipodi:role=\"line\">key:&lt;Primary&gt;Pause</tspan></text>",
"      <text",
"         id=\"text3872\"",
"         y=\"90\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#008000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"90\"",
"           x=\"5\"",
"           id=\"tspan3874\"",
"           sodipodi:role=\"line\">key:&lt;Shift&gt;Pause</tspan></text>",
"      <text",
"         id=\"text3876\"",
"         y=\"105\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#0000ff;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"105\"",
"           x=\"5\"",
"           id=\"tspan3878\"",
"           sodipodi:role=\"line\">key:&lt;Alt&gt;Pause</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-style:normal;font-weight:normal;font-size:12.00000048px;font-family:'Bitstream Vera Sans';fill:#464646;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;-inkscape-font-specification:'Bitstream Vera Sans, Normal';font-stretch:normal;font-variant:normal;text-anchor:start;text-align:start;writing-mode:lr;line-height:125%;\"",
"         x=\"5\"",
"         y=\"120\"",
"         id=\"text6747\"",
"         sodipodi:linespacing=\"125%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan6751\">key:&lt;Super&gt;Pause</tspan></text>",
"      <text",
"         sodipodi:linespacing=\"100%\"",
"         id=\"text3880\"",
"         y=\"135\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#808000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"135\"",
"           x=\"5\"",
"           id=\"tspan3882\"",
"           sodipodi:role=\"line\">key:&lt;Primary&gt;&lt;Shift&gt;Pause</tspan></text>",
"      <text",
"         sodipodi:linespacing=\"100%\"",
"         id=\"text3884\"",
"         y=\"150\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#ff00ff;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"150\"",
"           x=\"5\"",
"           id=\"tspan3886\"",
"           sodipodi:role=\"line\">key:&lt;Primary&gt;&lt;Alt&gt;Pause</tspan></text>",
"      <text",
"         sodipodi:linespacing=\"100%\"",
"         id=\"text3888\"",
"         y=\"165\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#008080;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"165\"",
"           x=\"5\"",
"           id=\"tspan3890\"",
"           sodipodi:role=\"line\">key:&lt;Shift&gt;&lt;Alt&gt;Pause</tspan></text>",
"    </g>",
"    <g",
"       transform=\"translate(1740,1450)\"",
"       id=\"g3892\">",
"      <rect",
"         style=\"fill:#0000ff;fill-opacity:0.1;stroke:#000000;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1\"",
"         id=\"rect3894\"",
"         width=\"170\"",
"         height=\"170\"",
"         x=\"0\"",
"         y=\"2.9921875\" />",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:48px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#000000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         x=\"40\"",
"         y=\"42.992188\"",
"         id=\"text3896\"",
"         sodipodi:linespacing=\"100%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan3898\"",
"           x=\"40\"",
"           y=\"42.992188\">ScrL</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#000000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"60\"",
"         id=\"text3900\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan3902\"",
"           x=\"5\"",
"           y=\"60\">key:Scroll_Lock</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12.00000048px;font-style:normal;font-weight:normal;fill:#ff0000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:'Bitstream Vera Sans';-inkscape-font-specification:'Bitstream Vera Sans, Normal';font-stretch:normal;font-variant:normal;text-anchor:start;text-align:start;writing-mode:lr;line-height:125%;\"",
"         x=\"5\"",
"         y=\"75\"",
"         id=\"text3904\"",
"         sodipodi:linespacing=\"125%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan6391\">key:&lt;Primary&gt;Scroll_Lock</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12.00000048px;font-style:normal;font-weight:normal;fill:#008000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:'Bitstream Vera Sans';-inkscape-font-specification:'Bitstream Vera Sans, Normal';font-stretch:normal;font-variant:normal;text-anchor:start;text-align:start;writing-mode:lr;line-height:125%;\"",
"         x=\"5\"",
"         y=\"90\"",
"         id=\"text3908\"",
"         sodipodi:linespacing=\"125%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan6393\">key:&lt;Shift&gt;Scroll_Lock</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12.00000048px;font-style:normal;font-weight:normal;fill:#0000ff;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:'Bitstream Vera Sans';-inkscape-font-specification:'Bitstream Vera Sans, Normal';font-stretch:normal;font-variant:normal;text-anchor:start;text-align:start;writing-mode:lr;line-height:125%;\"",
"         x=\"5\"",
"         y=\"105\"",
"         id=\"text3912\"",
"         sodipodi:linespacing=\"125%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan6396\">key:&lt;Alt&gt;Scroll_Lock</tspan></text>",
"      <text",
"         id=\"text6741\"",
"         y=\"120\"",
"         x=\"5\"",
"         style=\"font-style:normal;font-weight:normal;font-size:12.00000048px;font-family:'Bitstream Vera Sans';fill:#464646;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;-inkscape-font-specification:'Bitstream Vera Sans, Normal';font-stretch:normal;font-variant:normal;text-anchor:start;text-align:start;writing-mode:lr;line-height:125%;\"",
"         xml:space=\"preserve\"",
"         sodipodi:linespacing=\"125%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan6398\">key:&lt;Super&gt;Scroll_Lock</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12.00000048px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr;text-anchor:start;fill:#808000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:'Bitstream Vera Sans';-inkscape-font-specification:'Bitstream Vera Sans, Normal';\"",
"         x=\"5\"",
"         y=\"135\"",
"         id=\"text3916\"",
"         sodipodi:linespacing=\"100%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan6400\">key:&lt;Primary&gt;&lt;Shift&gt;Scroll_Lock</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12.00000048px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr;text-anchor:start;fill:#ff00ff;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:'Bitstream Vera Sans';-inkscape-font-specification:'Bitstream Vera Sans, Normal';\"",
"         x=\"5\"",
"         y=\"150\"",
"         id=\"text3920\"",
"         sodipodi:linespacing=\"100%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan6402\">key:&lt;Primary&gt;&lt;Alt&gt;Scroll_Lock</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12.00000048px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr;text-anchor:start;fill:#008080;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:'Bitstream Vera Sans';-inkscape-font-specification:'Bitstream Vera Sans, Normal';\"",
"         x=\"5\"",
"         y=\"165\"",
"         id=\"text3924\"",
"         sodipodi:linespacing=\"100%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan6404\">key:&lt;Shift&gt;&lt;Alt&gt;Scroll_Lock</tspan></text>",
"    </g>",
"    <g",
"       transform=\"translate(1920,1850)\"",
"       id=\"g3928\">",
"      <rect",
"         style=\"fill:#0000ff;fill-opacity:0.1;stroke:#000000;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1\"",
"         id=\"rect3930\"",
"         width=\"170\"",
"         height=\"170\"",
"         x=\"0\"",
"         y=\"2.9921875\" />",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:48px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#000000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         x=\"40\"",
"         y=\"42.992188\"",
"         id=\"text3932\"",
"         sodipodi:linespacing=\"100%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan3934\"",
"           x=\"40\"",
"           y=\"42.992188\">PgDn</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#000000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"60\"",
"         id=\"text3936\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan3938\"",
"           x=\"5\"",
"           y=\"60\">key:Page_Down</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#ff0000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"75\"",
"         id=\"text3940\"><tspan",
"           sodipodi:role=\"line\"",
"           x=\"5\"",
"           y=\"75\"",
"           id=\"tspan3942\">key:&lt;Primary&gt;Page_Down</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#008000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"90\"",
"         id=\"text3944\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan3946\"",
"           x=\"5\"",
"           y=\"90\">key:&lt;Shift&gt;Page_Down</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#0000ff;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"105\"",
"         id=\"text3948\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan3950\"",
"           x=\"5\"",
"           y=\"105\">key:&lt;Alt&gt;Page_Down</tspan></text>",
"      <text",
"         id=\"text6783\"",
"         y=\"120\"",
"         x=\"5\"",
"         style=\"font-style:normal;font-weight:normal;font-size:12.00000048px;font-family:'Bitstream Vera Sans';fill:#464646;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;-inkscape-font-specification:'Bitstream Vera Sans, Normal';font-stretch:normal;font-variant:normal;text-anchor:start;text-align:start;writing-mode:lr;line-height:125%;\"",
"         xml:space=\"preserve\"",
"         sodipodi:linespacing=\"125%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan6787\">key:&lt;Super&gt;Page_Down</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#808000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"135\"",
"         id=\"text3952\"",
"         sodipodi:linespacing=\"100%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan3954\"",
"           x=\"5\"",
"           y=\"135\">key:&lt;Primary&gt;&lt;Shift&gt;Page_Down</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#ff00ff;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"150\"",
"         id=\"text3956\"",
"         sodipodi:linespacing=\"100%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan3958\"",
"           x=\"5\"",
"           y=\"150\">key:&lt;Primary&gt;&lt;Alt&gt;Page_Down</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#008080;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"165\"",
"         id=\"text3960\"",
"         sodipodi:linespacing=\"100%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan3962\"",
"           x=\"5\"",
"           y=\"165\">key:&lt;Shift&gt;&lt;Alt&gt;Page_Down</tspan></text>",
"    </g>",
"    <g",
"       id=\"g3964\"",
"       transform=\"translate(1740,1850)\">",
"      <rect",
"         y=\"2.9921875\"",
"         x=\"0\"",
"         height=\"170\"",
"         width=\"170\"",
"         id=\"rect3966\"",
"         style=\"fill:#0000ff;fill-opacity:0.1;stroke:#000000;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1\" />",
"      <text",
"         sodipodi:linespacing=\"100%\"",
"         id=\"text3968\"",
"         y=\"42.992188\"",
"         x=\"40\"",
"         style=\"font-size:48px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#000000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"42.992188\"",
"           x=\"40\"",
"           id=\"tspan3970\"",
"           sodipodi:role=\"line\">End</tspan></text>",
"      <text",
"         id=\"text3972\"",
"         y=\"60\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#000000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"60\"",
"           x=\"5\"",
"           id=\"tspan3974\"",
"           sodipodi:role=\"line\">key:End</tspan></text>",
"      <text",
"         id=\"text3976\"",
"         y=\"75\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#ff0000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           id=\"tspan3978\"",
"           y=\"75\"",
"           x=\"5\"",
"           sodipodi:role=\"line\">key:&lt;Primary&gt;End</tspan></text>",
"      <text",
"         id=\"text3980\"",
"         y=\"90\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#008000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"90\"",
"           x=\"5\"",
"           id=\"tspan3982\"",
"           sodipodi:role=\"line\">key:&lt;Shift&gt;End</tspan></text>",
"      <text",
"         id=\"text3984\"",
"         y=\"105\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#0000ff;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"105\"",
"           x=\"5\"",
"           id=\"tspan3986\"",
"           sodipodi:role=\"line\">key:&lt;Alt&gt;End</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-style:normal;font-weight:normal;font-size:12.00000048px;font-family:'Bitstream Vera Sans';fill:#464646;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;-inkscape-font-specification:'Bitstream Vera Sans, Normal';font-stretch:normal;font-variant:normal;text-anchor:start;text-align:start;writing-mode:lr;line-height:125%;\"",
"         x=\"5\"",
"         y=\"120\"",
"         id=\"text6777\"",
"         sodipodi:linespacing=\"125%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan6781\">key:&lt;Super&gt;End</tspan></text>",
"      <text",
"         sodipodi:linespacing=\"100%\"",
"         id=\"text3988\"",
"         y=\"135\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#808000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"135\"",
"           x=\"5\"",
"           id=\"tspan3990\"",
"           sodipodi:role=\"line\">key:&lt;Primary&gt;&lt;Shift&gt;End</tspan></text>",
"      <text",
"         sodipodi:linespacing=\"100%\"",
"         id=\"text3992\"",
"         y=\"150\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#ff00ff;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"150\"",
"           x=\"5\"",
"           id=\"tspan3994\"",
"           sodipodi:role=\"line\">key:&lt;Primary&gt;&lt;Alt&gt;End</tspan></text>",
"      <text",
"         sodipodi:linespacing=\"100%\"",
"         id=\"text3996\"",
"         y=\"165\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#008080;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"165\"",
"           x=\"5\"",
"           id=\"tspan3998\"",
"           sodipodi:role=\"line\">key:&lt;Shift&gt;&lt;Alt&gt;End</tspan></text>",
"    </g>",
"    <g",
"       transform=\"translate(1560,1850)\"",
"       id=\"g4000\">",
"      <rect",
"         style=\"fill:#0000ff;fill-opacity:0.1;stroke:#000000;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1\"",
"         id=\"rect4002\"",
"         width=\"170\"",
"         height=\"170\"",
"         x=\"0\"",
"         y=\"2.9921875\" />",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:48px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#000000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         x=\"40\"",
"         y=\"42.992188\"",
"         id=\"text4004\"",
"         sodipodi:linespacing=\"100%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan4006\"",
"           x=\"40\"",
"           y=\"42.992188\">Del</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#000000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"60\"",
"         id=\"text4008\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan4010\"",
"           x=\"5\"",
"           y=\"60\">key:Delete</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#ff0000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"75\"",
"         id=\"text4012\"><tspan",
"           sodipodi:role=\"line\"",
"           x=\"5\"",
"           y=\"75\"",
"           id=\"tspan4014\">key:&lt;Primary&gt;Delete</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#008000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"90\"",
"         id=\"text4016\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan4018\"",
"           x=\"5\"",
"           y=\"90\">key:&lt;Shift&gt;Delete</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#0000ff;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"105\"",
"         id=\"text4020\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan4022\"",
"           x=\"5\"",
"           y=\"105\">key:&lt;Alt&gt;Delete</tspan></text>",
"      <text",
"         id=\"text6771\"",
"         y=\"120\"",
"         x=\"5\"",
"         style=\"font-style:normal;font-weight:normal;font-size:12.00000048px;font-family:'Bitstream Vera Sans';fill:#464646;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;-inkscape-font-specification:'Bitstream Vera Sans, Normal';font-stretch:normal;font-variant:normal;text-anchor:start;text-align:start;writing-mode:lr;line-height:125%;\"",
"         xml:space=\"preserve\"",
"         sodipodi:linespacing=\"125%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan6775\">key:&lt;Super&gt;Delete</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#808000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"135\"",
"         id=\"text4024\"",
"         sodipodi:linespacing=\"100%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan4026\"",
"           x=\"5\"",
"           y=\"135\">key:&lt;Primary&gt;&lt;Shift&gt;Delete</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#ff00ff;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"150\"",
"         id=\"text4028\"",
"         sodipodi:linespacing=\"100%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan4030\"",
"           x=\"5\"",
"           y=\"150\">key:&lt;Primary&gt;&lt;Alt&gt;Delete</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#008080;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"165\"",
"         id=\"text4032\"",
"         sodipodi:linespacing=\"100%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan4034\"",
"           x=\"5\"",
"           y=\"165\">key:&lt;Shift&gt;&lt;Alt&gt;Delete</tspan></text>",
"    </g>",
"    <g",
"       transform=\"translate(2140,1670)\"",
"       id=\"g4036\">",
"      <rect",
"         style=\"fill:#0000ff;fill-opacity:0.1;stroke:#000000;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1\"",
"         id=\"rect4038\"",
"         width=\"170\"",
"         height=\"170\"",
"         x=\"0\"",
"         y=\"2.9921875\" />",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:32px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr;text-anchor:start;fill:#000000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         x=\"10\"",
"         y=\"40\"",
"         id=\"text4040\"",
"         sodipodi:linespacing=\"100%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan4042\"",
"           x=\"10\"",
"           y=\"40\">NumLock</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#000000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"60\"",
"         id=\"text4044\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan4046\"",
"           x=\"5\"",
"           y=\"60\">key:NumLock</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#ff0000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"75\"",
"         id=\"text4048\"><tspan",
"           sodipodi:role=\"line\"",
"           x=\"5\"",
"           y=\"75\"",
"           id=\"tspan4050\">key:&lt;Primary&gt;NumLock</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#008000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"90\"",
"         id=\"text4052\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan4054\"",
"           x=\"5\"",
"           y=\"90\">key:&lt;Shift&gt;NumLock</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#0000ff;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"105\"",
"         id=\"text4056\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan4058\"",
"           x=\"5\"",
"           y=\"105\">key:&lt;Alt&gt;NumLock</tspan></text>",
"      <text",
"         id=\"text6813\"",
"         y=\"120\"",
"         x=\"5\"",
"         style=\"font-style:normal;font-weight:normal;font-size:12.00000048px;font-family:'Bitstream Vera Sans';fill:#464646;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;-inkscape-font-specification:'Bitstream Vera Sans, Normal';font-stretch:normal;font-variant:normal;text-anchor:start;text-align:start;writing-mode:lr;line-height:125%;\"",
"         xml:space=\"preserve\"",
"         sodipodi:linespacing=\"125%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan6817\">key:&lt;Super&gt;NumLock</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#808000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"135\"",
"         id=\"text4060\"",
"         sodipodi:linespacing=\"100%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan4062\"",
"           x=\"5\"",
"           y=\"135\">key:&lt;Primary&gt;&lt;Shift&gt;NumLock</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#ff00ff;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"150\"",
"         id=\"text4064\"",
"         sodipodi:linespacing=\"100%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan4066\"",
"           x=\"5\"",
"           y=\"150\">key:&lt;Primary&gt;&lt;Alt&gt;NumLock</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#008080;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"165\"",
"         id=\"text4068\"",
"         sodipodi:linespacing=\"100%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan4070\"",
"           x=\"5\"",
"           y=\"165\">key:&lt;Shift&gt;&lt;Alt&gt;NumLock</tspan></text>",
"    </g>",
"    <g",
"       id=\"g4072\"",
"       transform=\"translate(2680,1670)\">",
"      <rect",
"         y=\"2.9921875\"",
"         x=\"0\"",
"         height=\"170\"",
"         width=\"170\"",
"         id=\"rect4074\"",
"         style=\"fill:#0000ff;fill-opacity:0.1;stroke:#000000;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1\" />",
"      <text",
"         sodipodi:linespacing=\"100%\"",
"         id=\"text4076\"",
"         y=\"40\"",
"         x=\"10\"",
"         style=\"font-size:32px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#000000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"40\"",
"           x=\"10\"",
"           id=\"tspan4078\"",
"           sodipodi:role=\"line\">KP_Minus</tspan></text>",
"      <text",
"         id=\"text4080\"",
"         y=\"60\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#000000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"60\"",
"           x=\"5\"",
"           id=\"tspan4082\"",
"           sodipodi:role=\"line\">key:KP_Subtract</tspan></text>",
"      <text",
"         id=\"text4084\"",
"         y=\"75\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#ff0000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           id=\"tspan4086\"",
"           y=\"75\"",
"           x=\"5\"",
"           sodipodi:role=\"line\">key:&lt;Primary&gt;KP_Subtract</tspan></text>",
"      <text",
"         id=\"text4088\"",
"         y=\"90\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#008000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"90\"",
"           x=\"5\"",
"           id=\"tspan4090\"",
"           sodipodi:role=\"line\">key:&lt;Shift&gt;KP_Subtract</tspan></text>",
"      <text",
"         id=\"text4092\"",
"         y=\"105\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#0000ff;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"105\"",
"           x=\"5\"",
"           id=\"tspan4094\"",
"           sodipodi:role=\"line\">key:&lt;Alt&gt;KP_Subtract</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-style:normal;font-weight:normal;font-size:12.00000048px;font-family:'Bitstream Vera Sans';fill:#464646;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;-inkscape-font-specification:'Bitstream Vera Sans, Normal';font-stretch:normal;font-variant:normal;text-anchor:start;text-align:start;writing-mode:lr;line-height:125%;\"",
"         x=\"5\"",
"         y=\"120\"",
"         id=\"text6831\"",
"         sodipodi:linespacing=\"125%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan6835\">key:&lt;Super&gt;KP_Subtract</tspan></text>",
"      <text",
"         sodipodi:linespacing=\"100%\"",
"         id=\"text4096\"",
"         y=\"135\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#808000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"135\"",
"           x=\"5\"",
"           id=\"tspan4098\"",
"           sodipodi:role=\"line\">key:&lt;Primary&gt;&lt;Shift&gt;KP_Subtract</tspan></text>",
"      <text",
"         sodipodi:linespacing=\"100%\"",
"         id=\"text4100\"",
"         y=\"150\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#ff00ff;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"150\"",
"           x=\"5\"",
"           id=\"tspan4102\"",
"           sodipodi:role=\"line\">key:&lt;Primary&gt;&lt;Alt&gt;KP_Subtract</tspan></text>",
"      <text",
"         sodipodi:linespacing=\"100%\"",
"         id=\"text4104\"",
"         y=\"165\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#008080;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"165\"",
"           x=\"5\"",
"           id=\"tspan4106\"",
"           sodipodi:role=\"line\">key:&lt;Shift&gt;&lt;Alt&gt;KP_Subtract</tspan></text>",
"    </g>",
"    <g",
"       transform=\"translate(2500,1670)\"",
"       id=\"g4108\">",
"      <rect",
"         style=\"fill:#0000ff;fill-opacity:0.1;stroke:#000000;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1\"",
"         id=\"rect4110\"",
"         width=\"170\"",
"         height=\"170\"",
"         x=\"0\"",
"         y=\"2.9921875\" />",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:48px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#000000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         x=\"40\"",
"         y=\"42.992188\"",
"         id=\"text4112\"",
"         sodipodi:linespacing=\"100%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan4114\"",
"           x=\"40\"",
"           y=\"42.992188\">KP_*</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#000000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"60\"",
"         id=\"text4116\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan4118\"",
"           x=\"5\"",
"           y=\"60\">key:KP_Multiply</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#ff0000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"75\"",
"         id=\"text4120\"><tspan",
"           sodipodi:role=\"line\"",
"           x=\"5\"",
"           y=\"75\"",
"           id=\"tspan4122\">key:&lt;Primary&gt;KP_Multiply</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#008000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"90\"",
"         id=\"text4124\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan4126\"",
"           x=\"5\"",
"           y=\"90\">key:&lt;Shift&gt;KP_Multiply</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#0000ff;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"105\"",
"         id=\"text4128\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan4130\"",
"           x=\"5\"",
"           y=\"105\">key:&lt;Alt&gt;KP_Multiply</tspan></text>",
"      <text",
"         id=\"text6825\"",
"         y=\"120\"",
"         x=\"5\"",
"         style=\"font-style:normal;font-weight:normal;font-size:12.00000048px;font-family:'Bitstream Vera Sans';fill:#464646;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;-inkscape-font-specification:'Bitstream Vera Sans, Normal';font-stretch:normal;font-variant:normal;text-anchor:start;text-align:start;writing-mode:lr;line-height:125%;\"",
"         xml:space=\"preserve\"",
"         sodipodi:linespacing=\"125%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan6829\">key:&lt;Super&gt;KP_Multiply</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#808000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"135\"",
"         id=\"text4132\"",
"         sodipodi:linespacing=\"100%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan4134\"",
"           x=\"5\"",
"           y=\"135\">key:&lt;Primary&gt;&lt;Shift&gt;KP_Multiply</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#ff00ff;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"150\"",
"         id=\"text4136\"",
"         sodipodi:linespacing=\"100%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan4138\"",
"           x=\"5\"",
"           y=\"150\">key:&lt;Primary&gt;&lt;Alt&gt;KP_Multiply</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#008080;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"165\"",
"         id=\"text4140\"",
"         sodipodi:linespacing=\"100%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan4142\"",
"           x=\"5\"",
"           y=\"165\">key:&lt;Shift&gt;&lt;Alt&gt;KP_Multiply</tspan></text>",
"    </g>",
"    <g",
"       id=\"g4144\"",
"       transform=\"translate(2320,1670)\">",
"      <rect",
"         y=\"2.9921875\"",
"         x=\"0\"",
"         height=\"170\"",
"         width=\"170\"",
"         id=\"rect4146\"",
"         style=\"fill:#0000ff;fill-opacity:0.1;stroke:#000000;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1\" />",
"      <text",
"         sodipodi:linespacing=\"100%\"",
"         id=\"text4148\"",
"         y=\"42.992188\"",
"         x=\"40\"",
"         style=\"font-size:48px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#000000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"42.992188\"",
"           x=\"40\"",
"           id=\"tspan4150\"",
"           sodipodi:role=\"line\">KP_/</tspan></text>",
"      <text",
"         id=\"text4152\"",
"         y=\"60\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#000000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"60\"",
"           x=\"5\"",
"           id=\"tspan4154\"",
"           sodipodi:role=\"line\">key:KP_Divide</tspan></text>",
"      <text",
"         id=\"text4156\"",
"         y=\"75\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#ff0000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           id=\"tspan4158\"",
"           y=\"75\"",
"           x=\"5\"",
"           sodipodi:role=\"line\">key:&lt;Primary&gt;KP_Divide</tspan></text>",
"      <text",
"         id=\"text4160\"",
"         y=\"90\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#008000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"90\"",
"           x=\"5\"",
"           id=\"tspan4162\"",
"           sodipodi:role=\"line\">key:&lt;Shift&gt;KP_Divide</tspan></text>",
"      <text",
"         id=\"text4164\"",
"         y=\"105\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#0000ff;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"105\"",
"           x=\"5\"",
"           id=\"tspan4166\"",
"           sodipodi:role=\"line\">key:&lt;Alt&gt;KP_Divide</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-style:normal;font-weight:normal;font-size:12.00000048px;font-family:'Bitstream Vera Sans';fill:#464646;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;-inkscape-font-specification:'Bitstream Vera Sans, Normal';font-stretch:normal;font-variant:normal;text-anchor:start;text-align:start;writing-mode:lr;line-height:125%;\"",
"         x=\"5\"",
"         y=\"120\"",
"         id=\"text6819\"",
"         sodipodi:linespacing=\"125%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan6823\">key:&lt;Super&gt;KP_Divide</tspan></text>",
"      <text",
"         sodipodi:linespacing=\"100%\"",
"         id=\"text4168\"",
"         y=\"135\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#808000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"135\"",
"           x=\"5\"",
"           id=\"tspan4170\"",
"           sodipodi:role=\"line\">key:&lt;Primary&gt;&lt;Shift&gt;KP_Divide</tspan></text>",
"      <text",
"         sodipodi:linespacing=\"100%\"",
"         id=\"text4172\"",
"         y=\"150\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#ff00ff;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"150\"",
"           x=\"5\"",
"           id=\"tspan4174\"",
"           sodipodi:role=\"line\">key:&lt;Primary&gt;&lt;Alt&gt;KP_Divide</tspan></text>",
"      <text",
"         sodipodi:linespacing=\"100%\"",
"         id=\"text4176\"",
"         y=\"165\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#008080;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"165\"",
"           x=\"5\"",
"           id=\"tspan4178\"",
"           sodipodi:role=\"line\">key:&lt;Shift&gt;&lt;Alt&gt;KP_Divide</tspan></text>",
"    </g>",
"    <g",
"       id=\"g4180\"",
"       transform=\"translate(2500,1850)\">",
"      <rect",
"         y=\"2.9921875\"",
"         x=\"0\"",
"         height=\"170\"",
"         width=\"170\"",
"         id=\"rect4182\"",
"         style=\"fill:#0000ff;fill-opacity:0.1;stroke:#000000;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1\" />",
"      <text",
"         sodipodi:linespacing=\"100%\"",
"         id=\"text4184\"",
"         y=\"42.992188\"",
"         x=\"40\"",
"         style=\"font-size:48px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#000000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"42.992188\"",
"           x=\"40\"",
"           id=\"tspan4186\"",
"           sodipodi:role=\"line\">KP_9</tspan></text>",
"      <text",
"         id=\"text4188\"",
"         y=\"60\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#000000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"60\"",
"           x=\"5\"",
"           id=\"tspan4190\"",
"           sodipodi:role=\"line\">key:KP_9</tspan></text>",
"      <text",
"         id=\"text4192\"",
"         y=\"75\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#ff0000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           id=\"tspan4194\"",
"           y=\"75\"",
"           x=\"5\"",
"           sodipodi:role=\"line\">key:&lt;Primary&gt;KP_9</tspan></text>",
"      <text",
"         id=\"text4196\"",
"         y=\"90\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#008000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"90\"",
"           x=\"5\"",
"           id=\"tspan4198\"",
"           sodipodi:role=\"line\">key:&lt;Shift&gt;KP_9</tspan></text>",
"      <text",
"         id=\"text4200\"",
"         y=\"105\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#0000ff;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"105\"",
"           x=\"5\"",
"           id=\"tspan4202\"",
"           sodipodi:role=\"line\">key:&lt;Alt&gt;KP_9</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-style:normal;font-weight:normal;font-size:12.00000048px;font-family:'Bitstream Vera Sans';fill:#464646;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;-inkscape-font-specification:'Bitstream Vera Sans, Normal';font-stretch:normal;font-variant:normal;text-anchor:start;text-align:start;writing-mode:lr;line-height:125%;\"",
"         x=\"5\"",
"         y=\"120\"",
"         id=\"text6849\"",
"         sodipodi:linespacing=\"125%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan6853\">key:&lt;Super&gt;KP_9</tspan></text>",
"      <text",
"         sodipodi:linespacing=\"100%\"",
"         id=\"text4204\"",
"         y=\"135\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#808000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"135\"",
"           x=\"5\"",
"           id=\"tspan4206\"",
"           sodipodi:role=\"line\">key:&lt;Primary&gt;&lt;Shift&gt;KP_9</tspan></text>",
"      <text",
"         sodipodi:linespacing=\"100%\"",
"         id=\"text4208\"",
"         y=\"150\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#ff00ff;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"150\"",
"           x=\"5\"",
"           id=\"tspan4210\"",
"           sodipodi:role=\"line\">key:&lt;Primary&gt;&lt;Alt&gt;KP_9</tspan></text>",
"      <text",
"         sodipodi:linespacing=\"100%\"",
"         id=\"text4212\"",
"         y=\"165\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#008080;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"165\"",
"           x=\"5\"",
"           id=\"tspan4214\"",
"           sodipodi:role=\"line\">key:&lt;Shift&gt;&lt;Alt&gt;KP_9</tspan></text>",
"    </g>",
"    <g",
"       transform=\"translate(2320,1850)\"",
"       id=\"g4216\">",
"      <rect",
"         style=\"fill:#0000ff;fill-opacity:0.1;stroke:#000000;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1\"",
"         id=\"rect4218\"",
"         width=\"170\"",
"         height=\"170\"",
"         x=\"0\"",
"         y=\"2.9921875\" />",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:48px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#000000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         x=\"40\"",
"         y=\"42.992188\"",
"         id=\"text4220\"",
"         sodipodi:linespacing=\"100%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan4222\"",
"           x=\"40\"",
"           y=\"42.992188\">KP_8</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#000000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"60\"",
"         id=\"text4224\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan4226\"",
"           x=\"5\"",
"           y=\"60\">key:KP_8</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#ff0000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"75\"",
"         id=\"text4228\"><tspan",
"           sodipodi:role=\"line\"",
"           x=\"5\"",
"           y=\"75\"",
"           id=\"tspan4230\">key:&lt;Primary&gt;KP_8</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#008000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"90\"",
"         id=\"text4232\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan4234\"",
"           x=\"5\"",
"           y=\"90\">key:&lt;Shift&gt;KP_8</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#0000ff;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"105\"",
"         id=\"text4236\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan4238\"",
"           x=\"5\"",
"           y=\"105\">key:&lt;Alt&gt;KP_8</tspan></text>",
"      <text",
"         id=\"text6843\"",
"         y=\"120\"",
"         x=\"5\"",
"         style=\"font-style:normal;font-weight:normal;font-size:12.00000048px;font-family:'Bitstream Vera Sans';fill:#464646;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;-inkscape-font-specification:'Bitstream Vera Sans, Normal';font-stretch:normal;font-variant:normal;text-anchor:start;text-align:start;writing-mode:lr;line-height:125%;\"",
"         xml:space=\"preserve\"",
"         sodipodi:linespacing=\"125%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan6847\">key:&lt;Super&gt;KP_8</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#808000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"135\"",
"         id=\"text4240\"",
"         sodipodi:linespacing=\"100%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan4242\"",
"           x=\"5\"",
"           y=\"135\">key:&lt;Primary&gt;&lt;Shift&gt;KP_8</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#ff00ff;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"150\"",
"         id=\"text4244\"",
"         sodipodi:linespacing=\"100%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan4246\"",
"           x=\"5\"",
"           y=\"150\">key:&lt;Primary&gt;&lt;Alt&gt;KP_8</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#008080;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"165\"",
"         id=\"text4248\"",
"         sodipodi:linespacing=\"100%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan4250\"",
"           x=\"5\"",
"           y=\"165\">key:&lt;Shift&gt;&lt;Alt&gt;KP_8</tspan></text>",
"    </g>",
"    <g",
"       id=\"g4252\"",
"       transform=\"translate(2140,1850)\">",
"      <rect",
"         y=\"2.9921875\"",
"         x=\"0\"",
"         height=\"170\"",
"         width=\"170\"",
"         id=\"rect4254\"",
"         style=\"fill:#0000ff;fill-opacity:0.1;stroke:#000000;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1\" />",
"      <text",
"         sodipodi:linespacing=\"100%\"",
"         id=\"text4256\"",
"         y=\"42.992188\"",
"         x=\"40\"",
"         style=\"font-size:48px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#000000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"42.992188\"",
"           x=\"40\"",
"           id=\"tspan4258\"",
"           sodipodi:role=\"line\">KP_7</tspan></text>",
"      <text",
"         id=\"text4260\"",
"         y=\"60\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#000000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"60\"",
"           x=\"5\"",
"           id=\"tspan4262\"",
"           sodipodi:role=\"line\">key:KP_7</tspan></text>",
"      <text",
"         id=\"text4264\"",
"         y=\"75\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#ff0000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           id=\"tspan4266\"",
"           y=\"75\"",
"           x=\"5\"",
"           sodipodi:role=\"line\">key:&lt;Primary&gt;KP_7</tspan></text>",
"      <text",
"         id=\"text4268\"",
"         y=\"90\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#008000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"90\"",
"           x=\"5\"",
"           id=\"tspan4270\"",
"           sodipodi:role=\"line\">key:&lt;Shift&gt;KP_7</tspan></text>",
"      <text",
"         id=\"text4272\"",
"         y=\"105\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#0000ff;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"105\"",
"           x=\"5\"",
"           id=\"tspan4274\"",
"           sodipodi:role=\"line\">key:&lt;Alt&gt;KP_7</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-style:normal;font-weight:normal;font-size:12.00000048px;font-family:'Bitstream Vera Sans';fill:#464646;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;-inkscape-font-specification:'Bitstream Vera Sans, Normal';font-stretch:normal;font-variant:normal;text-anchor:start;text-align:start;writing-mode:lr;line-height:125%;\"",
"         x=\"5\"",
"         y=\"120\"",
"         id=\"text6837\"",
"         sodipodi:linespacing=\"125%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan6841\">key:&lt;Super&gt;KP_7</tspan></text>",
"      <text",
"         sodipodi:linespacing=\"100%\"",
"         id=\"text4276\"",
"         y=\"135\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#808000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"135\"",
"           x=\"5\"",
"           id=\"tspan4278\"",
"           sodipodi:role=\"line\">key:&lt;Primary&gt;&lt;Shift&gt;KP_7</tspan></text>",
"      <text",
"         sodipodi:linespacing=\"100%\"",
"         id=\"text4280\"",
"         y=\"150\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#ff00ff;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"150\"",
"           x=\"5\"",
"           id=\"tspan4282\"",
"           sodipodi:role=\"line\">key:&lt;Primary&gt;&lt;Alt&gt;KP_7</tspan></text>",
"      <text",
"         sodipodi:linespacing=\"100%\"",
"         id=\"text4284\"",
"         y=\"165\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#008080;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"165\"",
"           x=\"5\"",
"           id=\"tspan4286\"",
"           sodipodi:role=\"line\">key:&lt;Shift&gt;&lt;Alt&gt;KP_7</tspan></text>",
"    </g>",
"    <g",
"       transform=\"translate(2500,2210)\"",
"       id=\"g4288\">",
"      <rect",
"         style=\"fill:#0000ff;fill-opacity:0.1;stroke:#000000;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1\"",
"         id=\"rect4290\"",
"         width=\"170\"",
"         height=\"170\"",
"         x=\"0\"",
"         y=\"2.9921875\" />",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:48px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#000000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         x=\"40\"",
"         y=\"42.992188\"",
"         id=\"text4292\"",
"         sodipodi:linespacing=\"100%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan4294\"",
"           x=\"40\"",
"           y=\"42.992188\">KP_3</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#000000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"60\"",
"         id=\"text4296\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan4298\"",
"           x=\"5\"",
"           y=\"60\">key:KP_3</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#ff0000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"75\"",
"         id=\"text4300\"><tspan",
"           sodipodi:role=\"line\"",
"           x=\"5\"",
"           y=\"75\"",
"           id=\"tspan4302\">key:&lt;Primary&gt;KP_3</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#008000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"90\"",
"         id=\"text4304\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan4306\"",
"           x=\"5\"",
"           y=\"90\">key:&lt;Shift&gt;KP_3</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#0000ff;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"105\"",
"         id=\"text4308\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan4310\"",
"           x=\"5\"",
"           y=\"105\">key:&lt;Alt&gt;KP_3</tspan></text>",
"      <text",
"         id=\"text6891\"",
"         y=\"120\"",
"         x=\"5\"",
"         style=\"font-style:normal;font-weight:normal;font-size:12.00000048px;font-family:'Bitstream Vera Sans';fill:#464646;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;-inkscape-font-specification:'Bitstream Vera Sans, Normal';font-stretch:normal;font-variant:normal;text-anchor:start;text-align:start;writing-mode:lr;line-height:125%;\"",
"         xml:space=\"preserve\"",
"         sodipodi:linespacing=\"125%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan6895\">key:&lt;Super&gt;KP_3</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#808000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"135\"",
"         id=\"text4312\"",
"         sodipodi:linespacing=\"100%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan4314\"",
"           x=\"5\"",
"           y=\"135\">key:&lt;Primary&gt;&lt;Shift&gt;KP_3</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#ff00ff;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"150\"",
"         id=\"text4316\"",
"         sodipodi:linespacing=\"100%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan4318\"",
"           x=\"5\"",
"           y=\"150\">key:&lt;Primary&gt;&lt;Alt&gt;KP_3</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#008080;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"165\"",
"         id=\"text4320\"",
"         sodipodi:linespacing=\"100%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan4322\"",
"           x=\"5\"",
"           y=\"165\">key:&lt;Shift&gt;&lt;Alt&gt;KP_3</tspan></text>",
"    </g>",
"    <g",
"       id=\"g4324\"",
"       transform=\"translate(2320,2210)\">",
"      <rect",
"         y=\"2.9921875\"",
"         x=\"0\"",
"         height=\"170\"",
"         width=\"170\"",
"         id=\"rect4326\"",
"         style=\"fill:#0000ff;fill-opacity:0.1;stroke:#000000;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1\" />",
"      <text",
"         sodipodi:linespacing=\"100%\"",
"         id=\"text4328\"",
"         y=\"42.992188\"",
"         x=\"40\"",
"         style=\"font-size:48px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#000000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"42.992188\"",
"           x=\"40\"",
"           id=\"tspan4330\"",
"           sodipodi:role=\"line\">KP_2</tspan></text>",
"      <text",
"         id=\"text4332\"",
"         y=\"60\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#000000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"60\"",
"           x=\"5\"",
"           id=\"tspan4334\"",
"           sodipodi:role=\"line\">key:KP_2</tspan></text>",
"      <text",
"         id=\"text4336\"",
"         y=\"75\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#ff0000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           id=\"tspan4338\"",
"           y=\"75\"",
"           x=\"5\"",
"           sodipodi:role=\"line\">key:&lt;Primary&gt;KP_2</tspan></text>",
"      <text",
"         id=\"text4340\"",
"         y=\"90\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#008000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"90\"",
"           x=\"5\"",
"           id=\"tspan4342\"",
"           sodipodi:role=\"line\">key:&lt;Shift&gt;KP_2</tspan></text>",
"      <text",
"         id=\"text4344\"",
"         y=\"105\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#0000ff;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"105\"",
"           x=\"5\"",
"           id=\"tspan4346\"",
"           sodipodi:role=\"line\">key:&lt;Alt&gt;KP_2</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-style:normal;font-weight:normal;font-size:12.00000048px;font-family:'Bitstream Vera Sans';fill:#464646;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;-inkscape-font-specification:'Bitstream Vera Sans, Normal';font-stretch:normal;font-variant:normal;text-anchor:start;text-align:start;writing-mode:lr;line-height:125%;\"",
"         x=\"5\"",
"         y=\"120\"",
"         id=\"text6885\"",
"         sodipodi:linespacing=\"125%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan6889\">key:&lt;Super&gt;KP_2</tspan></text>",
"      <text",
"         sodipodi:linespacing=\"100%\"",
"         id=\"text4348\"",
"         y=\"135\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#808000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"135\"",
"           x=\"5\"",
"           id=\"tspan4350\"",
"           sodipodi:role=\"line\">key:&lt;Primary&gt;&lt;Shift&gt;KP_2</tspan></text>",
"      <text",
"         sodipodi:linespacing=\"100%\"",
"         id=\"text4352\"",
"         y=\"150\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#ff00ff;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"150\"",
"           x=\"5\"",
"           id=\"tspan4354\"",
"           sodipodi:role=\"line\">key:&lt;Primary&gt;&lt;Alt&gt;KP_2</tspan></text>",
"      <text",
"         sodipodi:linespacing=\"100%\"",
"         id=\"text4356\"",
"         y=\"165\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#008080;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"165\"",
"           x=\"5\"",
"           id=\"tspan4358\"",
"           sodipodi:role=\"line\">key:&lt;Shift&gt;&lt;Alt&gt;KP_2</tspan></text>",
"    </g>",
"    <g",
"       transform=\"translate(2500,2030)\"",
"       id=\"g4360\">",
"      <rect",
"         style=\"fill:#0000ff;fill-opacity:0.1;stroke:#000000;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1\"",
"         id=\"rect4362\"",
"         width=\"170\"",
"         height=\"170\"",
"         x=\"0\"",
"         y=\"2.9921875\" />",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:48px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#000000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         x=\"40\"",
"         y=\"42.992188\"",
"         id=\"text4364\"",
"         sodipodi:linespacing=\"100%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan4366\"",
"           x=\"40\"",
"           y=\"42.992188\">KP_6</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#000000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"60\"",
"         id=\"text4368\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan4370\"",
"           x=\"5\"",
"           y=\"60\">key:KP_6</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#ff0000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"75\"",
"         id=\"text4372\"><tspan",
"           sodipodi:role=\"line\"",
"           x=\"5\"",
"           y=\"75\"",
"           id=\"tspan4374\">key:&lt;Primary&gt;KP_6</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#008000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"90\"",
"         id=\"text4376\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan4378\"",
"           x=\"5\"",
"           y=\"90\">key:&lt;Shift&gt;KP_6</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#0000ff;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"105\"",
"         id=\"text4380\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan4382\"",
"           x=\"5\"",
"           y=\"105\">key:&lt;Alt&gt;KP_6</tspan></text>",
"      <text",
"         id=\"text6873\"",
"         y=\"120\"",
"         x=\"5\"",
"         style=\"font-style:normal;font-weight:normal;font-size:12.00000048px;font-family:'Bitstream Vera Sans';fill:#464646;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;-inkscape-font-specification:'Bitstream Vera Sans, Normal';font-stretch:normal;font-variant:normal;text-anchor:start;text-align:start;writing-mode:lr;line-height:125%;\"",
"         xml:space=\"preserve\"",
"         sodipodi:linespacing=\"125%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan6877\">key:&lt;Super&gt;KP_6</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#808000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"135\"",
"         id=\"text4384\"",
"         sodipodi:linespacing=\"100%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan4386\"",
"           x=\"5\"",
"           y=\"135\">key:&lt;Primary&gt;&lt;Shift&gt;KP_6</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#ff00ff;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"150\"",
"         id=\"text4388\"",
"         sodipodi:linespacing=\"100%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan4390\"",
"           x=\"5\"",
"           y=\"150\">key:&lt;Primary&gt;&lt;Alt&gt;KP_6</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#008080;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"165\"",
"         id=\"text4392\"",
"         sodipodi:linespacing=\"100%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan4394\"",
"           x=\"5\"",
"           y=\"165\">key:&lt;Shift&gt;&lt;Alt&gt;KP_6</tspan></text>",
"    </g>",
"    <g",
"       id=\"g4396\"",
"       transform=\"translate(2320,2030)\">",
"      <rect",
"         y=\"2.9921875\"",
"         x=\"0\"",
"         height=\"170\"",
"         width=\"170\"",
"         id=\"rect4398\"",
"         style=\"fill:#0000ff;fill-opacity:0.1;stroke:#000000;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1\" />",
"      <text",
"         sodipodi:linespacing=\"100%\"",
"         id=\"text4400\"",
"         y=\"42.992188\"",
"         x=\"40\"",
"         style=\"font-size:48px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#000000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"42.992188\"",
"           x=\"40\"",
"           id=\"tspan4402\"",
"           sodipodi:role=\"line\">KP_5</tspan></text>",
"      <text",
"         id=\"text4404\"",
"         y=\"60\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#000000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"60\"",
"           x=\"5\"",
"           id=\"tspan4406\"",
"           sodipodi:role=\"line\">key:KP_5</tspan></text>",
"      <text",
"         id=\"text4408\"",
"         y=\"75\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#ff0000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           id=\"tspan4410\"",
"           y=\"75\"",
"           x=\"5\"",
"           sodipodi:role=\"line\">key:&lt;Primary&gt;KP_5</tspan></text>",
"      <text",
"         id=\"text4412\"",
"         y=\"90\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#008000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"90\"",
"           x=\"5\"",
"           id=\"tspan4414\"",
"           sodipodi:role=\"line\">key:&lt;Shift&gt;KP_5</tspan></text>",
"      <text",
"         id=\"text4416\"",
"         y=\"105\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#0000ff;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"105\"",
"           x=\"5\"",
"           id=\"tspan4418\"",
"           sodipodi:role=\"line\">key:&lt;Alt&gt;KP_5</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-style:normal;font-weight:normal;font-size:12.00000048px;font-family:'Bitstream Vera Sans';fill:#464646;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;-inkscape-font-specification:'Bitstream Vera Sans, Normal';font-stretch:normal;font-variant:normal;text-anchor:start;text-align:start;writing-mode:lr;line-height:125%;\"",
"         x=\"5\"",
"         y=\"120\"",
"         id=\"text6867\"",
"         sodipodi:linespacing=\"125%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan6871\">key:&lt;Super&gt;KP_5</tspan></text>",
"      <text",
"         sodipodi:linespacing=\"100%\"",
"         id=\"text4420\"",
"         y=\"135\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#808000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"135\"",
"           x=\"5\"",
"           id=\"tspan4422\"",
"           sodipodi:role=\"line\">key:&lt;Primary&gt;&lt;Shift&gt;KP_5</tspan></text>",
"      <text",
"         sodipodi:linespacing=\"100%\"",
"         id=\"text4424\"",
"         y=\"150\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#ff00ff;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"150\"",
"           x=\"5\"",
"           id=\"tspan4426\"",
"           sodipodi:role=\"line\">key:&lt;Primary&gt;&lt;Alt&gt;KP_5</tspan></text>",
"      <text",
"         sodipodi:linespacing=\"100%\"",
"         id=\"text4428\"",
"         y=\"165\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#008080;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"165\"",
"           x=\"5\"",
"           id=\"tspan4430\"",
"           sodipodi:role=\"line\">key:&lt;Shift&gt;&lt;Alt&gt;KP_5</tspan></text>",
"    </g>",
"    <g",
"       transform=\"translate(2140,2210)\"",
"       id=\"g4432\">",
"      <rect",
"         style=\"fill:#0000ff;fill-opacity:0.1;stroke:#000000;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1\"",
"         id=\"rect4434\"",
"         width=\"170\"",
"         height=\"170\"",
"         x=\"0\"",
"         y=\"2.9921875\" />",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:48px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#000000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         x=\"40\"",
"         y=\"42.992188\"",
"         id=\"text4436\"",
"         sodipodi:linespacing=\"100%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan4438\"",
"           x=\"40\"",
"           y=\"42.992188\">KP_1</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#000000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"60\"",
"         id=\"text4440\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan4442\"",
"           x=\"5\"",
"           y=\"60\">key:KP_1</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#ff0000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"75\"",
"         id=\"text4444\"><tspan",
"           sodipodi:role=\"line\"",
"           x=\"5\"",
"           y=\"75\"",
"           id=\"tspan4446\">key:&lt;Primary&gt;KP_1</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#008000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"90\"",
"         id=\"text4448\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan4450\"",
"           x=\"5\"",
"           y=\"90\">key:&lt;Shift&gt;KP_1</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#0000ff;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"105\"",
"         id=\"text4452\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan4454\"",
"           x=\"5\"",
"           y=\"105\">key:&lt;Alt&gt;KP_1</tspan></text>",
"      <text",
"         id=\"text6879\"",
"         y=\"120\"",
"         x=\"5\"",
"         style=\"font-style:normal;font-weight:normal;font-size:12.00000048px;font-family:'Bitstream Vera Sans';fill:#464646;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;-inkscape-font-specification:'Bitstream Vera Sans, Normal';font-stretch:normal;font-variant:normal;text-anchor:start;text-align:start;writing-mode:lr;line-height:125%;\"",
"         xml:space=\"preserve\"",
"         sodipodi:linespacing=\"125%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan6883\">key:&lt;Super&gt;KP_1</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#808000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"135\"",
"         id=\"text4456\"",
"         sodipodi:linespacing=\"100%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan4458\"",
"           x=\"5\"",
"           y=\"135\">key:&lt;Primary&gt;&lt;Shift&gt;KP_1</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#ff00ff;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"150\"",
"         id=\"text4460\"",
"         sodipodi:linespacing=\"100%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan4462\"",
"           x=\"5\"",
"           y=\"150\">key:&lt;Primary&gt;&lt;Alt&gt;KP_1</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#008080;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"165\"",
"         id=\"text4464\"",
"         sodipodi:linespacing=\"100%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan4466\"",
"           x=\"5\"",
"           y=\"165\">key:&lt;Shift&gt;&lt;Alt&gt;KP_1</tspan></text>",
"    </g>",
"    <g",
"       id=\"g4468\"",
"       transform=\"translate(2140,2030)\">",
"      <rect",
"         y=\"2.9921875\"",
"         x=\"0\"",
"         height=\"170\"",
"         width=\"170\"",
"         id=\"rect4470\"",
"         style=\"fill:#0000ff;fill-opacity:0.1;stroke:#000000;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1\" />",
"      <text",
"         sodipodi:linespacing=\"100%\"",
"         id=\"text4472\"",
"         y=\"42.992188\"",
"         x=\"40\"",
"         style=\"font-size:48px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#000000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"42.992188\"",
"           x=\"40\"",
"           id=\"tspan4474\"",
"           sodipodi:role=\"line\">KP_4</tspan></text>",
"      <text",
"         id=\"text4476\"",
"         y=\"60\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#000000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"60\"",
"           x=\"5\"",
"           id=\"tspan4478\"",
"           sodipodi:role=\"line\">key:KP_4</tspan></text>",
"      <text",
"         id=\"text4480\"",
"         y=\"75\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#ff0000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           id=\"tspan4482\"",
"           y=\"75\"",
"           x=\"5\"",
"           sodipodi:role=\"line\">key:&lt;Primary&gt;KP_4</tspan></text>",
"      <text",
"         id=\"text4484\"",
"         y=\"90\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#008000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"90\"",
"           x=\"5\"",
"           id=\"tspan4486\"",
"           sodipodi:role=\"line\">key:&lt;Shift&gt;KP_4</tspan></text>",
"      <text",
"         id=\"text4488\"",
"         y=\"105\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#0000ff;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"105\"",
"           x=\"5\"",
"           id=\"tspan4490\"",
"           sodipodi:role=\"line\">key:&lt;Alt&gt;KP_4</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-style:normal;font-weight:normal;font-size:12.00000048px;font-family:'Bitstream Vera Sans';fill:#464646;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;-inkscape-font-specification:'Bitstream Vera Sans, Normal';font-stretch:normal;font-variant:normal;text-anchor:start;text-align:start;writing-mode:lr;line-height:125%;\"",
"         x=\"5\"",
"         y=\"120\"",
"         id=\"text6861\"",
"         sodipodi:linespacing=\"125%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan6865\">key:&lt;Super&gt;KP_4</tspan></text>",
"      <text",
"         sodipodi:linespacing=\"100%\"",
"         id=\"text4492\"",
"         y=\"135\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#808000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"135\"",
"           x=\"5\"",
"           id=\"tspan4494\"",
"           sodipodi:role=\"line\">key:&lt;Primary&gt;&lt;Shift&gt;KP_4</tspan></text>",
"      <text",
"         sodipodi:linespacing=\"100%\"",
"         id=\"text4496\"",
"         y=\"150\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#ff00ff;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"150\"",
"           x=\"5\"",
"           id=\"tspan4498\"",
"           sodipodi:role=\"line\">key:&lt;Primary&gt;&lt;Alt&gt;KP_4</tspan></text>",
"      <text",
"         sodipodi:linespacing=\"100%\"",
"         id=\"text4500\"",
"         y=\"165\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#008080;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"165\"",
"           x=\"5\"",
"           id=\"tspan4502\"",
"           sodipodi:role=\"line\">key:&lt;Shift&gt;&lt;Alt&gt;KP_4</tspan></text>",
"    </g>",
"    <g",
"       id=\"g4504\"",
"       transform=\"translate(2140,2390)\">",
"      <rect",
"         y=\"2.9921875\"",
"         x=\"0\"",
"         height=\"170\"",
"         width=\"350\"",
"         id=\"rect4506\"",
"         style=\"fill:#0000ff;fill-opacity:0.1;stroke:#000000;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1\" />",
"      <text",
"         sodipodi:linespacing=\"100%\"",
"         id=\"text4508\"",
"         y=\"42.992188\"",
"         x=\"40\"",
"         style=\"font-size:48px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#000000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"42.992188\"",
"           x=\"40\"",
"           id=\"tspan4510\"",
"           sodipodi:role=\"line\">KP_0</tspan></text>",
"      <text",
"         id=\"text4512\"",
"         y=\"60\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#000000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"60\"",
"           x=\"5\"",
"           id=\"tspan4514\"",
"           sodipodi:role=\"line\">key:KP_0</tspan></text>",
"      <text",
"         id=\"text4516\"",
"         y=\"75\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#ff0000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           id=\"tspan4518\"",
"           y=\"75\"",
"           x=\"5\"",
"           sodipodi:role=\"line\">key:&lt;Primary&gt;KP_0</tspan></text>",
"      <text",
"         id=\"text4520\"",
"         y=\"90\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#008000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"90\"",
"           x=\"5\"",
"           id=\"tspan4522\"",
"           sodipodi:role=\"line\">key:&lt;Shift&gt;KP_0</tspan></text>",
"      <text",
"         id=\"text4524\"",
"         y=\"105\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#0000ff;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"105\"",
"           x=\"5\"",
"           id=\"tspan4526\"",
"           sodipodi:role=\"line\">key:&lt;Alt&gt;KP_0</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-style:normal;font-weight:normal;font-size:12.00000048px;font-family:'Bitstream Vera Sans';fill:#464646;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;-inkscape-font-specification:'Bitstream Vera Sans, Normal';font-stretch:normal;font-variant:normal;text-anchor:start;text-align:start;writing-mode:lr;line-height:125%;\"",
"         x=\"5\"",
"         y=\"120\"",
"         id=\"text6903\"",
"         sodipodi:linespacing=\"125%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan6907\">key:&lt;Super&gt;KP_0</tspan></text>",
"      <text",
"         sodipodi:linespacing=\"100%\"",
"         id=\"text4528\"",
"         y=\"135\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#808000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"135\"",
"           x=\"5\"",
"           id=\"tspan4530\"",
"           sodipodi:role=\"line\">key:&lt;Primary&gt;&lt;Shift&gt;KP_0</tspan></text>",
"      <text",
"         sodipodi:linespacing=\"100%\"",
"         id=\"text4532\"",
"         y=\"150\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#ff00ff;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"150\"",
"           x=\"5\"",
"           id=\"tspan4534\"",
"           sodipodi:role=\"line\">key:&lt;Primary&gt;&lt;Alt&gt;KP_0</tspan></text>",
"      <text",
"         sodipodi:linespacing=\"100%\"",
"         id=\"text4536\"",
"         y=\"165\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#008080;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"165\"",
"           x=\"5\"",
"           id=\"tspan4538\"",
"           sodipodi:role=\"line\">key:&lt;Shift&gt;&lt;Alt&gt;KP_0</tspan></text>",
"    </g>",
"    <g",
"       transform=\"translate(2500,2390)\"",
"       id=\"g4540\">",
"      <rect",
"         style=\"fill:#0000ff;fill-opacity:0.1;stroke:#000000;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1\"",
"         id=\"rect4542\"",
"         width=\"170\"",
"         height=\"170\"",
"         x=\"0\"",
"         y=\"2.9921875\" />",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:32px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr;text-anchor:start;fill:#000000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         x=\"10\"",
"         y=\"40\"",
"         id=\"text4544\"",
"         sodipodi:linespacing=\"100%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan4546\"",
"           x=\"10\"",
"           y=\"40\">KP_Delete</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#000000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"60\"",
"         id=\"text4548\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan4550\"",
"           x=\"5\"",
"           y=\"60\">key:KP_Delete</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#ff0000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"75\"",
"         id=\"text4552\"><tspan",
"           sodipodi:role=\"line\"",
"           x=\"5\"",
"           y=\"75\"",
"           id=\"tspan4554\">key:&lt;Primary&gt;KP_Delete</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#008000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"90\"",
"         id=\"text4556\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan4558\"",
"           x=\"5\"",
"           y=\"90\">key:&lt;Shift&gt;KP_Delete</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#0000ff;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"105\"",
"         id=\"text4560\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan4562\"",
"           x=\"5\"",
"           y=\"105\">key:&lt;Alt&gt;KP_Delete</tspan></text>",
"      <text",
"         id=\"text6909\"",
"         y=\"120\"",
"         x=\"5\"",
"         style=\"font-style:normal;font-weight:normal;font-size:12.00000048px;font-family:'Bitstream Vera Sans';fill:#464646;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;-inkscape-font-specification:'Bitstream Vera Sans, Normal';font-stretch:normal;font-variant:normal;text-anchor:start;text-align:start;writing-mode:lr;line-height:125%;\"",
"         xml:space=\"preserve\"",
"         sodipodi:linespacing=\"125%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan6913\">key:&lt;Super&gt;KP_Delete</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#808000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"135\"",
"         id=\"text4564\"",
"         sodipodi:linespacing=\"100%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan4566\"",
"           x=\"5\"",
"           y=\"135\">key:&lt;Primary&gt;&lt;Shift&gt;KP_Delete</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#ff00ff;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"150\"",
"         id=\"text4568\"",
"         sodipodi:linespacing=\"100%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan4570\"",
"           x=\"5\"",
"           y=\"150\">key:&lt;Primary&gt;&lt;Alt&gt;KP_Delete</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#008080;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"165\"",
"         id=\"text4572\"",
"         sodipodi:linespacing=\"100%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan4574\"",
"           x=\"5\"",
"           y=\"165\">key:&lt;Shift&gt;&lt;Alt&gt;KP_Delete</tspan></text>",
"    </g>",
"    <g",
"       id=\"g4576\"",
"       transform=\"translate(2680,2210)\">",
"      <rect",
"         y=\"2.9921875\"",
"         x=\"0\"",
"         height=\"350\"",
"         width=\"170\"",
"         id=\"rect4578\"",
"         style=\"fill:#0000ff;fill-opacity:0.1;stroke:#000000;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1\" />",
"      <text",
"         sodipodi:linespacing=\"100%\"",
"         id=\"text4580\"",
"         y=\"40\"",
"         x=\"10\"",
"         style=\"font-size:32px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr;text-anchor:start;fill:#000000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"40\"",
"           x=\"10\"",
"           id=\"tspan4582\"",
"           sodipodi:role=\"line\">KP_Enter</tspan></text>",
"      <text",
"         id=\"text4584\"",
"         y=\"60\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#000000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"60\"",
"           x=\"5\"",
"           id=\"tspan4586\"",
"           sodipodi:role=\"line\">key:KP_Enter</tspan></text>",
"      <text",
"         id=\"text4588\"",
"         y=\"75\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#ff0000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           id=\"tspan4590\"",
"           y=\"75\"",
"           x=\"5\"",
"           sodipodi:role=\"line\">key:&lt;Primary&gt;KP_Enter</tspan></text>",
"      <text",
"         id=\"text4592\"",
"         y=\"90\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#008000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"90\"",
"           x=\"5\"",
"           id=\"tspan4594\"",
"           sodipodi:role=\"line\">key:&lt;Shift&gt;KP_Enter</tspan></text>",
"      <text",
"         id=\"text4596\"",
"         y=\"105\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#0000ff;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"105\"",
"           x=\"5\"",
"           id=\"tspan4598\"",
"           sodipodi:role=\"line\">key:&lt;Alt&gt;KP_Enter</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-style:normal;font-weight:normal;font-size:12.00000048px;font-family:'Bitstream Vera Sans';fill:#464646;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;-inkscape-font-specification:'Bitstream Vera Sans, Normal';font-stretch:normal;font-variant:normal;text-anchor:start;text-align:start;writing-mode:lr;line-height:125%;\"",
"         x=\"5\"",
"         y=\"120\"",
"         id=\"text6897\"",
"         sodipodi:linespacing=\"125%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan6901\">key:&lt;Super&gt;KP_Enter</tspan></text>",
"      <text",
"         sodipodi:linespacing=\"100%\"",
"         id=\"text4600\"",
"         y=\"135\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#808000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"135\"",
"           x=\"5\"",
"           id=\"tspan4602\"",
"           sodipodi:role=\"line\">key:&lt;Primary&gt;&lt;Shift&gt;KP_Enter</tspan></text>",
"      <text",
"         sodipodi:linespacing=\"100%\"",
"         id=\"text4604\"",
"         y=\"150\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#ff00ff;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"150\"",
"           x=\"5\"",
"           id=\"tspan4606\"",
"           sodipodi:role=\"line\">key:&lt;Primary&gt;&lt;Alt&gt;KP_Enter</tspan></text>",
"      <text",
"         sodipodi:linespacing=\"100%\"",
"         id=\"text4608\"",
"         y=\"165\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#008080;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"165\"",
"           x=\"5\"",
"           id=\"tspan4610\"",
"           sodipodi:role=\"line\">key:&lt;Shift&gt;&lt;Alt&gt;KP_Enter</tspan></text>",
"    </g>",
"    <g",
"       transform=\"translate(2680,1850)\"",
"       id=\"g4612\">",
"      <rect",
"         style=\"fill:#0000ff;fill-opacity:0.1;stroke:#000000;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1\"",
"         id=\"rect4614\"",
"         width=\"170\"",
"         height=\"350\"",
"         x=\"0\"",
"         y=\"2.9921875\" />",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:32px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#000000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         x=\"10\"",
"         y=\"40\"",
"         id=\"text4616\"",
"         sodipodi:linespacing=\"100%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan4618\"",
"           x=\"10\"",
"           y=\"40\">KP_Plus</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#000000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"60\"",
"         id=\"text4620\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan4622\"",
"           x=\"5\"",
"           y=\"60\">key:KP_Add</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#ff0000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"75\"",
"         id=\"text4624\"><tspan",
"           sodipodi:role=\"line\"",
"           x=\"5\"",
"           y=\"75\"",
"           id=\"tspan4626\">key:&lt;Primary&gt;KP_Add</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#008000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"90\"",
"         id=\"text4628\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan4630\"",
"           x=\"5\"",
"           y=\"90\">key:&lt;Shift&gt;KP_Add</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#0000ff;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"105\"",
"         id=\"text4632\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan4634\"",
"           x=\"5\"",
"           y=\"105\">key:&lt;Alt&gt;KP_Add</tspan></text>",
"      <text",
"         id=\"text6855\"",
"         y=\"120\"",
"         x=\"5\"",
"         style=\"font-style:normal;font-weight:normal;font-size:12.00000048px;font-family:'Bitstream Vera Sans';fill:#464646;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;-inkscape-font-specification:'Bitstream Vera Sans, Normal';font-stretch:normal;font-variant:normal;text-anchor:start;text-align:start;writing-mode:lr;line-height:125%;\"",
"         xml:space=\"preserve\"",
"         sodipodi:linespacing=\"125%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan6859\">key:&lt;Super&gt;KP_Add</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#808000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"135\"",
"         id=\"text4636\"",
"         sodipodi:linespacing=\"100%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan4638\"",
"           x=\"5\"",
"           y=\"135\">key:&lt;Primary&gt;&lt;Shift&gt;KP_Add</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#ff00ff;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"150\"",
"         id=\"text4640\"",
"         sodipodi:linespacing=\"100%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan4642\"",
"           x=\"5\"",
"           y=\"150\">key:&lt;Primary&gt;&lt;Alt&gt;KP_Add</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#008080;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"165\"",
"         id=\"text4644\"",
"         sodipodi:linespacing=\"100%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan4646\"",
"           x=\"5\"",
"           y=\"165\">key:&lt;Shift&gt;&lt;Alt&gt;KP_Add</tspan></text>",
"    </g>",
"    <g",
"       transform=\"translate(1920,2390)\"",
"       id=\"g4648\">",
"      <rect",
"         style=\"fill:#0000ff;fill-opacity:0.1;stroke:#000000;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1\"",
"         id=\"rect4650\"",
"         width=\"170\"",
"         height=\"170\"",
"         x=\"0\"",
"         y=\"2.9921875\" />",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:48px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#000000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         x=\"40\"",
"         y=\"42.992188\"",
"         id=\"text4652\"",
"         sodipodi:linespacing=\"100%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan4654\"",
"           x=\"40\"",
"           y=\"42.992188\">Right</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#000000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"60\"",
"         id=\"text4656\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan4658\"",
"           x=\"5\"",
"           y=\"60\">key:Right</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#ff0000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"75\"",
"         id=\"text4660\"><tspan",
"           sodipodi:role=\"line\"",
"           x=\"5\"",
"           y=\"75\"",
"           id=\"tspan4662\">key:&lt;Primary&gt;Right</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#008000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"90\"",
"         id=\"text4664\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan4666\"",
"           x=\"5\"",
"           y=\"90\">key:&lt;Shift&gt;Right</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#0000ff;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"105\"",
"         id=\"text4668\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan4670\"",
"           x=\"5\"",
"           y=\"105\">key:&lt;Alt&gt;Right</tspan></text>",
"      <text",
"         id=\"text6807\"",
"         y=\"120\"",
"         x=\"5\"",
"         style=\"font-style:normal;font-weight:normal;font-size:12.00000048px;font-family:'Bitstream Vera Sans';fill:#464646;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;-inkscape-font-specification:'Bitstream Vera Sans, Normal';font-stretch:normal;font-variant:normal;text-anchor:start;text-align:start;writing-mode:lr;line-height:125%;\"",
"         xml:space=\"preserve\"",
"         sodipodi:linespacing=\"125%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan6811\">key:&lt;Super&gt;Right</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#808000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"135\"",
"         id=\"text4672\"",
"         sodipodi:linespacing=\"100%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan4674\"",
"           x=\"5\"",
"           y=\"135\">key:&lt;Primary&gt;&lt;Shift&gt;Right</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#ff00ff;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"150\"",
"         id=\"text4676\"",
"         sodipodi:linespacing=\"100%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan4678\"",
"           x=\"5\"",
"           y=\"150\">key:&lt;Primary&gt;&lt;Alt&gt;Right</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#008080;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"165\"",
"         id=\"text4680\"",
"         sodipodi:linespacing=\"100%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan4682\"",
"           x=\"5\"",
"           y=\"165\">key:&lt;Shift&gt;&lt;Alt&gt;Right</tspan></text>",
"    </g>",
"    <g",
"       id=\"g4684\"",
"       transform=\"translate(1740,2210)\">",
"      <rect",
"         y=\"2.9921875\"",
"         x=\"0\"",
"         height=\"170\"",
"         width=\"170\"",
"         id=\"rect4686\"",
"         style=\"fill:#0000ff;fill-opacity:0.1;stroke:#000000;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1\" />",
"      <text",
"         sodipodi:linespacing=\"100%\"",
"         id=\"text4688\"",
"         y=\"42.992188\"",
"         x=\"40\"",
"         style=\"font-size:48px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#000000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"42.992188\"",
"           x=\"40\"",
"           id=\"tspan4690\"",
"           sodipodi:role=\"line\">Up</tspan></text>",
"      <text",
"         id=\"text4692\"",
"         y=\"60\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#000000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"60\"",
"           x=\"5\"",
"           id=\"tspan4694\"",
"           sodipodi:role=\"line\">key:Up</tspan></text>",
"      <text",
"         id=\"text4696\"",
"         y=\"75\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#ff0000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           id=\"tspan4698\"",
"           y=\"75\"",
"           x=\"5\"",
"           sodipodi:role=\"line\">key:&lt;Primary&gt;Up</tspan></text>",
"      <text",
"         id=\"text4700\"",
"         y=\"90\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#008000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"90\"",
"           x=\"5\"",
"           id=\"tspan4702\"",
"           sodipodi:role=\"line\">key:&lt;Shift&gt;Up</tspan></text>",
"      <text",
"         id=\"text4704\"",
"         y=\"105\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#0000ff;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"105\"",
"           x=\"5\"",
"           id=\"tspan4706\"",
"           sodipodi:role=\"line\">key:&lt;Alt&gt;Up</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-style:normal;font-weight:normal;font-size:12.00000048px;font-family:'Bitstream Vera Sans';fill:#464646;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;-inkscape-font-specification:'Bitstream Vera Sans, Normal';font-stretch:normal;font-variant:normal;text-anchor:start;text-align:start;writing-mode:lr;line-height:125%;\"",
"         x=\"5\"",
"         y=\"120\"",
"         id=\"text6795\"",
"         sodipodi:linespacing=\"125%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan6799\">key:&lt;Super&gt;Up</tspan></text>",
"      <text",
"         sodipodi:linespacing=\"100%\"",
"         id=\"text4708\"",
"         y=\"135\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#808000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"135\"",
"           x=\"5\"",
"           id=\"tspan4710\"",
"           sodipodi:role=\"line\">key:&lt;Primary&gt;&lt;Shift&gt;Up</tspan></text>",
"      <text",
"         sodipodi:linespacing=\"100%\"",
"         id=\"text4712\"",
"         y=\"150\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#ff00ff;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"150\"",
"           x=\"5\"",
"           id=\"tspan4714\"",
"           sodipodi:role=\"line\">key:&lt;Primary&gt;&lt;Alt&gt;Up</tspan></text>",
"      <text",
"         sodipodi:linespacing=\"100%\"",
"         id=\"text4716\"",
"         y=\"165\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#008080;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"165\"",
"           x=\"5\"",
"           id=\"tspan4718\"",
"           sodipodi:role=\"line\">key:&lt;Shift&gt;&lt;Alt&gt;Up</tspan></text>",
"    </g>",
"    <g",
"       transform=\"translate(1560,2390)\"",
"       id=\"g4720\">",
"      <rect",
"         style=\"fill:#0000ff;fill-opacity:0.1;stroke:#000000;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1\"",
"         id=\"rect4722\"",
"         width=\"170\"",
"         height=\"170\"",
"         x=\"0\"",
"         y=\"2.9921875\" />",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:48px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#000000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         x=\"40\"",
"         y=\"42.992188\"",
"         id=\"text4724\"",
"         sodipodi:linespacing=\"100%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan4726\"",
"           x=\"40\"",
"           y=\"42.992188\">Left</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#000000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"60\"",
"         id=\"text4728\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan4730\"",
"           x=\"5\"",
"           y=\"60\">key:Left</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#ff0000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"75\"",
"         id=\"text4732\"><tspan",
"           sodipodi:role=\"line\"",
"           x=\"5\"",
"           y=\"75\"",
"           id=\"tspan4734\">key:&lt;Primary&gt;Left</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#008000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"90\"",
"         id=\"text4736\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan4738\"",
"           x=\"5\"",
"           y=\"90\">key:&lt;Shift&gt;Left</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#0000ff;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"105\"",
"         id=\"text4740\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan4742\"",
"           x=\"5\"",
"           y=\"105\">key:&lt;Alt&gt;Left</tspan></text>",
"      <text",
"         id=\"text6789\"",
"         y=\"120\"",
"         x=\"5\"",
"         style=\"font-style:normal;font-weight:normal;font-size:12.00000048px;font-family:'Bitstream Vera Sans';fill:#464646;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;-inkscape-font-specification:'Bitstream Vera Sans, Normal';font-stretch:normal;font-variant:normal;text-anchor:start;text-align:start;writing-mode:lr;line-height:125%;\"",
"         xml:space=\"preserve\"",
"         sodipodi:linespacing=\"125%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan6793\">key:&lt;Super&gt;Left</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#808000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"135\"",
"         id=\"text4744\"",
"         sodipodi:linespacing=\"100%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan4746\"",
"           x=\"5\"",
"           y=\"135\">key:&lt;Primary&gt;&lt;Shift&gt;Left</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#ff00ff;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"150\"",
"         id=\"text4748\"",
"         sodipodi:linespacing=\"100%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan4750\"",
"           x=\"5\"",
"           y=\"150\">key:&lt;Primary&gt;&lt;Alt&gt;Left</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#008080;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"165\"",
"         id=\"text4752\"",
"         sodipodi:linespacing=\"100%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan4754\"",
"           x=\"5\"",
"           y=\"165\">key:&lt;Shift&gt;&lt;Alt&gt;Left</tspan></text>",
"    </g>",
"    <g",
"       id=\"g4756\"",
"       transform=\"translate(1740,2390)\">",
"      <rect",
"         y=\"2.9921875\"",
"         x=\"0\"",
"         height=\"170\"",
"         width=\"170\"",
"         id=\"rect4758\"",
"         style=\"fill:#0000ff;fill-opacity:0.1;stroke:#000000;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1\" />",
"      <text",
"         sodipodi:linespacing=\"100%\"",
"         id=\"text4760\"",
"         y=\"42.992188\"",
"         x=\"40\"",
"         style=\"font-size:48px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#000000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"42.992188\"",
"           x=\"40\"",
"           id=\"tspan4762\"",
"           sodipodi:role=\"line\">Down</tspan></text>",
"      <text",
"         id=\"text4764\"",
"         y=\"60\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#000000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"60\"",
"           x=\"5\"",
"           id=\"tspan4766\"",
"           sodipodi:role=\"line\">key:Down</tspan></text>",
"      <text",
"         id=\"text4768\"",
"         y=\"75\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#ff0000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           id=\"tspan4770\"",
"           y=\"75\"",
"           x=\"5\"",
"           sodipodi:role=\"line\">key:&lt;Primary&gt;Down</tspan></text>",
"      <text",
"         id=\"text4772\"",
"         y=\"90\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#008000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"90\"",
"           x=\"5\"",
"           id=\"tspan4774\"",
"           sodipodi:role=\"line\">key:&lt;Shift&gt;Down</tspan></text>",
"      <text",
"         id=\"text4776\"",
"         y=\"105\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#0000ff;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"105\"",
"           x=\"5\"",
"           id=\"tspan4778\"",
"           sodipodi:role=\"line\">key:&lt;Alt&gt;Down</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-style:normal;font-weight:normal;font-size:12.00000048px;font-family:'Bitstream Vera Sans';fill:#464646;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;-inkscape-font-specification:'Bitstream Vera Sans, Normal';font-stretch:normal;font-variant:normal;text-anchor:start;text-align:start;writing-mode:lr;line-height:125%;\"",
"         x=\"5\"",
"         y=\"120\"",
"         id=\"text6801\"",
"         sodipodi:linespacing=\"125%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan6805\">key:&lt;Super&gt;Down</tspan></text>",
"      <text",
"         sodipodi:linespacing=\"100%\"",
"         id=\"text4780\"",
"         y=\"135\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#808000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"135\"",
"           x=\"5\"",
"           id=\"tspan4782\"",
"           sodipodi:role=\"line\">key:&lt;Primary&gt;&lt;Shift&gt;Down</tspan></text>",
"      <text",
"         sodipodi:linespacing=\"100%\"",
"         id=\"text4784\"",
"         y=\"150\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#ff00ff;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"150\"",
"           x=\"5\"",
"           id=\"tspan4786\"",
"           sodipodi:role=\"line\">key:&lt;Primary&gt;&lt;Alt&gt;Down</tspan></text>",
"      <text",
"         sodipodi:linespacing=\"100%\"",
"         id=\"text4788\"",
"         y=\"165\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#008080;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"165\"",
"           x=\"5\"",
"           id=\"tspan4790\"",
"           sodipodi:role=\"line\">key:&lt;Shift&gt;&lt;Alt&gt;Down</tspan></text>",
"    </g>",
"    <g",
"       id=\"g4792\"",
"       transform=\"translate(160,1450)\">",
"      <rect",
"         y=\"2.9921875\"",
"         x=\"0\"",
"         height=\"530\"",
"         width=\"170\"",
"         id=\"rect4794\"",
"         style=\"fill:#0000ff;fill-opacity:0.1;stroke:#000000;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1\" />",
"      <text",
"         sodipodi:linespacing=\"100%\"",
"         id=\"text4796\"",
"         y=\"42.992188\"",
"         x=\"40\"",
"         style=\"font-size:48px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#000000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"42.992188\"",
"           x=\"40\"",
"           id=\"tspan4798\"",
"           sodipodi:role=\"line\">M1</tspan></text>",
"      <text",
"         id=\"text4800\"",
"         y=\"60\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#000000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"60\"",
"           x=\"5\"",
"           id=\"tspan4802\"",
"           sodipodi:role=\"line\">key:M1</tspan></text>",
"      <text",
"         id=\"text4804\"",
"         y=\"75\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#ff0000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           id=\"tspan4806\"",
"           y=\"75\"",
"           x=\"5\"",
"           sodipodi:role=\"line\">key:&lt;Primary&gt;M1</tspan></text>",
"      <text",
"         id=\"text4808\"",
"         y=\"90\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#008000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"90\"",
"           x=\"5\"",
"           id=\"tspan4810\"",
"           sodipodi:role=\"line\">key:&lt;Shift&gt;M1</tspan></text>",
"      <text",
"         id=\"text4812\"",
"         y=\"105\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#0000ff;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"105\"",
"           x=\"5\"",
"           id=\"tspan4814\"",
"           sodipodi:role=\"line\">key:&lt;Alt&gt;M1</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-style:normal;font-weight:normal;font-size:12.00000048px;font-family:'Bitstream Vera Sans';fill:#464646;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;-inkscape-font-specification:'Bitstream Vera Sans, Normal';font-stretch:normal;font-variant:normal;text-anchor:start;text-align:start;writing-mode:lr;line-height:125%;\"",
"         x=\"5\"",
"         y=\"120\"",
"         id=\"text6707\"",
"         sodipodi:linespacing=\"125%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan6711\">key:&lt;Super&gt;M1</tspan></text>",
"      <text",
"         sodipodi:linespacing=\"100%\"",
"         id=\"text4816\"",
"         y=\"135\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#808000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"135\"",
"           x=\"5\"",
"           id=\"tspan4818\"",
"           sodipodi:role=\"line\">key:&lt;Primary&gt;&lt;Shift&gt;M1</tspan></text>",
"      <text",
"         sodipodi:linespacing=\"100%\"",
"         id=\"text4820\"",
"         y=\"150\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#ff00ff;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"150\"",
"           x=\"5\"",
"           id=\"tspan4822\"",
"           sodipodi:role=\"line\">key:&lt;Primary&gt;&lt;Alt&gt;M1</tspan></text>",
"      <text",
"         sodipodi:linespacing=\"100%\"",
"         id=\"text4824\"",
"         y=\"165\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#008080;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"165\"",
"           x=\"5\"",
"           id=\"tspan4826\"",
"           sodipodi:role=\"line\">key:&lt;Shift&gt;&lt;Alt&gt;M1</tspan></text>",
"    </g>",
"    <g",
"       transform=\"translate(520,1450)\"",
"       id=\"g4828\">",
"      <rect",
"         style=\"fill:#0000ff;fill-opacity:0.1;stroke:#000000;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1\"",
"         id=\"rect4830\"",
"         width=\"170\"",
"         height=\"530\"",
"         x=\"0\"",
"         y=\"2.9921875\" />",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:48px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#000000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         x=\"40\"",
"         y=\"42.992188\"",
"         id=\"text4832\"",
"         sodipodi:linespacing=\"100%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan4834\"",
"           x=\"40\"",
"           y=\"42.992188\">M3</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#000000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"60\"",
"         id=\"text4836\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan4838\"",
"           x=\"5\"",
"           y=\"60\">key:M3</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#ff0000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"75\"",
"         id=\"text4840\"><tspan",
"           sodipodi:role=\"line\"",
"           x=\"5\"",
"           y=\"75\"",
"           id=\"tspan4842\">key:&lt;Primary&gt;M3</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#008000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"90\"",
"         id=\"text4844\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan4846\"",
"           x=\"5\"",
"           y=\"90\">key:&lt;Shift&gt;M3</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#0000ff;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"105\"",
"         id=\"text4848\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan4850\"",
"           x=\"5\"",
"           y=\"105\">key:&lt;Alt&gt;M3</tspan></text>",
"      <text",
"         id=\"text6731\"",
"         y=\"120\"",
"         x=\"5\"",
"         style=\"font-style:normal;font-weight:normal;font-size:12px;font-family:'Bitstream Vera Sans';fill:#464646;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"105\"",
"           x=\"5\"",
"           id=\"tspan6733\"",
"           sodipodi:role=\"line\">key:&lt;Alt&gt;M3</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#808000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"135\"",
"         id=\"text4852\"",
"         sodipodi:linespacing=\"100%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan4854\"",
"           x=\"5\"",
"           y=\"135\">key:&lt;Primary&gt;&lt;Shift&gt;M3</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#ff00ff;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"150\"",
"         id=\"text4856\"",
"         sodipodi:linespacing=\"100%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan4858\"",
"           x=\"5\"",
"           y=\"150\">key:&lt;Primary&gt;&lt;Alt&gt;M3</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#008080;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"165\"",
"         id=\"text4860\"",
"         sodipodi:linespacing=\"100%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan4862\"",
"           x=\"5\"",
"           y=\"165\">key:&lt;Shift&gt;&lt;Alt&gt;M3</tspan></text>",
"    </g>",
"    <g",
"       id=\"g4864\"",
"       transform=\"translate(340,1450)\">",
"      <rect",
"         y=\"2.9921875\"",
"         x=\"0\"",
"         height=\"170\"",
"         width=\"170\"",
"         id=\"rect4866\"",
"         style=\"fill:#0000ff;fill-opacity:0.1;stroke:#000000;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1\" />",
"      <text",
"         sodipodi:linespacing=\"100%\"",
"         id=\"text4868\"",
"         y=\"42.992188\"",
"         x=\"40\"",
"         style=\"font-size:48px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#000000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"42.992188\"",
"           x=\"40\"",
"           id=\"tspan4870\"",
"           sodipodi:role=\"line\">M2</tspan></text>",
"      <text",
"         id=\"text4872\"",
"         y=\"60\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#000000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"60\"",
"           x=\"5\"",
"           id=\"tspan4874\"",
"           sodipodi:role=\"line\">key:M2</tspan></text>",
"      <text",
"         id=\"text4876\"",
"         y=\"75\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#ff0000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           id=\"tspan4878\"",
"           y=\"75\"",
"           x=\"5\"",
"           sodipodi:role=\"line\">key:&lt;Primary&gt;M2</tspan></text>",
"      <text",
"         id=\"text4880\"",
"         y=\"90\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#008000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"90\"",
"           x=\"5\"",
"           id=\"tspan4882\"",
"           sodipodi:role=\"line\">key:&lt;Shift&gt;M2</tspan></text>",
"      <text",
"         id=\"text4884\"",
"         y=\"105\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#0000ff;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"105\"",
"           x=\"5\"",
"           id=\"tspan4886\"",
"           sodipodi:role=\"line\">key:&lt;Alt&gt;M2</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-style:normal;font-weight:normal;font-size:12.00000048px;font-family:'Bitstream Vera Sans';fill:#464646;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;-inkscape-font-specification:'Bitstream Vera Sans, Normal';font-stretch:normal;font-variant:normal;text-anchor:start;text-align:start;writing-mode:lr;line-height:125%;\"",
"         x=\"5\"",
"         y=\"120\"",
"         id=\"text6713\"",
"         sodipodi:linespacing=\"125%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan6717\">key:&lt;Super&gt;M2</tspan></text>",
"      <text",
"         sodipodi:linespacing=\"100%\"",
"         id=\"text4888\"",
"         y=\"135\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#808000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"135\"",
"           x=\"5\"",
"           id=\"tspan4890\"",
"           sodipodi:role=\"line\">key:&lt;Primary&gt;&lt;Shift&gt;M2</tspan></text>",
"      <text",
"         sodipodi:linespacing=\"100%\"",
"         id=\"text4893\"",
"         y=\"150\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#ff00ff;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"150\"",
"           x=\"5\"",
"           id=\"tspan4895\"",
"           sodipodi:role=\"line\">key:&lt;Primary&gt;&lt;Alt&gt;M2</tspan></text>",
"      <text",
"         sodipodi:linespacing=\"100%\"",
"         id=\"text4897\"",
"         y=\"165\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#008080;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"165\"",
"           x=\"5\"",
"           id=\"tspan4899\"",
"           sodipodi:role=\"line\">key:&lt;Shift&gt;&lt;Alt&gt;M2</tspan></text>",
"    </g>",
"    <g",
"       transform=\"translate(340,1810)\"",
"       id=\"g4901\">",
"      <rect",
"         style=\"fill:#0000ff;fill-opacity:0.1;stroke:#000000;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1\"",
"         id=\"rect4903\"",
"         width=\"170\"",
"         height=\"170\"",
"         x=\"0\"",
"         y=\"2.9921875\" />",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:48px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#000000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         x=\"40\"",
"         y=\"42.992188\"",
"         id=\"text4905\"",
"         sodipodi:linespacing=\"100%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan4907\"",
"           x=\"40\"",
"           y=\"42.992188\">MW5</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#000000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"60\"",
"         id=\"text4909\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan4911\"",
"           x=\"5\"",
"           y=\"60\">key:MW5</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#ff0000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"75\"",
"         id=\"text4913\"><tspan",
"           sodipodi:role=\"line\"",
"           x=\"5\"",
"           y=\"75\"",
"           id=\"tspan4915\">key:&lt;Primary&gt;MW5</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#008000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"90\"",
"         id=\"text4917\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan4919\"",
"           x=\"5\"",
"           y=\"90\">key:&lt;Shift&gt;MW5</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#0000ff;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"105\"",
"         id=\"text4921\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan4923\"",
"           x=\"5\"",
"           y=\"105\">key:&lt;Alt&gt;MW5</tspan></text>",
"      <text",
"         id=\"text6725\"",
"         y=\"120\"",
"         x=\"5\"",
"         style=\"font-style:normal;font-weight:normal;font-size:12.00000048px;font-family:'Bitstream Vera Sans';fill:#464646;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;-inkscape-font-specification:'Bitstream Vera Sans, Normal';font-stretch:normal;font-variant:normal;text-anchor:start;text-align:start;writing-mode:lr;line-height:125%;\"",
"         xml:space=\"preserve\"",
"         sodipodi:linespacing=\"125%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan6729\">key:&lt;Super&gt;MW5</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#808000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"135\"",
"         id=\"text4925\"",
"         sodipodi:linespacing=\"100%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan4927\"",
"           x=\"5\"",
"           y=\"135\">key:&lt;Primary&gt;&lt;Shift&gt;MW5</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#ff00ff;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"150\"",
"         id=\"text4929\"",
"         sodipodi:linespacing=\"100%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan4931\"",
"           x=\"5\"",
"           y=\"150\">key:&lt;Primary&gt;&lt;Alt&gt;MW5</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-size:12px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#008080;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         x=\"5\"",
"         y=\"165\"",
"         id=\"text4933\"",
"         sodipodi:linespacing=\"100%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan4935\"",
"           x=\"5\"",
"           y=\"165\">key:&lt;Shift&gt;&lt;Alt&gt;MW5</tspan></text>",
"    </g>",
"    <g",
"       id=\"g4937\"",
"       transform=\"translate(340,1630)\">",
"      <rect",
"         y=\"2.9921875\"",
"         x=\"0\"",
"         height=\"170\"",
"         width=\"170\"",
"         id=\"rect4939\"",
"         style=\"fill:#0000ff;fill-opacity:0.1;stroke:#000000;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1\" />",
"      <text",
"         sodipodi:linespacing=\"100%\"",
"         id=\"text4941\"",
"         y=\"42.992188\"",
"         x=\"40\"",
"         style=\"font-size:48px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#000000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"42.992188\"",
"           x=\"40\"",
"           id=\"tspan4943\"",
"           sodipodi:role=\"line\">MW4</tspan></text>",
"      <text",
"         id=\"text4945\"",
"         y=\"60\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#000000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"60\"",
"           x=\"5\"",
"           id=\"tspan4947\"",
"           sodipodi:role=\"line\">key:MW4</tspan></text>",
"      <text",
"         id=\"text4949\"",
"         y=\"75\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#ff0000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           id=\"tspan4951\"",
"           y=\"75\"",
"           x=\"5\"",
"           sodipodi:role=\"line\">key:&lt;Primary&gt;MW4</tspan></text>",
"      <text",
"         id=\"text4953\"",
"         y=\"90\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#008000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"90\"",
"           x=\"5\"",
"           id=\"tspan4955\"",
"           sodipodi:role=\"line\">key:&lt;Shift&gt;MW4</tspan></text>",
"      <text",
"         id=\"text4957\"",
"         y=\"105\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-weight:normal;fill:#0000ff;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"105\"",
"           x=\"5\"",
"           id=\"tspan4959\"",
"           sodipodi:role=\"line\">key:&lt;Alt&gt;MW4</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-style:normal;font-weight:normal;font-size:12.00000048px;font-family:'Bitstream Vera Sans';fill:#464646;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;-inkscape-font-specification:'Bitstream Vera Sans, Normal';font-stretch:normal;font-variant:normal;text-anchor:start;text-align:start;writing-mode:lr;line-height:125%;\"",
"         x=\"5\"",
"         y=\"120\"",
"         id=\"text6719\"",
"         sodipodi:linespacing=\"125%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan6723\">key:&lt;Super&gt;MW4</tspan></text>",
"      <text",
"         sodipodi:linespacing=\"100%\"",
"         id=\"text4961\"",
"         y=\"135\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#808000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"135\"",
"           x=\"5\"",
"           id=\"tspan4963\"",
"           sodipodi:role=\"line\">key:&lt;Primary&gt;&lt;Shift&gt;MW4</tspan></text>",
"      <text",
"         sodipodi:linespacing=\"100%\"",
"         id=\"text4965\"",
"         y=\"150\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#ff00ff;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"150\"",
"           x=\"5\"",
"           id=\"tspan4967\"",
"           sodipodi:role=\"line\">key:&lt;Primary&gt;&lt;Alt&gt;MW4</tspan></text>",
"      <text",
"         sodipodi:linespacing=\"100%\"",
"         id=\"text4969\"",
"         y=\"165\"",
"         x=\"5\"",
"         style=\"font-size:12px;font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;text-align:start;line-height:100%;writing-mode:lr-tb;text-anchor:start;fill:#008080;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;font-family:Bitstream Vera Sans;-inkscape-font-specification:Bitstream Vera Sans\"",
"         xml:space=\"preserve\"><tspan",
"           y=\"165\"",
"           x=\"5\"",
"           id=\"tspan4971\"",
"           sodipodi:role=\"line\">key:&lt;Shift&gt;&lt;Alt&gt;MW4</tspan></text>",
"    </g>",
"    <path",
"       style=\"fill:none;fill-opacity:0.75;fill-rule:evenodd;stroke:#000000;stroke-width:3.75;stroke-linecap:round;stroke-linejoin:round;stroke-miterlimit:4;stroke-dasharray:none;stroke-opacity:1\"",
"       d=\"M 140,1432.9922 L 710,1432.9922 L 710,2262.9922 C 710,2262.9922 720,2502.9922 420,2502.9922 C 120,2502.9922 140,2262.9922 140,2262.9922 L 140,1432.9922 z\"",
"       id=\"path4973\"",
"       sodipodi:nodetypes=\"ccczcc\" />",
"    <g",
"       transform=\"translate(1017.269,1432.8172)\"",
"       id=\"g2942-3\">",
"      <rect",
"         style=\"fill:#0000ff;fill-opacity:0.1;stroke:#000000;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1\"",
"         id=\"rect2944-6\"",
"         width=\"170\"",
"         height=\"170\"",
"         x=\"0.71428585\"",
"         y=\"-1.2935265\" />",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;font-size:44.8187491px;line-height:100%;font-family:'Bitstream Vera Sans';-inkscape-font-specification:'Bitstream Vera Sans, Normal';text-align:start;writing-mode:lr;text-anchor:start;fill:#000000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;\"",
"         x=\"27.144716\"",
"         y=\"40.809456\"",
"         id=\"text2946-7\"",
"         sodipodi:linespacing=\"100%\"",
"         transform=\"scale(0.93371642,1.070989)\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan5325\">Colors</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-style:normal;font-weight:normal;font-size:12.00000048px;font-family:'Bitstream Vera Sans';fill:#000000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;-inkscape-font-specification:'Bitstream Vera Sans, Normal';font-stretch:normal;font-variant:normal;text-anchor:start;text-align:start;writing-mode:lr;line-height:125%;\"",
"         x=\"5\"",
"         y=\"60\"",
"         id=\"text2950-3\"",
"         sodipodi:linespacing=\"125%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan9406\">key</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-style:normal;font-weight:normal;font-size:12.00000048px;font-family:'Bitstream Vera Sans';fill:#ff0000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;-inkscape-font-specification:'Bitstream Vera Sans, Normal';font-stretch:normal;font-variant:normal;text-anchor:start;text-align:start;writing-mode:lr;line-height:125%;\"",
"         x=\"5\"",
"         y=\"75\"",
"         id=\"text2954-6\"",
"         sodipodi:linespacing=\"125%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan9408\">&lt;Primary&gt;key</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-style:normal;font-weight:normal;font-size:12.00000048px;font-family:'Bitstream Vera Sans';fill:#008000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;-inkscape-font-specification:'Bitstream Vera Sans, Normal';font-stretch:normal;font-variant:normal;text-anchor:start;text-align:start;writing-mode:lr;line-height:125%;\"",
"         x=\"5\"",
"         y=\"90\"",
"         id=\"text2958-9\"",
"         sodipodi:linespacing=\"125%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan9410\">&lt;Shift&gt;key</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-style:normal;font-weight:normal;font-size:12.00000048px;font-family:'Bitstream Vera Sans';fill:#0000ff;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;-inkscape-font-specification:'Bitstream Vera Sans, Normal';font-stretch:normal;font-variant:normal;text-anchor:start;text-align:start;writing-mode:lr;line-height:125%;\"",
"         x=\"5\"",
"         y=\"105\"",
"         id=\"text2962-2\"",
"         sodipodi:linespacing=\"125%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan9412\">&lt;Alt&gt;key</tspan></text>",
"      <text",
"         sodipodi:linespacing=\"125%\"",
"         id=\"text6391\"",
"         y=\"120\"",
"         x=\"5\"",
"         style=\"font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;font-size:12.00000048px;line-height:125%;font-family:'Bitstream Vera Sans';-inkscape-font-specification:'Bitstream Vera Sans, Normal';text-align:start;writing-mode:lr;text-anchor:start;fill:#464646;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;\"",
"         xml:space=\"preserve\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan6395\">&lt;Super&gt;key</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;font-size:12.00000048px;line-height:100%;font-family:'Bitstream Vera Sans';-inkscape-font-specification:'Bitstream Vera Sans, Normal';text-align:start;writing-mode:lr;text-anchor:start;fill:#808000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;\"",
"         x=\"5\"",
"         y=\"135\"",
"         id=\"text2966-0\"",
"         sodipodi:linespacing=\"100%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan9414\">&lt;Primary&gt;&lt;Shift&gt;key</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;font-size:12.00000048px;line-height:100%;font-family:'Bitstream Vera Sans';-inkscape-font-specification:'Bitstream Vera Sans, Normal';text-align:start;writing-mode:lr;text-anchor:start;fill:#ff00ff;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;\"",
"         x=\"5\"",
"         y=\"150\"",
"         id=\"text2970-3\"",
"         sodipodi:linespacing=\"100%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan9416\">&lt;Primary&gt;&lt;Alt&gt;key</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;font-size:12.00000048px;line-height:100%;font-family:'Bitstream Vera Sans';-inkscape-font-specification:'Bitstream Vera Sans, Normal';text-align:start;writing-mode:lr;text-anchor:start;fill:#008080;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;\"",
"         x=\"5\"",
"         y=\"165\"",
"         id=\"text2974-0\"",
"         sodipodi:linespacing=\"100%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan9418\">&lt;Shift&gt;&lt;Alt&gt;key</tspan></text>",
"    </g>",
"    <g",
"       id=\"g2978-8\"",
"       transform=\"translate(408.57143,891.42857)\">",
"      <rect",
"         y=\"2.9921875\"",
"         x=\"0\"",
"         height=\"170\"",
"         width=\"170\"",
"         id=\"rect2980-3\"",
"         style=\"fill:#0000ff;fill-opacity:0.1;stroke:#000000;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1\" />",
"      <text",
"         sodipodi:linespacing=\"100%\"",
"         id=\"text2982-8\"",
"         y=\"42.992188\"",
"         x=\"40\"",
"         style=\"font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;font-size:48.00000191px;line-height:100%;font-family:'Bitstream Vera Sans';-inkscape-font-specification:'Bitstream Vera Sans, Normal';text-align:start;writing-mode:lr;text-anchor:start;fill:#000000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;\"",
"         xml:space=\"preserve\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan11457\">\\</tspan></text>",
"      <text",
"         id=\"text2986-3\"",
"         y=\"60\"",
"         x=\"5\"",
"         style=\"font-style:normal;font-weight:normal;font-size:12.00000048px;font-family:'Bitstream Vera Sans';fill:#000000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;-inkscape-font-specification:'Bitstream Vera Sans, Normal';font-stretch:normal;font-variant:normal;text-anchor:start;text-align:start;writing-mode:lr;line-height:125%;\"",
"         xml:space=\"preserve\"",
"         sodipodi:linespacing=\"125%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan11459\">key:backslash</tspan></text>",
"      <text",
"         id=\"text2990-8\"",
"         y=\"75\"",
"         x=\"5\"",
"         style=\"font-style:normal;font-weight:normal;font-size:12.00000048px;font-family:'Bitstream Vera Sans';fill:#ff0000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;-inkscape-font-specification:'Bitstream Vera Sans, Normal';font-stretch:normal;font-variant:normal;text-anchor:start;text-align:start;writing-mode:lr;line-height:125%;\"",
"         xml:space=\"preserve\"",
"         sodipodi:linespacing=\"125%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan11461\">key:&lt;Primary&gt;backslash</tspan></text>",
"      <text",
"         id=\"text2994-4\"",
"         y=\"90\"",
"         x=\"5\"",
"         style=\"font-style:normal;font-weight:normal;font-size:12.00000048px;font-family:'Bitstream Vera Sans';fill:#008000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;-inkscape-font-specification:'Bitstream Vera Sans, Normal';font-stretch:normal;font-variant:normal;text-anchor:start;text-align:start;writing-mode:lr;line-height:125%;\"",
"         xml:space=\"preserve\"",
"         sodipodi:linespacing=\"125%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan11463\">key:&lt;Shift&gt;bar</tspan></text>",
"      <text",
"         id=\"text2998-6\"",
"         y=\"105\"",
"         x=\"5\"",
"         style=\"font-style:normal;font-weight:normal;font-size:12.00000048px;font-family:'Bitstream Vera Sans';fill:#0000ff;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;-inkscape-font-specification:'Bitstream Vera Sans, Normal';font-stretch:normal;font-variant:normal;text-anchor:start;text-align:start;writing-mode:lr;line-height:125%;\"",
"         xml:space=\"preserve\"",
"         sodipodi:linespacing=\"125%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan11465\">key:&lt;Alt&gt;backslash</tspan></text>",
"      <text",
"         sodipodi:linespacing=\"125%\"",
"         xml:space=\"preserve\"",
"         style=\"font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;font-size:12.00000048px;line-height:125%;font-family:'Bitstream Vera Sans';-inkscape-font-specification:'Bitstream Vera Sans, Normal';text-align:start;writing-mode:lr;text-anchor:start;fill:#464646;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;\"",
"         x=\"5\"",
"         y=\"120\"",
"         id=\"text6631\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan6635\">key:&lt;Super&gt;backslash</tspan></text>",
"      <text",
"         sodipodi:linespacing=\"100%\"",
"         id=\"text3002-9\"",
"         y=\"135\"",
"         x=\"5\"",
"         style=\"font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;font-size:12.00000048px;line-height:100%;font-family:'Bitstream Vera Sans';-inkscape-font-specification:'Bitstream Vera Sans, Normal';text-align:start;writing-mode:lr;text-anchor:start;fill:#808000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;\"",
"         xml:space=\"preserve\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan11467\">key:&lt;Primary&gt;&lt;Shift&gt;bar</tspan></text>",
"      <text",
"         sodipodi:linespacing=\"100%\"",
"         id=\"text3006-6\"",
"         y=\"150\"",
"         x=\"5\"",
"         style=\"font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;font-size:12.00000048px;line-height:100%;font-family:'Bitstream Vera Sans';-inkscape-font-specification:'Bitstream Vera Sans, Normal';text-align:start;writing-mode:lr;text-anchor:start;fill:#ff00ff;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;\"",
"         xml:space=\"preserve\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan11469\">key:&lt;Primary&gt;&lt;Alt&gt;backslash</tspan></text>",
"      <text",
"         sodipodi:linespacing=\"100%\"",
"         id=\"text3010-7\"",
"         y=\"165\"",
"         x=\"5\"",
"         style=\"font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;font-size:12.00000048px;line-height:100%;font-family:'Bitstream Vera Sans';-inkscape-font-specification:'Bitstream Vera Sans, Normal';text-align:start;writing-mode:lr;text-anchor:start;fill:#008080;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;\"",
"         xml:space=\"preserve\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan11471\">key:&lt;Shift&gt;&lt;Alt&gt;bar</tspan></text>",
"    </g>",
"    <g",
"       transform=\"translate(2477.8571,710.00001)\"",
"       id=\"g3698-7\">",
"      <rect",
"         style=\"fill:#0000ff;fill-opacity:0.1;stroke:#000000;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1\"",
"         id=\"rect3700-3\"",
"         width=\"170\"",
"         height=\"170\"",
"         x=\"0\"",
"         y=\"2.9921875\" />",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;font-size:48.00000191px;line-height:100%;font-family:'Bitstream Vera Sans';-inkscape-font-specification:'Bitstream Vera Sans, Normal';text-align:start;writing-mode:lr;text-anchor:start;fill:#000000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;\"",
"         x=\"40\"",
"         y=\"42.992188\"",
"         id=\"text3702-2\"",
"         sodipodi:linespacing=\"100%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan11563\">#</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-style:normal;font-weight:normal;font-size:12.00000048px;font-family:'Bitstream Vera Sans';fill:#000000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;-inkscape-font-specification:'Bitstream Vera Sans, Normal';font-stretch:normal;font-variant:normal;text-anchor:start;text-align:start;writing-mode:lr;line-height:125%;\"",
"         x=\"5\"",
"         y=\"60\"",
"         id=\"text3706-5\"",
"         sodipodi:linespacing=\"125%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan11565\">key:numbersign</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-style:normal;font-weight:normal;font-size:12.00000048px;font-family:'Bitstream Vera Sans';fill:#ff0000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;-inkscape-font-specification:'Bitstream Vera Sans, Normal';font-stretch:normal;font-variant:normal;text-anchor:start;text-align:start;writing-mode:lr;line-height:125%;\"",
"         x=\"5\"",
"         y=\"75\"",
"         id=\"text3710-6\"",
"         sodipodi:linespacing=\"125%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan11567\">key:&lt;Primary&gt;numbersign</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-style:normal;font-weight:normal;font-size:12.00000048px;font-family:'Bitstream Vera Sans';fill:#008000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;-inkscape-font-specification:'Bitstream Vera Sans, Normal';font-stretch:normal;font-variant:normal;text-anchor:start;text-align:start;writing-mode:lr;line-height:125%;\"",
"         x=\"5\"",
"         y=\"90\"",
"         id=\"text3714-8\"",
"         sodipodi:linespacing=\"125%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan11569\">key:&lt;Shift&gt;asciitilde</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-style:normal;font-weight:normal;font-size:12.00000048px;font-family:'Bitstream Vera Sans';fill:#0000ff;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;-inkscape-font-specification:'Bitstream Vera Sans, Normal';font-stretch:normal;font-variant:normal;text-anchor:start;text-align:start;writing-mode:lr;line-height:125%;\"",
"         x=\"5\"",
"         y=\"105\"",
"         id=\"text3718-9\"",
"         sodipodi:linespacing=\"125%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan11571\">key:&lt;Alt&gt;numbersign</tspan></text>",
"      <text",
"         sodipodi:linespacing=\"125%\"",
"         id=\"text6625\"",
"         y=\"120\"",
"         x=\"5\"",
"         style=\"font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;font-size:12.00000048px;line-height:125%;font-family:'Bitstream Vera Sans';-inkscape-font-specification:'Bitstream Vera Sans, Normal';text-align:start;writing-mode:lr;text-anchor:start;fill:#464646;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;\"",
"         xml:space=\"preserve\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan6629\">key:&lt;Super&gt;numbersign</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;font-size:12.00000048px;line-height:100%;font-family:'Bitstream Vera Sans';-inkscape-font-specification:'Bitstream Vera Sans, Normal';text-align:start;writing-mode:lr;text-anchor:start;fill:#808000;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;\"",
"         x=\"5\"",
"         y=\"135\"",
"         id=\"text3722-0\"",
"         sodipodi:linespacing=\"100%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan11573\">key:&lt;Primary&gt;&lt;Shift&gt;asciitilde</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;font-size:12.00000048px;line-height:100%;font-family:'Bitstream Vera Sans';-inkscape-font-specification:'Bitstream Vera Sans, Normal';text-align:start;writing-mode:lr;text-anchor:start;fill:#ff00ff;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;\"",
"         x=\"5\"",
"         y=\"150\"",
"         id=\"text3726-1\"",
"         sodipodi:linespacing=\"100%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan11575\">key:&lt;Primary&gt;&lt;Alt&gt;numbersign</tspan></text>",
"      <text",
"         xml:space=\"preserve\"",
"         style=\"font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;font-size:12.00000048px;line-height:100%;font-family:'Bitstream Vera Sans';-inkscape-font-specification:'Bitstream Vera Sans, Normal';text-align:start;writing-mode:lr;text-anchor:start;fill:#008080;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1;\"",
"         x=\"5\"",
"         y=\"165\"",
"         id=\"text3730-4\"",
"         sodipodi:linespacing=\"100%\"><tspan",
"           sodipodi:role=\"line\"",
"           id=\"tspan11577\">key:&lt;Shift&gt;&lt;Alt&gt;asciitilde</tspan></text>",
"    </g>",
"    <text",
"       id=\"text3106-8-2\"",
"       y=\"468.38324\"",
"       x=\"5\"",
"       style=\"font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;font-size:12.00000095px;line-height:125%;font-family:'Bitstream Vera Sans';-inkscape-font-specification:'Bitstream Vera Sans, Normal';text-align:start;writing-mode:lr-tb;text-anchor:start;fill:#464646;fill-opacity:1;stroke:none;stroke-width:1pt;stroke-linecap:butt;stroke-linejoin:miter;stroke-opacity:1\"",
"       xml:space=\"preserve\"",
"       sodipodi:linespacing=\"125%\"><tspan",
"         sodipodi:role=\"line\"",
"         id=\"tspan6191-5\"",
"         x=\"5\"",
"         y=\"468.38324\" /></text>",
"  </g>",
"</svg>",
NULL,
};
